#include "EG3000.h"
#include "../eg3000/i_eg3000.h"
#include "hardware/control3030.hpp"
#include "mainwindow.h"
#include <QSettings>
#include <QTimer>
#include <QApplication>
#include <QFile>

USB_READ_THREAD *_rThread = 0;
extern USB_WRITE_THREAD *_wThread;//from modules.cpp
extern Control3030 *_ControllerStub3030; //from modules.cpp

double settings_double(QSettings &sett, const QString &path, double def)
{
    bool ok = false;
    double val = sett.value(path, def).toDouble(&ok);
    return ok ? val : def;
}


inline
double linear_calib(double value, double from_min, double from_max, double to_min, double to_max)
{
    return (value - from_min) / (from_max - from_min) * (to_max - to_min) + to_min;
}

QHash<const char *, double> analog_map_min, analog_map_max;
//relies on cc/ld duplicate string compression
void init_analog_calib_cache()
{
    analog_map_min.clear();
    analog_map_max.clear();

    QSettings settings("Param/analog.conf", QSettings::IniFormat);

    QList<const char *> lst;
    lst << "adc/v1" << "adc/v1mag" << "adc/v2" << "adc/v2mag" << "adc/v3" << "adc/v4" << "adc/v5" << "adc/v1hr" << "adc/v2hr" << "dac/vtip";
    foreach (const char *name, lst) {
        double a = settings_double(settings, QString() + name + "_min", -10);
        double b = settings_double(settings, QString() + name + "_max", 10);

        analog_map_min.insert(name, a);
        analog_map_max.insert(name, b);
    }
}

QHash<BYTE, double> analog_ch_gain;
double input_gain(BYTE ch)
{
    return analog_ch_gain.value(ch, 1);
}
void input_gain_set(BYTE ch, double gain)
{
    analog_ch_gain.insert(ch, gain);
}
void input_gain_clear(BYTE ch)
{
    analog_ch_gain.remove(ch);
}

QHash<BYTE, double> analog_ch_shift;
double input_shift(BYTE ch)
{
    return analog_ch_shift.value(ch, 0);
}
void input_shift_set(BYTE ch, double shift)
{
    analog_ch_shift.insert(ch, shift);
}
void input_shift_clear(BYTE ch)
{
    analog_ch_shift.remove(ch);
}

//from [a..b] to [a0..b0]
double analog_calib(double value, const char *name, double a0 = -10., double b0 = 10.)
{
    QSettings settings("Param/analog.conf", QSettings::IniFormat);
    double a = settings_double(settings, QString() + name + "_min", a0);
    double b = settings_double(settings, QString() + name + "_max", b0);
    return linear_calib(value, a, b, a0, b0);
}

double analog_calib_inverse(double value, const char *name, double a0 = -10., double b0 = 10., double a = -10., double b = 10.)
{
    QSettings settings("Param/analog.conf", QSettings::IniFormat);
    a = settings_double(settings, QString() + name + "_min", a);
    b = settings_double(settings, QString() + name + "_max", b);
    return linear_calib(value, a0, b0, a, b);
}
double analog_calib_inverse_cached(double value, const char *name, double a0 = -10., double b0 = 10., double a = -10., double b = 10.)
{
    if (!analog_map_min.contains(name) || !analog_map_max.contains(name))
        return analog_calib_inverse(value, name, a0, b0, a, b);

    a = analog_map_min.value(name, a);
    b = analog_map_max.value(name, b);
    return linear_calib(value, a0, b0, a, b);
}

//z nlc
/*
template<class T>
double NewtonResult(T arr[5], double x0, double value, double ac, DWORD max = 100)
{
    double x_new = x0;
    WORD N = 0;

    do {
        x0 = x_new;
        x_new = x0 -
                (((((arr[0] * x0 + arr[1]) * x0 + arr[2]) * x0 + arr[3]) * x0 + arr[4]) - value) /
                (((4*arr[0] * x0 + 3*arr[1]) * x0 + 2*arr[2]) * x0 + arr[3]);
        N++;
    } while ((fabs(x0 - x_new) > ac) && (N < max));

    return x_new;
}

double ZheightCalibRangeCL = 1;
double ZheightShiftCL = 0;
inline void init_reverse_znlc_params(float *Scanner_data_pid_cal_range)
{
    ZheightShiftCL = NewtonResult(Scanner_data_pid_cal_range + 21, 10, 0, 0.001); //21-25.26
    ZheightCalibRangeCL = NewtonResult(Scanner_data_pid_cal_range + 21, 10, 65535, 0.001) - ZheightShiftCL;
}
inline double nonlinear_zheight_to_mkm(float c, float *Scanner_data_pid_cal_range)
{
    double C = NewtonResult(Scanner_data_pid_cal_range + 21, 10, c, 0.0001); //21-25.26
    return (1. - (C - ZheightShiftCL) / ZheightCalibRangeCL) * Scanner_data_pid_cal_range[47] * 1e-6;
}*/
//

double input_calib(BYTE ch, double val, double zmax)
{
    double gain = input_gain(ch);
    double shift = input_shift(ch);

    const char *path = "";
    switch (ch) {
    default:
        return (linear_calib(val, 0, 65535, -10, 10) + shift) * gain;
        break;
    case srcHeight: {
        /*if (*(_wThread->FB_Scanner ? _wThread->Nlc_status_ol_z2 : _wThread->Nlc_status_ol_z)) {
            float *calib = _wThread->FB_Scanner ? _wThread->Scanner2_data_pid_cal_range : _wThread->Scanner_data_pid_cal_range;
            init_reverse_znlc_params(calib);
            double mkm = nonlinear_zheight_to_mkm(val, calib);
            if (qIsFinite(mkm)) return mkm;
        }*/
        return (65535.0 - val) / 65535. * zmax * 1e-6;
    }
        break;
    case srcADC1:
    case srcDFL:
        path = "adc/v1";
        break;
    case srcMag1:
    case srcMagCos1:
    case srcMagSin1:
        path = "adc/v1mag";
        break;
    case srcADC2:
    case srcLFL:
        path = "adc/v2";
        break;
    case srcMag2:
        path = "adc/v2mag";
        break;
    case srcADC3:
    case srcIntensity:
        path = "adc/v3";
        break;
    case srcADC4:
        path = "adc/v4";
        break;
    case srcADC5:
        path = "adc/v5";
        break;
    case srcADCHR1:
        path = "adc/v1hr";
        break;
    case srcADCHR2:
        path = "adc/v2hr";
        break;
    case srcVTip:
        path = "dac/vtip";
        break;
    }

    return (analog_calib_inverse_cached(val, path, 0., 65535.) + shift) * gain;
}


QHash<BYTE, QString> analog_ch_unit;
QString input_units(BYTE ch)
{
    return analog_ch_unit.value(ch, ch == srcHeight ? QObject::tr("m") : QObject::tr("V"));
}
void input_unit_set(BYTE ch, QString unit)
{
    analog_ch_unit.insert(ch, unit);
}
void input_unit_clear(BYTE ch)
{
    analog_ch_unit.remove(ch);
}


//inline //not in gcc
extern
double setpoint_limit_top(double value, bool cl_on)
{
    if (cl_on)
        return value;

    QSettings settings("Param/controller.conf", QSettings::IniFormat);
    double k = settings_double(settings, "limit_scaner_range", 0);
    if (k < 0.01)
        return value;
    return value*(1 - k);
}


void go_x(double x_value)
{
    _wThread->Go_to_X(qMin(qMax(0., x_value), (double)_wThread->Scanner_data_pid_cal_range[45]));
}
void go_x2(double x_value)
{
    _wThread->Go_to_X2(qMin(qMax(0., x_value), (double)_wThread->Scanner2_data_pid_cal_range[45]));
}
void go_y(double y_value)
{
    _wThread->Go_to_Y(qMin(qMax(0., y_value), (double)_wThread->Scanner_data_pid_cal_range[46]));
}
void go_y2(double y_value)
{
    _wThread->Go_to_Y2(qMin(qMax(0., y_value), (double)_wThread->Scanner2_data_pid_cal_range[46]));
}
void go_relative_x(double x_diff)
{
    go_x(_wThread->Xsp + x_diff);
}
void go_relative_x2(double x_diff)
{
    go_x2(_wThread->Xsp2 + x_diff);
}
void go_relative_y(double y_diff)
{
    go_y(_wThread->Ysp + y_diff);
}
void go_relative_y2(double y_diff)
{
    go_y2(_wThread->Ysp2 + y_diff);
}

double get_objective_z()
{
    QSettings settings("Param/controller.conf", QSettings::IniFormat);
    int af_z_channel = settings.value("video_af_z_channel", 1).toInt();
    if (!_ControllerStub3030 && (af_z_channel == 2 || af_z_channel == 3)) af_z_channel = 1;

    switch (af_z_channel) {
    case 0:
        return _wThread->Zsp;
        break;
    case 1:
        return _wThread->Zsp2;
        break;
    case 2:
        return _ControllerStub3030->Zsp;
        break;
    case 3:
        return _ControllerStub3030->Zsp2;
        break;
    case 4:
        return _wThread->Zsp3;
        break;
    default:
        return _wThread->Zsp2;
        break;
    };
}
double get_max_objective_z()
{
    QSettings settings("Param/controller.conf", QSettings::IniFormat);
    int af_z_channel = settings.value("video_af_z_channel", 1).toInt();
    if (!_ControllerStub3030 && (af_z_channel == 2 || af_z_channel == 3)) af_z_channel = 1;

    switch (af_z_channel) {
    case 0:
        return _wThread->Scanner_data_pid_cal_range[47];
        break;
    case 1:
        return _wThread->Scanner2_data_pid_cal_range[47];
        break;
    case 2:
        return _ControllerStub3030->Scanner_data_pid_cal_range[47];
        break;
    case 3:
        return _ControllerStub3030->Scanner2_data_pid_cal_range[47];
        break;
    case 4:
        return _wThread->Scanner3_data_pid_cal_range[47];
        break;
    default:
        return _wThread->Scanner2_data_pid_cal_range[47];
        break;
    };
}
void go_relative_objective_z(double z_diff)
{
    QSettings settings("Param/controller.conf", QSettings::IniFormat);
    int af_z_channel = settings.value("video_af_z_channel", 1).toInt();
    if (!_ControllerStub3030 && (af_z_channel == 2 || af_z_channel == 3)) af_z_channel = 1;

    switch (af_z_channel) {
    case 0:
        z_diff += _wThread->Zsp;
        break;
    case 1:
        z_diff += _wThread->Zsp2;
        break;
    case 2:
        z_diff += _ControllerStub3030->Zsp;
        break;
    case 3:
        z_diff += _ControllerStub3030->Zsp2;
        break;
    case 4:
        z_diff += _wThread->Zsp3;
        break;
    default:
        z_diff += _wThread->Zsp2;
        break;
    };

    go_objective_z(z_diff);
}
void go_objective_z(double z_value)
{
    QSettings settings("Param/controller.conf", QSettings::IniFormat);
    int af_z_channel = settings.value("video_af_z_channel", 1).toInt();
    if (!_ControllerStub3030 && (af_z_channel == 2 || af_z_channel == 3)) af_z_channel = 1;

    switch (af_z_channel) {
    case 0:
        _wThread->Go_to_Z(qMin(qMax(0., z_value), get_max_objective_z()));
        break;

    case 1:
        _wThread->Go_to_Zobject(qMin(qMax(0., z_value), get_max_objective_z()));
        break;

    case 2:
        //TODO:copypaste head xyz positioning - now same as stage
        _ControllerStub3030->Go_to_Zobject(qMin(qMax(0., z_value), get_max_objective_z()));
        break;

    case 3:
        _ControllerStub3030->Go_to_Zobject(qMin(qMax(0., z_value), get_max_objective_z()));
        break;

    case 4:
        _wThread->Go_to_Z3(qMin(qMax(0., z_value), get_max_objective_z()));
        break;

    default:
        _wThread->Go_to_Zobject(qMin(qMax(0., z_value), get_max_objective_z()));
        break;
    };
}

bool validateHex(QString filename)
{
    QFile file(filename);
    if (!file.open(QIODevice::ReadOnly | QIODevice::Text)) {
        qDebug() << "validateHex can't open file";
        return false;
    }

    bool end = false;
    int c = 0;
    while (!file.atEnd()) {
        c++;
        QByteArray line = file.readLine();
        line.chop(1); // windows, "\cr\lf\", \cr remove

        if (line.size() < 11) {
            qDebug() << "validateHex short line" << c;
            file.close();
            return false;
        }
        if (line.at(0) != ':') {
            qDebug() << "validateHex no ':'" << c;
            file.close();
            return false;
        }
        if ((line.size() - 11) != 2*((QString(line.at(1)).toInt(0, 16) << 4) + QString(line.at(2)).toInt(0, 16))) {
            qDebug() << "validateHex wrong data len" << c;
            file.close();
            return false;
        }
        if (line.at(7) != '0') {
            qDebug() << "validateHex wrong record type1" << c;
            file.close();
            return false;
        }
        if (line.at(8) != '0' && line.at(8) != '1' && line.at(8) != '2' &&
                line.at(8) != '3' && line.at(8) != '4' && line.at(8) != '5') {
            qDebug() << "validateHex wrong record type2" << c;
            file.close();
            return false;
        }
        if (line.at(8) != '0' &&
                (line.at(3) != '0' ||
                 line.at(4) != '0' ||
                 line.at(5) != '0' ||
                 line.at(6) != '0')) {
            qDebug() << "validateHex address not 0" << c;
            file.close();
            return false;
        }

        if (line.at(8) == '1')
            end = true;
        if (end && !file.atEnd()) {
            qDebug() << "validateHex end record not last" << c;
            file.close();
            return false;
        }

        unsigned long checksum = 0;
        for (int i = 0; i < (line.size() - 1)/2; i++) {
            checksum += (QString(line.at(2*i+1)).toInt(0, 16) << 4) + QString(line.at(2*i+2)).toInt(0, 16);
        }
        if ((0xFF & checksum) != 0) {
            qDebug() << "validateHex wrong checksum" << c;
            file.close();
            return false;
        }
    }

    if (!end)
        qDebug() << "validateHex no eof record";

    return end;
}

unsigned long bufferHex(QString filename, unsigned char *buffer, unsigned int max_size)
{
    QFile file(filename);
    if (!file.open(QIODevice::ReadOnly | QIODevice::Text)) {
        qDebug() << "bufferHex can't open file";
        return 0;
    }

    memset(buffer, 0xFF, max_size);

    unsigned long max_addr = 0;
    unsigned long high_addr = 0;
    while (!file.atEnd()) {
        QByteArray line = file.readLine();
        line.chop(1); // windows, "\cr\lf\", \cr remove

        if (line.at(8) == '0') {
            if (high_addr < 0x1D00 || high_addr > 0x1D08)
                continue;

            unsigned long size = (QString(line.at(1)).toInt(0, 16) << 4) + QString(line.at(2)).toInt(0, 16);
            unsigned long addr = (((QString(line.at(3)).toInt(0, 16) << 4) + QString(line.at(4)).toInt(0, 16)) << 8) +
                    (QString(line.at(5)).toInt(0, 16) << 4) + QString(line.at(6)).toInt(0, 16);
            addr += (high_addr - 0x1D00) << 16;

            for (unsigned int i = 0; i < size && addr + i < max_size; i++) {
                buffer[addr + i] = (QString(line.at(i*2+9)).toInt(0, 16) << 4) + QString(line.at(i*2+10)).toInt(0, 16);
            }

            if (addr + size >= max_size)
                qDebug() << "bufferHex overflow!" << addr + size << max_size;

            if (addr + size > max_addr)
                max_addr = addr + size;
        } else if (line.at(8) == '1') {
            break;
        } else if (line.at(8) == '4') {
            high_addr = (((QString(line.at(9)).toInt(0, 16) << 4) + QString(line.at(10)).toInt(0, 16)) << 8) +
                    (QString(line.at(11)).toInt(0, 16) << 4) + QString(line.at(12)).toInt(0, 16);
        }
    }

    return max_addr;
}

//
void *ftHandle = NULL;
struct eg3000_read_handle read_handle;

//-----------USB Thread---------------------
USB_CONTROL::USB_CONTROL()
{
    m_closeThread = false;
    m_waitMainInit = true;
    need_program = false;

    init_analog_calib_cache();

    eg3000_init_read_handle(&read_handle);
}
USB_CONTROL::~USB_CONTROL()
{
    eg3000_close_read_handle(&read_handle);
}

bool program_hex(QString filename)
{
    if (!validateHex(filename))
        return false;

    unsigned char *buf = new unsigned char[520000];
    unsigned size = 4 + ((~3) & bufferHex(filename, buf, 520000));

    bool ok = RET_TRUE == eg3000_bl_program(ftHandle, buf, size, &read_handle, 0);
    eg3000_bl_stats(ftHandle);
    delete[] buf;

    return ok;
}
BYTE invert_byte(BYTE b)
{ //from SO sth
    b = (b & 0xF0) >> 4 | (b & 0x0F) << 4; //swap every 4 bits
    b = (b & 0xCC) >> 2 | (b & 0x33) << 2; //swap every 2 bits
    b = (b & 0xAA) >> 1 | (b & 0x55) << 1; //swap every 1 bits
    return b;
 }
bool program_bin(QString filename)
{
    QFile bin(filename);
    bin.open(QIODevice::ReadOnly);
    QByteArray buf = bin.readAll();
    bin.close();

    if (filename.endsWith(".rpd")) //flash lsb first
        for (int i = 0; i < buf.size(); i++)
            buf[i] = invert_byte(buf.at(i));

    bool ok = RET_TRUE == eg3000_bl_program(ftHandle, (BYTE *)buf.data(), buf.size(), &read_handle, 1);
    return ok;
}

QMutex read_lock;
QMutex write_lock;

void USB_CONTROL::program(QString val)
{
    need_program = true;
    path = val;
    emit device_status(2);
}

void USB_CONTROL::s_boot_timeout()
{
    boot_timeout = true;
}

extern QMutex *ftdi_mutex;
bool eg_open(WORD vid, WORD pid, bool open_small = false, bool open_motors = false) {
    WORD size;
    bool st;

    QMutexLocker locker(ftdi_mutex);
    st = eg3000_get_number(vid, pid, &size);
    if (!st || size == 0)
        return false;

    struct s_info *array;
    try {
        array = new struct s_info[size];
    } catch (...) {
        qDebug() << "eg_open new exception";
        return false;
    }

    st = eg3000_list(vid, pid, &size, array, size);
    if (st && size != 0) {
        st = false; //not found
        for (int i = 0; i < size; i++) {
            if (array[i].flags == (open_motors ? F_MOTORS : F_NO)) {
                st = eg3000_open(array[i].handle, vid, pid, &ftHandle);
                break;
            }
        }

        if (!st && open_small) {
            for (int i = 0; i < size; i++) {
                if (array[i].flags & F_EG1000) {
                    st = eg3000_open(array[i].handle, vid, pid, &ftHandle);
                    break;
                }
            }
        }
    }

    delete [] array;
    return st;
}

bool motor_driver = false;
void USB_CONTROL::run()
{
    m_closeThread = false;
    m_waitMainInit = true;

    if (testMode) {
        while (m_waitMainInit && !m_closeThread) {
            msleep(50);
        }

        emit updateUsbStatus(true);
        emit usbStatus(EDok);

        while (!m_closeThread) {
            msleep(50);
        }

        return;
    }

    boot_timeout_t = new QTimer();
    boot_timeout_t->setInterval(2500);
    boot_timeout_t->setSingleShot(true);
    connect(boot_timeout_t, SIGNAL(timeout()), this, SLOT(s_boot_timeout()), Qt::QueuedConnection);
    connect(this, SIGNAL(restart_boot_timer()), boot_timeout_t, SLOT(start()), Qt::QueuedConnection);
    connect(this, SIGNAL(destroyed()), boot_timeout_t, SLOT(stop()), Qt::QueuedConnection);
    connect(this, SIGNAL(destroyed()), boot_timeout_t, SLOT(deleteLater()), Qt::QueuedConnection);
    boot_timeout_t->moveToThread(qApp->thread());

    bool Packet_Recieved = false;
    DWORD PacketSize = 0;
    BYTE Address = 0;
    BYTE SubAddress = 0;

    bool need_boot = true;
    bool is_flash_good = true;

    st = false;
    boot_timeout = false;
    int prevStatus = -1;

#ifndef _NO_FTDI
    bool use_4000 = false;
#else
    bool use_4000 = true;
#endif
    bool slept = true;

    while (!m_closeThread) {
        if (st) st = eg3000_is_ok(ftHandle);

        if (st && need_program && !need_boot) {
            emit updateUsbStatus(false);
            prevStatus = EDconn;
            emit usbStatus(EDconn);

            while (_wThread->isRunning())
                msleep(10);

            eg3000_fb_control(ftHandle, fbOff); //disable feedback
            eg3000_scan_startstop(ftHandle, scHead, 1); //start scan, to replace scan struct, for correct scan off
            eg3000_scan_startstop(ftHandle, scHead, 0); //stop all scans & ccd
            eg3000_osc_startstop(ftHandle, 0); //stop oscilloscope
            eg3000_status_onoff(ftHandle, 0); //disable spm info send
            eg3000_set_head_dac(ftHandle, 32767); //set tip voltage ~0
            eg3000_send_word_packet(ftHandle, ADR_STEP_MOTORS, SMD_MOTOR_ITOA(1) + REG_CONTROL, 0);
            eg3000_send_word_packet(ftHandle, ADR_STEP_MOTORS, SMD_MOTOR_ITOA(2) + REG_CONTROL, 0);
            eg3000_send_word_packet(ftHandle, ADR_STEP_MOTORS, SMD_MOTOR_ITOA(3) + REG_CONTROL, 0);
            eg3000_send_word_packet(ftHandle, ADR_STEP_MOTORS, SMD_MOTOR_ITOA(4) + REG_CONTROL, 0);
            BYTE stp = 7;
            eg3000_send_packet(ftHandle, ADR_THERMO, 0, &stp, 1);
            eg3000_hv_zero(ftHandle); //hv outs to 0
            qDebug() << "sent controller shutdown";

            msleep(500);

            eg3000_send_empty_packet(ftHandle, ADR_CPU, CPU_GO_REBOOT);
            qDebug() << "sent controller reset";

            need_boot = true;
            is_flash_good = true;

            msleep(500);

            eg3000_clear(ftHandle);
            eg3000_clear_read_handle(&read_handle);

            st = eg3000_is_ok(ftHandle);
        }

        if (st && need_boot && !m_closeThread) {
            read_lock.lock();
            write_lock.lock();

            if (!need_program) {
                //query
                eg3000_bl_stats(ftHandle); //ping bootloader
                eg3000_cpu_loopback(ftHandle); //ping cpu
                st = eg3000_is_ok(ftHandle);
            }

            emit restart_boot_timer();
            boot_timeout = false;

            if (!(st && need_boot && !m_closeThread)) {
                read_lock.unlock();
                write_lock.unlock();
            }
        }
        bool boot_retried = false;
        BYTE boot_timeout_count = 0;
        while (st && need_boot && !m_closeThread) {
            st = eg3000_update_parse(ftHandle, &read_handle);
            Packet_Recieved = read_handle.packet_recieved;
            PacketSize = read_handle.packet_size;
            Address = read_handle.address;
            SubAddress = read_handle.sub_address;

            //packet parsing
            if (Packet_Recieved) {
                if (Address == ADR_BOOTLOADER) {
                    if (SubAddress == BL_ERROR) {
                        qDebug() << "bl error" << (void*)((BYTE *)read_handle.buffer)[0];
                    }
                    if (SubAddress == BL_STATS) {
                        //unsigned int flash_size = ((DWORD*)read_handle.buffer)[0];
                        is_flash_good = ((BYTE *)read_handle.buffer)[4];

                        if (!need_program) {
                            if (!is_flash_good) {
                                if (boot_timeout_count == 0) {
                                    if (!m_waitMainInit)
                                        boot_timeout_count = 4; //skip second warning
                                    qDebug() << "no flash in controller!";
                                    emit device_status(0);
                                }
                            } else { //auto_boot
                                st = eg3000_bl_run(ftHandle);
                                boot_retried = false;
                            }
                        }
                    }
                    if (SubAddress == BL_NO_FLASH) {
                        if (!need_program) {
                            if (boot_retried) {
                                if (boot_timeout_count == 0) {
                                    if (!m_waitMainInit)
                                        boot_timeout_count = 4; //skip second warning
                                    qDebug() << "no flash in controller!";
                                    emit device_status(0);
                                }
                            } else {
                                boot_retried = true;
                                msleep(100);
                                st = eg3000_bl_run(ftHandle);
                            }
                        }
                    }

                    if (SubAddress == BL_BOOT_OK) { //booted
                        st = eg3000_cpu_loopback(ftHandle); //ping
                    }
                }
                if (Address == ADR_CPU) {
                    need_boot = false;
                }

                eg3000_discard_packet(&read_handle);
            } else
                msleep(10);

            //program
            if (need_program && need_boot) {
                emit device_status(3);
                bool ok = path.endsWith(".hex") ? program_hex(path) : program_bin(path);
                st = eg3000_is_ok(ftHandle);
                if (ok && st) {
                    need_program = false;
                    emit device_status(4);
                } else {
                    qDebug() << "retry program";
                    emit device_status(2);
                }

                emit restart_boot_timer();
                boot_timeout = false;
                boot_timeout_count = 0;

                msleep(100);
            }

            if (st && (boot_timeout || read_handle.packet_malformed) && need_boot && !m_closeThread && is_flash_good) {
                emit restart_boot_timer();
                boot_timeout = false;

                if (read_handle.packet_malformed) {
                    eg3000_clear_read_handle(&read_handle);
                    eg3000_send_empty_packet(ftHandle, ADR_CPU, CPU_USB_BUF_RST);
                    msleep(100);
                } else if (!m_waitMainInit)
                    boot_timeout_count++;

                //resend query
                eg3000_bl_stats(ftHandle); //ping bootloader
                eg3000_cpu_loopback(ftHandle); //ping cpu
                st = eg3000_is_ok(ftHandle);

                if (boot_timeout_count == 3) {
                    qDebug() << "boot timeout";
                    emit device_status(1);
                }
            }

            if (!(st && need_boot && !m_closeThread)) {
                read_lock.unlock();
                write_lock.unlock();
            }
        }

        if (!m_waitMainInit) {
            emit updateUsbStatus(st);
            int stat = st ? (need_boot ? EDconn : EDok) : EDno;
            if (prevStatus != stat) {
                prevStatus = stat;
                emit usbStatus(stat);
            }
        }

        if (!st && ftHandle != NULL) {
            read_lock.lock();
            write_lock.lock();

            if (eg3000_is_ok(ftHandle))
                eg3000_send_empty_packet(ftHandle, ADR_CPU, CPU_USB_BUF_RST); //l2k soft reset = need cleanup

            eg3000_close(ftHandle);
            ftHandle = NULL;

            read_lock.unlock();
            write_lock.unlock();
        }

        if (!st && !m_closeThread) {
            QSettings settings("Param/controller.conf", QSettings::IniFormat);
#ifndef _NO_FTDI
            bool use = settings.value("use_eg4000", false).toBool();
            if (use_4000 != use)
                eg3000_set_backend(use ? 1 : 0); //libusb:ftdi
            use_4000 = use;
#endif

            need_boot = true;
            is_flash_good = true;
            motor_driver = false;

            st = eg_open(VID, PID+(use_4000 ? 5 : 0), settings.value("use_3030", false).toBool());
            if (!st && !use_4000 && settings.value("use_motors", false).toBool()) {
                st = eg_open(VID, PID_FTDI, false, true);
                need_boot = !st;
                motor_driver = st;
            }

            if (st) {
                st = eg3000_clear(ftHandle);
                eg3000_clear_read_handle(&read_handle);
                if (slept) {
                    slept = false;
                    continue;
                }
            }
        }

        slept = true;
        for (int i = 0; i < (m_waitMainInit ? 1 : 10) && !m_closeThread; i++)
            msleep(50);
    }

    if (st && !need_boot) {
        qDebug() << "fin reset commands sent";
        eg3000_fb_control(ftHandle, fbOff); //disable feedback
        eg3000_scan_startstop(ftHandle, scHead, 1); //start scan, to replace scan struct, for correct scan off
        eg3000_scan_startstop(ftHandle, scHead, 0); //stop all scans & ccd
        eg3000_osc_startstop(ftHandle, 0); //stop oscilloscope
        eg3000_status_onoff(ftHandle, 0); //disable spm info send
        eg3000_set_head_dac(ftHandle, 32767); //set tip voltage ~0
        eg3000_send_word_packet(ftHandle, ADR_STEP_MOTORS, SMD_MOTOR_ITOA(1) + REG_CONTROL, 0);
        eg3000_send_word_packet(ftHandle, ADR_STEP_MOTORS, SMD_MOTOR_ITOA(2) + REG_CONTROL, 0);
        eg3000_send_word_packet(ftHandle, ADR_STEP_MOTORS, SMD_MOTOR_ITOA(3) + REG_CONTROL, 0);
        eg3000_send_word_packet(ftHandle, ADR_STEP_MOTORS, SMD_MOTOR_ITOA(4) + REG_CONTROL, 0);
        eg3000_send_word_packet(ftHandle, ADR_CPU, CPU_LAND_M4_Z, 0);
        BYTE stp = 7;
        eg3000_send_packet(ftHandle, ADR_THERMO, 0, &stp, 1);
        eg3000_hv_zero(ftHandle); //hv outs to 0
        eg3000_send_empty_packet(ftHandle, ADR_CPU, CPU_USB_BUF_RST);
    }

    if (ftHandle != NULL) {
        eg3000_close(ftHandle);
        ftHandle = NULL;
    }
}

void USB_CONTROL::thread_close(USB_WRITE_THREAD *wThread, USB_READ_THREAD *rThread)
{
    wThread->thread_close();
    rThread->thread_close();
    while (wThread->isRunning() && rThread->isRunning()) {
        qApp->processEvents();
        msleep(3);
    }

    m_closeThread = true;

    while (isRunning()) {
        qApp->processEvents();
        msleep(3);
    }
}
void USB_CONTROL::drop_connect()
{
    st = false;
}
void USB_CONTROL::SetTestMode(bool value)
{
    testMode=value;
}

//=====================����� ������ � USB=================
USB_WRITE_THREAD::USB_WRITE_THREAD()
{
    for (int i = 0; i < 4; i++)
        Active_Ch[i] = 0;

    motorRst = 0;
    updateSensorOF = 0;

    updateQcontrolGain = false;
    updateDFLXtalk = false;
    updateMAGXtalk = false;
    updateKelvin = false;
    updatePiezoTip = false;
    updatePiezoTip2 = false;
    updateCapMode = false;
    updateCapMode2 = false;
    updateEFMplus = false;
    updatelflCurrent = false;
    updateextCurrent = false;
    updatedrive2 = false;
    updatefreq2 = false;
    {
        QSettings settings("Param/gui.conf", QSettings::IniFormat);
        kelvinP = settings_double(settings, "kelvin/p", 1);
    }

    head_dac = 32767;
    updateHeadDAC = false;
    updateHeadDAC2 = false;
    updateCCDTime = false;
    updateSPMPrecision = false;
    updateJM_Lift = false;
    updateJM_Precision = false;
    updateJM_div = false;

    updateFirstDelay = false;
    updateLineDelay = false;
    updatePointDelay = false;
    updateFastReturn = false;
    updateSyncMode = false;

    Action_1 = false;
    Action_2 = false;
    Action_3 = false;
    for (int i = 0; i < 3; i++)
        Action_5[i] = false;
    Action_6 = false;
    Action_61 = false;
    Action_65 = false;
    Action_7 = false;
    Action_71 = false;
    Action_72 = false;
    Action_73 = false;
    Action_76 = false;
    Action_77 = false;
    Action_78 = false;
    Action_79 = false;

    Action_8 = false;
    Action_81 = false;
    Action_82 = false;
    Action_83 = false;
    Action_84 = false;
    Action_85 = false;
    Action_852 = false;
    Action_86 = false;
    Action_87 = false;
    Action_88 = false;
    Action_89 = false;
    Action_9 = false;
    Action_10 = false;
    Action_11 = false;
    updLanding = false;
    Action_17 = false;
    Action_18 = false;
    Action_19 = false;
    Action_20 = false;
    Action_21 = false;
    Action_22 = false;
    Action_23 = false;

    sendLMheight = false;
    sendPIDnlc = false;

    QSettings settings("Param/controller.conf", QSettings::IniFormat);
    csb_speed_idx = qMin(3, settings.value("csb_speed_settings_idx", 1).toInt());
    csb_speed_extended = false;

    {
        QSettings settings("Param/gui.conf", QSettings::IniFormat);
        scanner_scan = settings.value("display/oscilloscope/head", true).toBool() ? scHead : scStage;
    }
    fast_scan_correction = false;
    scan_precise_mode = false;

    RawMotor_pwm = false;
    RawMotor_bl_speed = false;
    RawMotorLandSafeCalib = false;
    updateCsbMotorAddr = updateCsbMotorCount = false;
    updMotor[0] = false;
    RawMotor_control[0] = false;
    for (int i = 0; i < 8; i++) {
        RawMotor_inv_end[i] = false;
        RawMotor_bl_steps[i] = false;
        updateRemapMotor[i] = false;
        updMotor[i+1] = false;
        RawMotor_control[i+1] = false;
    }
    RawMotor = false;

    Raw_Thermo = false;
    doGetXY = false;
    doGetCZ = false;
    doGetData = false;
    updateLandingM4 = false;


    FB_sign = 0;
    FB_Scanner = scHead;
    FBZ_config = fbOff;
    Spm_FB_Channel = srcDFL;
    Spm_I = 0;
    Spm_P = 0.1;

    LmEn=0;
    LmEnA=0;
    LmEnB=0;
    LmEnC=0;
    LmEnD=0;
    LmHeight=0;

    thread_started = false;
    thread_closed = false;

    LithoMode = ltmNone;
    LithoDACIndex=0;
    LithoScanner = scHead;
    LithoContinuous = 0;
    LithPacketInd=0;
    newShape=false;
    Xsp=0; Ysp=0; Zsp=0;
    Xsp2=0;Ysp2=0;Zsp2=0;
    Zsp3 = 0;
}

//---------------------- Extra calibration---------------------------------
void USB_WRITE_THREAD::SetExtraCalibOut(double *adcCalib_p, double *dacCalib_p)
{
    adcCalib = adcCalib_p;
    dacCalib = dacCalib_p;
}
void USB_WRITE_THREAD::SendExtraCalib()
{
    Action_22=true;
}
void USB_WRITE_THREAD::sl_csb_speed(int idx)
{
    csb_speed_idx = idx;
    Action_23 = true;
}
void USB_WRITE_THREAD::ccd_readout_time(WORD time_ms)
{
    m_ccd_readout_time = time_ms;
    updateCCDTime = true;
}
void USB_WRITE_THREAD::scan_first_delay(double time)
{
    m_scan_first_delay = time*1000.;
    updateFirstDelay = true;
}
void USB_WRITE_THREAD::scan_line_delay(double time)
{
    m_scan_line_delay = time*1000.;
    updateLineDelay = true;
}
void USB_WRITE_THREAD::scan_point_delay(double time)
{
    m_scan_point_delay = time*1000.*10.;
    updatePointDelay = true;
}
void USB_WRITE_THREAD::scan_fast_return(bool on)
{
    m_scan_fast_return = on;
    updateFastReturn = true;
}
void USB_WRITE_THREAD::scan_sync_mode(int mode)
{
    m_scan_sync_mode = mode;
    updateSyncMode = true;
}
void USB_WRITE_THREAD::spm_pid_precision(double delta)
{
    m_spm_pid_precision = 65535. * delta / 20. ;
    updateSPMPrecision = true;
}
void USB_WRITE_THREAD::spm_jm_lift(DWORD dz)
{
    m_jm_lift = dz;
    updateJM_Lift = true;
}
void USB_WRITE_THREAD::spm_jm_precision(WORD msb, WORD lsb)
{
    m_jm_precision = msb;
    m_jm_precision_lsb = lsb;
    updateJM_Precision = true;
}
void USB_WRITE_THREAD::spm_jm_div(int div)
{
    m_jm_div = div;
    updateJM_div = true;
}
void USB_WRITE_THREAD::ScanCorrection(int st)
{
    fast_scan_correction = st;
}
void USB_WRITE_THREAD::scanPreciseMode(bool enabled)
{
    scan_precise_mode = enabled;
}
void USB_WRITE_THREAD::SetTestMode(bool value)
{
    testMode=value;
}
void USB_WRITE_THREAD::start_stop(bool st)
{
#if (QT_VERSION < QT_VERSION_CHECK(5, 0, 0))
    if (testMode && st && !thread_started)
        emit started();
#else
#pragma message ("fuck qt5 private signals")
#endif
    thread_started = st;
    if (st && !testMode)
        start(QThread::HighPriority);
}
void USB_WRITE_THREAD::thread_close()
{
    thread_closed = true;

    while (isRunning()) {
        qApp->processEvents();
        msleep(3);
    }
}

void USB_WRITE_THREAD::Set_Scan_Parameters(unsigned short int NXn, unsigned short int NYn,
                                           double Vscn, double Step_Xn, double Step_Yn, double St_point_Xn, double St_point_Yn,
                                           unsigned char scan_dir_st_pointn, unsigned char scan_an, unsigned char scan_bn, unsigned char scan_cn, unsigned char scan_dn,
                                           bool LmEn_n, bool LmEnA_n, bool LmEnB_n, bool LmEnC_n, bool LmEnD_n, double LmHeight_n)
{
    NX = NXn;
    NY = NYn;
    Vsc = Vscn;
    Step_X = Step_Xn;
    Step_Y = Step_Yn;
    St_point_X = St_point_Xn;
    St_point_Y = St_point_Yn;
    scan_dir_st_point = scan_dir_st_pointn;
    scan_a = scan_an;
    scan_b = scan_bn;
    scan_c = scan_cn;
    scan_d = scan_dn;

    LmEn=LmEn_n;
    LmEnA=LmEnA_n;
    LmEnB=LmEnB_n;
    LmEnC=LmEnC_n;
    LmEnD=LmEnD_n;
    LmHeight=LmHeight_n;

    Action_7 = true;
}
void USB_WRITE_THREAD::SetccdScanMode(int sc, int mode, int nx, int ny, int st_pos, double x0, double y0, double step_x, double step_y, bool fast_scan)
{
    ccdNX = nx;
    ccdNY = ny;
    ccdStep_X = step_x;
    ccdStep_Y = step_y;
    ccdSt_point_X = x0;
    ccdSt_point_Y = y0;
    ccdscan_dir_st_point = st_pos;
    ccdScanner = sc;
    ccdMode=mode;
    ccdFastScan=fast_scan;
    Action_76 = true;
}
//++++++++++++++++++++LOCK_IN+++++++++++++++++++++++++++++++++++++
void USB_WRITE_THREAD::freq_scan_start(double start_n, double end_n, int Channel)
{
    fr_start = start_n;
    fr_end = end_n;
    fr_channel=Channel;

    Action_81 = true;
    if (testMode) emit sig_lock_test();

}
void USB_WRITE_THREAD::set_freq(double freq_n)
{
    current_frequency = freq_n * 1000; //khz->hz
    Action_8 = true;
}
void USB_WRITE_THREAD::set_cutoff_freq(double cutoff_n)
{
    cutoff = cutoff_n;
    Action_82 = true;
}
void USB_WRITE_THREAD::set_ampl_gains(int gain1_n, int gain2_n, int gain3_n)
{
    gain1 = gain1_n;
    gain2 = gain2_n;
    gain3 = gain3_n;
    Action_85 = true;
}
void USB_WRITE_THREAD::set_ampl_gains2(int gain1_n, int gain2_n, int gain3_n)
{
    gain12 = gain1_n;
    gain22 = gain2_n;
    gain32 = gain3_n;
    Action_852 = true;
}
void USB_WRITE_THREAD::set_phase(double phase_n)
{
    phase = phase_n;
    Action_83 = true;
}
void USB_WRITE_THREAD::set_piezodrive_volt(double volt_n)
{
    volt_n = analog_calib(volt_n, "dac/vdrive", 0.);
    volt = volt_n;
    Action_84 = true;
}

void USB_WRITE_THREAD::setLVM(BYTE head, BYTE stage)
{
    lvm_bits = (head ? (0x3 << 0) : 0) + (stage ? (0x3 << 2) : 0);
    updLVM = true;
}

void USB_WRITE_THREAD::setSensorOverflowMode(BYTE scanner, BYTE mode)
{
    SensorOFon = mode;
    updateSensorOF = scanner+1;
}

void USB_WRITE_THREAD::setKelvinOn(bool val)
{
    kelvinOn = val;
    updateKelvin = true;
}
void USB_WRITE_THREAD::setKelvinP(double val)
{
    QSettings settings("Param/gui.conf", QSettings::IniFormat);
    settings.setValue("kelvin/p", val);

    kelvinP = val;
    updateKelvin = true;
}

void USB_WRITE_THREAD::setDFLXtalk(double val)
{
    dfl_xtalk = analog_calib(val / input_gain(srcDFL), "adc/v1") / 20 * 65535;
    updateDFLXtalk = true;
}
void USB_WRITE_THREAD::setMAGXtalk(double val)
{
    mag_xtalk = analog_calib(val / input_gain(srcMag1), "adc/v1mag") / 20 * 65535;
    updateMAGXtalk = true;
}

void USB_WRITE_THREAD::setQcontrolGain(double val)
{
    qcontrol_gain = val;
    updateQcontrolGain = true;
}

void USB_WRITE_THREAD::setPiezoTip(bool on)
{
    piezoTip = on;
    updatePiezoTip = true;
}
void USB_WRITE_THREAD::setPiezoTip2(bool on)
{
    piezoTip2 = on;
    updatePiezoTip2 = true;
}
void USB_WRITE_THREAD::setLFLCurrent(bool on)
{
    lflCurrent = on;
    updatelflCurrent = true;
}
void USB_WRITE_THREAD::setExtCurrent(bool on)
{
    extCurrent = on;
    updateextCurrent = true;
}
void USB_WRITE_THREAD::setCapMode(bool on)
{
    capmode = on;
    updateCapMode = true;
}
void USB_WRITE_THREAD::setCapMode2(bool on)
{
    capmode2 = on;
    updateCapMode2 = true;
}
void USB_WRITE_THREAD::setEFMplus(bool on)
{
    efmplus = on;
    updateEFMplus = true;
}
void USB_WRITE_THREAD::setDrive2(double val)
{
    val = analog_calib(val, "dac/vdrive", 0.);
    drive2 = val;
    updatedrive2 = true;
}
void USB_WRITE_THREAD::setFreq2(double val)
{
    freq2 = val;
    updatefreq2 = true;
}

void USB_WRITE_THREAD::set_Ex_dac1(double value)
{
    value = analog_calib(value, "dac/v1");
    Ex_dac1 = value;
    Action_86 = true;
}
void USB_WRITE_THREAD::set_Ex_dac2(double value)
{
    value = analog_calib(value, "dac/v2");
    Ex_dac2 = value;
    Action_87 = true;
}
void USB_WRITE_THREAD::set_Ex_dac3(double value)
{
    value = analog_calib(value, "dac/v3hr");
    Ex_dac3 = value;
    Action_88 = true;
}
void USB_WRITE_THREAD::set_Ex_dac4(double value)
{
    value = analog_calib(value, "dac/v4hr");
    Ex_dac4 = value;
    Action_89 = true;
}

void USB_WRITE_THREAD::set_raw_motors_control(unsigned char motor, unsigned short control)
{
    if (motor == 250) {
        set_raw_motors_control(1, control);
        set_raw_motors_control(2, control);
        set_raw_motors_control(3, control);
        return;
    }
    motor_cont_byte[motor] = control;
    RawMotor_control[motor] = true;
}
void USB_WRITE_THREAD::set_raw_motors_inv_end(unsigned char motor, unsigned short inv)
{
    if (motor == 250) {
        set_raw_motors_inv_end(0, inv);
        set_raw_motors_inv_end(1, inv);
        set_raw_motors_inv_end(2, inv);
        return;
    }
    motor_inv_byte[motor] = inv;
    RawMotor_inv_end[motor] = true;
}

void USB_WRITE_THREAD::set_step_motors_leveling(int motor, unsigned char motor_cont_byte_n, double motor_step_n, double motor_speed_n)
{
    if (motor_cont_byte_n == smStop && motor != 250) {
        set_raw_motors_control(motor, motor_cont_byte_n);
        return;
    }
    if (motor == 250) motor = 0;
    motor_step[motor] = motor_step_n;
    motor_speed[motor] = motor_speed_n;
    motor_cont_byte[motor] = motor_cont_byte_n;
    RawMotor_control[motor] = false;
    updMotor[motor] = true;
}
void USB_WRITE_THREAD::set_landing(BYTE mode, BYTE direction, double speed, double speed_slow, double close_perc)
{
    Landing_mode = mode;
    Landing_direction = direction;
    Landing_speed = speed;
    Landing_speed_close = speed_slow;
    Landing_close_perc = close_perc;
    updLanding = true;
}
void USB_WRITE_THREAD::set_landing_m4(int val)
{
    landing_m4_mode = val;
    updateLandingM4 = true;
}

//+++++++++++Spectroscopy+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
void USB_WRITE_THREAD::set_spectroscopy(int st, double z1,double z2, double speed, int channel, int points)
{
    spect_onn_off = st;
    Z1spectr=z1;
    Z2spectr=z2;
    SpeedSpectr=speed;
    SpectrChannel=channel;
    SpectrPoints=points;
    Action_17 = true;
    if (testMode) emit sig_Spectr_test();
}

//+++++++++++Head data++++++++++++++++++++++++++++++
void USB_WRITE_THREAD::sl_Head_Data(int LaserOnOf_n, int AC_DC_n, int Preamp_n)
{
    LaserOnOf = LaserOnOf_n;
    AC_DC = AC_DC_n;
    Preamp=Preamp_n;
    Action_18=true;
}

void USB_WRITE_THREAD::HVXYZzero()
{
    Action_19 = true;
}

//---------SPM Feed Back----------------------------------------
void USB_WRITE_THREAD::Set_FB_sign(bool sign)
{
    if (sign) FB_sign=1;
    else      FB_sign=0;
    Action_61=true;
}

void USB_WRITE_THREAD::Set_FB_Scanner(int sc)
{
    FB_Scanner = sc;
    Action_61 = true;
}

void USB_WRITE_THREAD::SET_SPM_P(double Pk)
{
    Spm_P = Pk/100.;
    Action_61 = true;
}
void USB_WRITE_THREAD::SET_SPM_I(double Ik)
{
    Spm_I = Ik/10000.;
    Action_61 = true;
}
void USB_WRITE_THREAD::SET_SPM_Sp(double Sp)
{
    Spm_SetPoint = Sp;
    Action_65 = true;
}
void USB_WRITE_THREAD::SET_SPM_FB_Channel(int Channel)
{
    Spm_FB_Channel = Channel;
    Action_61 = true;
}

//scan
void USB_WRITE_THREAD::set_scanner(int sc)
{
    scanner_scan = sc;
}
void USB_WRITE_THREAD::SCAN_Start(void)
{
    emit sig_scan_test(true);

    if (!testMode)
        Action_71 = true;
}
void USB_WRITE_THREAD::SCAN_Stop(void)
{
    emit sig_scan_test(false);

    if (!testMode)
        Action_72 = true;
}
void USB_WRITE_THREAD::Set_Speed(double V)
{
    Vsc = V;
    Action_73 = true;
    if (testMode)
        _rThread->scan_wait = qMax<int>(1, 1000/4/Vsc);
}
void USB_WRITE_THREAD::scan_lm_height(double h)
{
    scan_lm_h = h/1000.;
    sendLMheight = true;
}

//-----------------------CCD Scan--------------------
void USB_WRITE_THREAD::CCD_SCAN_Start(void)
{
    if (testMode)
        emit sig_scan_test_ccd(true);
    if (ccdMode != ccdSingleScanMode)
        emit sig_scan_test(true);

    if (!testMode)
        Action_77 = true;
}
void USB_WRITE_THREAD::CCD_SCAN_Stop(void)
{
    if (testMode)
        emit sig_scan_test_ccd(false);
    if (ccdMode != ccdSingleScanMode)
        emit sig_scan_test(false);

    if (!testMode)
        Action_78 = true;
}
void USB_WRITE_THREAD::CCD_SCAN_Ready(void)
{
    if (testMode) {
        emit sig_scan_test_ccd_ready();
        //_rThread->sl_set_scan_test_ccd_ready();
    }
    else
        Action_79 = true;
}
//----------------lithography---------------------------------------

void USB_WRITE_THREAD::setLithResOut(const graphed::out::TResultsList* res)
{
    lithoResList=res;
}

void USB_WRITE_THREAD::slLitoStartStop(bool sig)
{
    if (!sig) {
        //send canceled
        Action_21 = true;
        return;
    }

    LithoPar.LithoMode = LithoMode;
    LithoPar.Scanner = LithoScanner;
    LithoPar.isContinuous = LithoContinuous;
    LithoPar.DACIndex = LithoDACIndex;

    ResIt = lithoResList->begin();
    newShape = true;

    sl_LithoNextpacket();
}

void USB_WRITE_THREAD::sl_LithoNextpacket()
{
    if (ResIt == lithoResList->end()) {
        emit sig_LithoTotalReady();
        return;
    }

    LithPacketInd = 0;

    double xr, yr;
    if (LithoScanner == 0) {
        xr = Scanner_data_pid_cal_range[45];
        yr = Scanner_data_pid_cal_range[46];
    } else {
        xr = Scanner2_data_pid_cal_range[45];
        yr = Scanner2_data_pid_cal_range[46];
    }

    graphed::out::TResults &Result = **ResIt;
    if (newShape) {
        PointIt = Result.PointList().begin();
        newShape = false;
    }

    LithoPar.Speed = Result.Speed();
    LithoPar.V = Result.VtData1(PointIt).first;
    LithoPar.T = Result.VtData1(PointIt).second;
    if (LithoMode == ltmSync) LithoPar.V = 1;
    if (Spm_FB_Channel >= 4 && Spm_FB_Channel <= 6) LithoPar.V -= 10;

    for (; PointIt != Result.PointList().end(); PointIt++) {
        QPointF fl1 = *PointIt;
        LithoPar.XY[LithPacketInd*2] = 1e6*fl1.rx();
        LithoPar.XY[LithPacketInd*2+1] = yr-1e6*fl1.ry();
        LithPacketInd++;

        if (LithPacketInd >= LITHO_DATA_SIZE) {
            PointIt++;
            break;
        }
        if ((PointIt+1 != Result.PointList().end()) &&
                fabs(Result.VtData1(PointIt).first - Result.VtData1(PointIt+1).first) >= 1e-5) {
            PointIt++;
            break;
        }
    }

    if (PointIt == Result.PointList().end()) {
        ResIt++;
        newShape = true;
    }

    LithoPar.Size = LithPacketInd;
    Action_20 = true;

    if (testMode) {
        emit sig_Lit_test();
    }
}

void USB_WRITE_THREAD::set_Litho_continuous(bool st)
{
    LithoContinuous = !st;
}
void USB_WRITE_THREAD::set_Litho_Type(int lt)
{
    LithoMode = lt+1;
}
void USB_WRITE_THREAD::set_Litho_Scanner(int sc)
{
    LithoScanner = sc;
}
void USB_WRITE_THREAD::set_Litho_VOut(int out)
{
    LithoDACIndex = out;
}

void USB_WRITE_THREAD::sl_get_curr_xy()
{
    doGetXY = true;
}
void USB_WRITE_THREAD::sl_get_curr_cz()
{
    doGetCZ = true;
}
void USB_WRITE_THREAD::sl_get_curr_data(int ch)
{
    getDataCh = ch;
    doGetData = true;
}

//------------------------------------------------------------------
void USB_WRITE_THREAD::Set_Scanner_Data_in(float* scanner_data_pid_cal_range, DWORD *scanner_data_scale, long* scanner_data_Time,
                                           bool* nlc_c_xy, bool* ncl_xy, bool* nlc_c_z, bool* ncl_z,
                                           bool* nlc_o_xy, bool* nlc_o_z, float *_free_move_speed,
                                           float* scanner2_data_pid_cal_range, DWORD *scanner2_data_scale, long* scanner2_data_Time,
                                           bool* nlc2_c_xy, bool* ncl2_xy, bool* nlc2_c_z, bool* ncl2_z,
                                           bool* nlc2_o_xy, bool* nlc2_o_z, float *_free_move_speed2)
{
    Scanner_data_pid_cal_range = scanner_data_pid_cal_range;
    Scanner_data_scale = scanner_data_scale;
    Scanner_data_Shift = scanner_data_Time;

    Nlc_status_cl_xy = nlc_c_xy;
    CL_status_xy = ncl_xy;
    Nlc_status_cl_z = nlc_c_z;
    CL_status_z = ncl_z;
    Nlc_status_ol_xy = nlc_o_xy;
    Nlc_status_ol_z = nlc_o_z;

    free_move_speed = _free_move_speed;

    Scanner2_data_pid_cal_range = scanner2_data_pid_cal_range;
    Scanner2_data_scale = scanner2_data_scale;
    Scanner2_data_Shift = scanner2_data_Time;

    Nlc_status_cl_xy2 = nlc2_c_xy;
    CL_status_xy2 = ncl2_xy;
    Nlc_status_cl_z2 = nlc2_c_z;
    CL_status_z2 = ncl2_z;
    Nlc_status_ol_xy2 = nlc2_o_xy;
    Nlc_status_ol_z2 = nlc2_o_z;

    free_move_speed2 = _free_move_speed2;
}
void USB_WRITE_THREAD::Osc_Start()
{
    Action_11 = false;
    Action_10 = true;
    if (testMode) emit sig_osc_test(true);
}
void USB_WRITE_THREAD::Osc_Stop(void)
{
    Action_11 = true;
    if (testMode) emit sig_osc_test(false);

}
void USB_WRITE_THREAD::Data_to_Cont()
{
    Action_2 = true;
}
void USB_WRITE_THREAD::Save_to_EEPROM(void)
{
    Action_3 = true;
}
void USB_WRITE_THREAD::Usb_Load_Data()
{
    Action_1 = true;
}

void USB_WRITE_THREAD::Go_to_XYZ(double Xsp_n, double Ysp_n, double Zsp_n)
{
    Xsp = Xsp_n;
    Ysp = Ysp_n;
    Zsp = Zsp_n;

    emit xUpdate(Xsp);
    emit yUpdate(Ysp);
    emit zUpdate(Zsp);
    Action_5[0] = true;
}
void USB_WRITE_THREAD::Go_to_XYZ2(double Xsp_n, double Ysp_n, double Zsp_n)
{
    Xsp2 = Xsp_n;
    Ysp2 = Ysp_n;
    Zsp2 = Zsp_n;

    emit x2Update(Xsp2);
    emit y2Update(Ysp2);
    emit z2Update(Zsp2);
    Action_5[1] = true;
}
void USB_WRITE_THREAD::Go_to_XYZ_CCD(int sc, double Xsp_n, double Ysp_n)
{
    if (sc!=0)
    {
        if (sc-1 == scHead) Go_to_XYZ(Xsp_n,Ysp_n,Zsp);
        else                 Go_to_XYZ2(Xsp_n,Ysp_n,Zsp2);
    }
}

void USB_WRITE_THREAD::Go_to_Z(double Zsp_n)
{
    Zsp = Zsp_n;
    emit zUpdate(Zsp);
    Action_5[0] = true;
}
void USB_WRITE_THREAD::Go_to_Zobject(double Zsp_n)
{
    Zsp2 = Zsp_n;
    emit z2Update(Zsp2);
    Action_5[1] = true;
}
void USB_WRITE_THREAD::Go_to_Z3(double Zsp_n)
{
    Zsp3 = Zsp_n;
    emit z3Update(Zsp3);
    Action_5[2] = true;
}
void USB_WRITE_THREAD::Go_to_X(double Xsp_n)
{
    Xsp = Xsp_n;
    emit xUpdate(Xsp);
    Action_5[0] = true;
}
void USB_WRITE_THREAD::Go_to_X2(double Xsp_n)
{
    Xsp2 = Xsp_n;
    emit x2Update(Xsp2);
    Action_5[1] = true;
}
void USB_WRITE_THREAD::Go_to_Y(double Ysp_n)
{
    Ysp = Ysp_n;
    emit yUpdate(Ysp);
    Action_5[0] = true;
}
void USB_WRITE_THREAD::Go_to_Y2(double Ysp_n)
{
    Ysp2 = Ysp_n;
    emit y2Update(Ysp2);
    Action_5[1] = true;
}
void USB_WRITE_THREAD::FBZ_send(unsigned char FBZ_config_n)
{
    FBZ_config=FBZ_config_n;
    Action_6 = true;
}
//------------------------------------------------------------
void USB_WRITE_THREAD::Set_osc_Channel(int idx, int ch)
{
    Active_Ch[idx] = ch;
    Action_9 = true;
}
void USB_WRITE_THREAD::Set_headDAC(double val)
{
    val = analog_calib(val, "dac/vtip");
    head_dac = 65535.0 * (val + 10.0) / 20.0;
    updateHeadDAC = true;
}
void USB_WRITE_THREAD::Set_headDAC2(double val)
{
    val = analog_calib(val, "dac/vtip");
    head_dac2 = 65535.0 * (val + 10.0) / 20.0;
    updateHeadDAC2 = true;
}

//------------------------------------------------------------------------
//TODO members
BYTE StepMotorDirs[16];
BYTE StepMotorRemap[16];

void StepMotorRun(BYTE MotorN, DWORD Steps, WORD Speed, WORD StepSize, BYTE Direction) {
    eg3000_raw_motor_i2c(ftHandle, SMD_MOTOR_ITOA(MotorN) + REG_STEPS_HW, Steps >> 16);
    eg3000_raw_motor_i2c(ftHandle, SMD_MOTOR_ITOA(MotorN) + REG_STEPS_LW, Steps);
    eg3000_raw_motor_i2c(ftHandle, SMD_MOTOR_ITOA(MotorN) + REG_SPEED, Speed);
    eg3000_raw_motor_i2c(ftHandle, SMD_MOTOR_ITOA(MotorN) + REG_CONTROL, StepSize + RC_ENABLED + RC_RUN + Direction);
    StepMotorDirs[MotorN - 1] = Direction & RC_DIR_CCW; //direction flag mask
}
void StepMotorRunForever(BYTE MotorN, WORD Speed, WORD StepSize, BYTE Direction) {
    eg3000_raw_motor_i2c(ftHandle, SMD_MOTOR_ITOA(MotorN) + REG_SPEED, Speed);
    eg3000_raw_motor_i2c(ftHandle, SMD_MOTOR_ITOA(MotorN) + REG_CONTROL, StepSize + RC_ENABLED + RC_RUN + RC_FOREVER + Direction);
    StepMotorDirs[MotorN - 1] = Direction & RC_DIR_CCW; //direction flag mask
}
void StepMotorStop(BYTE MotorN) {
    eg3000_raw_motor_i2c(ftHandle, SMD_MOTOR_ITOA(MotorN) + REG_CONTROL, 0);
}

void SetMotor(BYTE motor, BYTE mode, DWORD steps, WORD speed) {
    switch (mode) {
        case smStop:
            StepMotorStop(motor); // 
            break;
        case smUp: // 
            StepMotorRun(motor, steps, speed, STEP_16, RC_DIR_CW);
            break;
        case smDown: // 
            StepMotorRun(motor, steps, speed, STEP_16, RC_DIR_CCW);
            break;
        case smUpToLimit: //   
            StepMotorRunForever(motor, speed, STEP_16, RC_DIR_CW);
            break;
        case smDownToLimit: //   
            StepMotorRunForever(motor, speed, STEP_16, RC_DIR_CCW);
            break;
    }
}

//------------------------------------------------------------------------
void USB_WRITE_THREAD::run()
{
    if (motor_driver) { //motor command preprocessing
        for (int i = 0; i < 16; i++) StepMotorRemap[i] = i+1;

        while (thread_started && !thread_closed) {
            if (!write_lock.tryLock(10)) {
                msleep(1);
                continue;
            }

            _wThread->Action_5[0] = _wThread->Action_5[1] = false;

            if (RawMotor_pwm) {
                RawMotor_pwm = false;
                eg3000_raw_motor_i2c(ftHandle, SMD_MOTOR_ITOA(1) + REG_PWM_DC, motor_pwm);
            }
            if (RawMotor_bl_speed) {
                RawMotor_bl_speed = false;
                eg3000_raw_motor_i2c(ftHandle, SMD_MOTOR_ITOA(1) + REG_BL_SPEED, motor_bl_speed);
            }
            for (int i = 0; i < 8; i++) {
                if (updateRemapMotor[i]) {
                    updateRemapMotor[i] = false;
                    StepMotorRemap[i] = remap_motor_to[i];
                    if (!StepMotorRemap[i]) StepMotorRemap[i] = i+1;
                }
                if (RawMotor_inv_end[i]) {
                    RawMotor_inv_end[i] = false;
                    eg3000_raw_motor_i2c(ftHandle, SMD_MOTOR_ITOA(StepMotorRemap[i]) + REG_INV_END, motor_inv_byte[i]);
                }
                if (RawMotor_bl_steps[i]) {
                    RawMotor_bl_steps[i] = false;
                    eg3000_raw_motor_i2c(ftHandle, SMD_MOTOR_ITOA(StepMotorRemap[i]) + REG_BL_STEPS_H, (motor_bl_steps[i] >> 16) & 0xFFFF);
                    eg3000_raw_motor_i2c(ftHandle, SMD_MOTOR_ITOA(StepMotorRemap[i]) + REG_BL_STEPS_L, motor_bl_steps[i] & 0xFFFF);
                }
            }
            for (int i = 0; i < 8+1; i++) {
                if (updMotor[i]) {
                    updMotor[i] = false;
                    SetMotor(StepMotorRemap[i == 0 ? 0 : i-1], motor_cont_byte[i], motor_step[i], motor_speed[i]);
                    if (i == 0) {
                        SetMotor(StepMotorRemap[1], motor_cont_byte[i], motor_step[i], motor_speed[i]);
                        SetMotor(StepMotorRemap[2], motor_cont_byte[i], motor_step[i], motor_speed[i]);
                    }
                }
                if (RawMotor_control[i] && i != 0) {
                    RawMotor_control[i] = false;
                    eg3000_raw_motor_i2c(ftHandle, SMD_MOTOR_ITOA(StepMotorRemap[i-1]) + REG_CONTROL, motor_cont_byte[i]);
                }
            }

            write_lock.unlock();
            msleep(1);
        }

        motorRst = 0;
        return;
    }

    if (thread_started && !thread_closed) {
        write_lock.lock();

        QSettings settings("Param/controller.conf", QSettings::IniFormat);
        qDebug() << "init reset commands sent";

        eg3000_fb_control(ftHandle, fbOff); //disable feedback

        eg3000_scan_params(ftHandle, scHead, 30, 30,
                                1., 1., 1., 0., 0.,
                                spTopLeftAxisX, srcNone, srcNone, srcNone, srcNone,
                                false, false, false, false, false, 0., false, 0.); //send default head params
        eg3000_scan_params(ftHandle, scStage, 30, 30,
                                1., 1., 1., 0., 0.,
                                spTopLeftAxisX, srcNone, srcNone, srcNone, srcNone,
                                false, false, false, false, false, 0., false, 0.); //send default stage params
        eg3000_scan_startstop(ftHandle, scHead, 1); //start scan, to replace scan struct, for correct scan off
        eg3000_scan_startstop(ftHandle, scHead, 0); //stop all scans & ccd
        eg3000_osc_startstop(ftHandle, 0); //stop oscilloscope
        eg3000_osc_set_data(ftHandle, 0, 0, 0, 0);

        eg3000_status_onoff(ftHandle, 1);
        eg3000_set_head_dac(ftHandle, 32767); //set tip voltage ~0
        eg3000_send_word_packet(ftHandle, ADR_STEP_MOTORS, SMD_MOTOR_ITOA(1) + REG_CONTROL, 0);
        eg3000_send_word_packet(ftHandle, ADR_STEP_MOTORS, SMD_MOTOR_ITOA(2) + REG_CONTROL, 0);
        eg3000_send_word_packet(ftHandle, ADR_STEP_MOTORS, SMD_MOTOR_ITOA(3) + REG_CONTROL, 0);
        eg3000_send_word_packet(ftHandle, ADR_STEP_MOTORS, SMD_MOTOR_ITOA(4) + REG_CONTROL, 0);
        eg3000_send_word_packet(ftHandle, ADR_CPU, CPU_LAND_M4_Z, 0);

        csb_speed_extended = false;
        BYTE data = csb_speed_idx = qMin(3, settings.value("csb_speed_settings_idx", 1).toInt());
        eg3000_send_packet(ftHandle, ADR_CPU, CPU_SET_CSB_RATE, &data, 1);

        data = settings.value("double_csb_rate", false).toBool() ? 1 : 0;
        eg3000_send_packet(ftHandle, ADR_CPU, CPU_CSB_DUALSPEED, &data, 1);

        data = settings.value("double_c_output", 0).toInt();
        eg3000_csb_double_c_out(ftHandle, scStage, data);

        eg3000_send_empty_packet(ftHandle, ADR_CPU, CPU_CAP_BITS);

        write_lock.unlock();
    }

    bool st = eg3000_is_ok(ftHandle);
    while (thread_started && !thread_closed && st) {
        if (!write_lock.tryLock(10)) {
            msleep(1);
            continue;
        }

        if (updateSensorOF) {
            BYTE v = updateSensorOF - 1;
            updateSensorOF = 0;
            eg3000_csb_overflow(ftHandle, v, SensorOFon);
        }
        if (updateKelvin) {
            updateKelvin = false;
            eg3000_kelvin_params(ftHandle, kelvinOn, kelvinP / 100.);
        }
        if (updateDFLXtalk) {
            updateDFLXtalk = false;
            eg3000_send_packet(ftHandle, ADR_CPU, CPU_DFL_XTALK, &dfl_xtalk, sizeof(float));
        }
        if (updateMAGXtalk) {
            updateMAGXtalk = false;
            eg3000_send_packet(ftHandle, ADR_CPU, CPU_MAG_XTALK, &mag_xtalk, sizeof(float));
        }
        if (updateQcontrolGain) {
            updateQcontrolGain = false;
            float qcoeff = -qcontrol_gain;
            if (qcontrol_gain > 0)
                qcoeff /= 5;
            eg3000_send_packet(ftHandle, ADR_CPU, CPU_SET_QCONTROL, &qcoeff, sizeof(float));
        }
        if (updatePiezoTip) {
            updatePiezoTip = false;
            eg3000_set_head_piezotip(ftHandle, piezoTip);
        }
        if (updatePiezoTip2) {
            updatePiezoTip2 = false;
            eg3000_set_head_piezotip2(ftHandle, piezoTip2);
        }
        if (updatelflCurrent) {
            updatelflCurrent = false;
            eg3000_set_head_lfl_current(ftHandle, lflCurrent);
        }
        if (updateextCurrent) {
            updateextCurrent = false;
            eg3000_set_head_ext_current(ftHandle, extCurrent);
        }
        if (updateCapMode) {
            updateCapMode = false;
            eg3000_set_cap_mode(ftHandle, capmode);
        }
        if (updateCapMode2) {
            updateCapMode2 = false;
            eg3000_set_cap_mode2(ftHandle, capmode2);
        }
        if (updateEFMplus) {
            updateEFMplus = false;
            eg3000_send_word_packet(ftHandle, ADR_CPU, CPU_EFM_PLUS, efmplus);
        }
        if (updatedrive2) {
            updatedrive2 = false;
            eg3000_set_piezo_drive_voltage_lift(ftHandle, drive2);
        }
        if (updatefreq2) {
            updatefreq2 = false;
            eg3000_set_frequency2(ftHandle, freq2*1000.0f);
        }
        if (updateHeadDAC) {
            updateHeadDAC = false;
            eg3000_set_head_dac(ftHandle, head_dac); //set tip voltage
        }
        if (updateHeadDAC2) {
            updateHeadDAC2 = false;
            eg3000_set_head_dac2(ftHandle, head_dac2); //set tip voltage
        }
        if (updateCCDTime) {
            updateCCDTime = false;
            eg3000_ccd_readout_time(ftHandle, m_ccd_readout_time);
        }
        if (updateFirstDelay) {
            updateFirstDelay = false;
            eg3000_send_word_packet(ftHandle, ADR_CPU, CPU_FIRST_DELAY, m_scan_first_delay);
        }
        if (updateLineDelay) {
            updateLineDelay = false;
            eg3000_send_word_packet(ftHandle, ADR_CPU, CPU_LINE_DELAY, m_scan_line_delay);
        }
        if (updatePointDelay) {
            updatePointDelay = false;
            eg3000_send_word_packet(ftHandle, ADR_CPU, CPU_SCAN_POINT_DELAY, m_scan_point_delay);
        }
        if (updateFastReturn) {
            updateFastReturn = false;
            eg3000_send_word_packet(ftHandle, ADR_CPU, CPU_FAST_RETURN, m_scan_fast_return ? 1 : 0);
        }
        if (updateSyncMode) {
            updateSyncMode = false;
            eg3000_send_word_packet(ftHandle, ADR_CPU, CPU_SCAN_WAIT_SYNC, m_scan_sync_mode);
        }
        if (updateSPMPrecision) {
            updateSPMPrecision = false;
            eg3000_spm_pid_precision(ftHandle, m_spm_pid_precision);
        }
        if (updateJM_Precision) {
            updateJM_Precision = false;
            eg3000_send_word_packet(ftHandle, ADR_CPU, CPU_JM_PREC_LSB, m_jm_precision_lsb);
            eg3000_send_word_packet(ftHandle, ADR_CPU, CPU_JM_PREC, m_jm_precision);
        }
        if (updateJM_Lift) {
            updateJM_Lift = false;
            eg3000_send_packet(ftHandle, ADR_CPU, CPU_JM_LIFT, &m_jm_lift, sizeof(DWORD));
        }
        if (updateJM_div) {
            updateJM_div = false;
            BYTE d =  m_jm_div;
            eg3000_send_packet(ftHandle, ADR_CPU, CPU_JM_DIV, &d, sizeof(BYTE));
        }
        if (updLVM) {
            updLVM = false;
            eg3000_send_word_packet(ftHandle, ADR_CPU, CPU_HW_LVM, lvm_bits);
        }
        if (sendPIDnlc) {
            sendPIDnlc = false;

            EG_PartCalData cd;
            cd.caldataID = cdNLC_CPID;
            for (int i = 0; i < 6; i++) {
                cd.deviceID = i;
                cd.data = nlcPIDOn[i];
                for (int j = 0; j < 6; j++) {
                    cd.fcaldata[j] = nlcPID[i*6 + j];
                }
                eg3000_send_packet(ftHandle, ADR_CPU, CPU_SET_PARTCALDATA, &cd, sizeof(EG_PartCalData));
            }
        }
        if (Action_1) {
            Action_1 = false;
            eg3000_read_caldata(ftHandle);
        }
        if (Action_2) {
            Action_2 = false;
            eg3000_send_caldata(ftHandle, scHead,
                                     Scanner_data_pid_cal_range, Scanner_data_scale, Scanner_data_Shift,
                                     *Nlc_status_cl_xy, *CL_status_xy, *free_move_speed,
                                     *Nlc_status_cl_z, *CL_status_z, *Nlc_status_ol_xy, *Nlc_status_ol_z);

            eg3000_send_caldata(ftHandle, scStage,
                                     Scanner2_data_pid_cal_range, Scanner2_data_scale, Scanner2_data_Shift,
                                     *Nlc_status_cl_xy2, *CL_status_xy2, *free_move_speed2,
                                     *Nlc_status_cl_z2, *CL_status_z2, *Nlc_status_ol_xy2, *Nlc_status_ol_z2);

            eg3000_send_caldata(ftHandle, scZ3,
                                     Scanner3_data_pid_cal_range, Scanner3_data_scale, Scanner3_data_Shift,
                                     0, *CL3_status_z, *free_move_speed3,
                                     0, *CL3_status_z, 0, 0);
        }
        if (Action_3) {
            Action_3 = false;
            eg3000_save_caldata(ftHandle);
        }
        if (Action_5[0]) {
            Action_5[0] = false;
            eg3000_set_point_xyz(ftHandle, 0,
                                      setpoint_limit_top(Xsp, *CL_status_xy),
                                      setpoint_limit_top(Ysp, *CL_status_xy),
                                      Zsp);
        }
        if (Action_5[1]) {
            Action_5[1] = false;
            eg3000_set_point_xyz(ftHandle, 1,
                                      setpoint_limit_top(Xsp2, *CL_status_xy2),
                                      setpoint_limit_top(Ysp2, *CL_status_xy2),
                                      Zsp2);
        }
        if (Action_5[2]) {
            Action_5[2] = false;
            eg3000_set_point_xyz(ftHandle, 2, 0, 0, Zsp3);
        }
        if (Action_6) {
            Action_6 = false;
            eg3000_fb_control(ftHandle, FBZ_config);
        }
        if (Action_61) {
            QSettings settings("Param/controller.conf", QSettings::IniFormat);
            double scale = settings_double(settings, (FB_Scanner == scHead) ? "mult_spm_pid_head" : "mult_spm_pid_stage", 1.0);
            eg3000_fb_pid_params(ftHandle, FB_Scanner, Spm_P*scale, Spm_I*scale, 0., Spm_FB_Channel, FB_sign);
            Action_61 = false;
        }
        if (Action_65) {
            eg3000_fb_pid_setpoint(ftHandle, Spm_SetPoint);
            Action_65 = false;
        }
        if (Action_7) {
            bool cl_on_xy = scanner_scan == scHead ? *CL_status_xy : *CL_status_xy2;
            eg3000_scan_params(ftHandle, scanner_scan, NX, NY, Vsc,
                                    setpoint_limit_top(Step_X, cl_on_xy),
                                    setpoint_limit_top(Step_Y, cl_on_xy),
                                    setpoint_limit_top(St_point_X, cl_on_xy),
                                    setpoint_limit_top(St_point_Y, cl_on_xy),
                                    scan_dir_st_point, scan_a, scan_b, scan_c, scan_d,
                                    LmEn, LmEnA, LmEnB, LmEnC, LmEnD, LmHeight, fast_scan_correction, scan_precise_mode);
            Action_7 = false;
        }
        if (Action_73) {
            Action_73 = false;
            eg3000_scan_speed(ftHandle, scanner_scan, Vsc);
        }
        if (sendLMheight) {
            sendLMheight = false;
            eg3000_send_packet(ftHandle, ADR_CPU, CPU_SET_SCAN_LM_HEIGHT, &scan_lm_h, sizeof(float));
        }
        if (Action_76) {
            eg3000_ccd_scan_params(ftHandle, ccdScanner, ccdMode, ccdNX, ccdNY, ccdStep_X, ccdStep_Y, ccdSt_point_X, ccdSt_point_Y, ccdscan_dir_st_point, ccdFastScan);
            Action_76 = false;
        }
        if (Action_71) {
            Action_71 = false;
            eg3000_scan_startstop(ftHandle, scanner_scan, 1);
        }
        if (Action_72) {
            Action_72 = false;
            eg3000_scan_startstop(ftHandle, scanner_scan, 0);
        }
        if (Action_77) {
            Action_77 = false;
            eg3000_ccd_scan_startstop(ftHandle, 1);
        }
        if (Action_78) {
            Action_78 = false;
            eg3000_ccd_scan_startstop(ftHandle, 0);
        }
        if (Action_79) {
            Action_79 = false;
            eg3000_ccd_ready(ftHandle);
        }
        if (Action_8) {
            Action_8 = false;
            eg3000_set_frequency(ftHandle, current_frequency);
        }
        if (Action_81) {
            Action_81 = false;
            eg3000_frequency_scan_start(ftHandle, fr_start, fr_end, fr_channel);
        }
        if (Action_82) {
            Action_82 = false;
            eg3000_set_cutoff_frequency(ftHandle, cutoff);
        }
        if (Action_83) {
            Action_83 = false;
            eg3000_set_phase(ftHandle, phase);
        }
        if (Action_84) {
            Action_84 = false;
            eg3000_set_piezo_drive_voltage(ftHandle, volt);
        }
        if (Action_85) {
            Action_85 = false;
            eg3000_set_ampl_gains(ftHandle, gain1, gain2, gain3);
        }
        if (Action_852) {
            Action_852 = false;
            eg3000_set_ampl_gains2(ftHandle, gain12, gain22, gain32);
        }
        if (Action_86) {
            Action_86 = false;
            eg3000_set_extra_DAC1(ftHandle, Ex_dac1);
        }
        if (Action_87) {
            Action_87 = false;
            eg3000_set_extra_DAC2(ftHandle, Ex_dac2);
        }
        if (Action_88) {
            Action_88 = false;
            eg3000_set_extra_DAC3(ftHandle, Ex_dac3);
        }
        if (Action_89) {
            Action_89 = false;
            eg3000_set_extra_DAC4(ftHandle, Ex_dac4);
        }
        if (Action_9) {
            Action_9 = false;
            eg3000_osc_set_data(ftHandle, Active_Ch[0], Active_Ch[1], Active_Ch[2], Active_Ch[3]);
        }
        if (Action_10) {
            Action_10 = false;
            eg3000_osc_startstop(ftHandle, 1);
        }
        if (Action_11) {
            Action_11 = false;
            eg3000_osc_startstop(ftHandle, 0);
        }
        if (motorRst == 1) { //do rst
            motorRst = 2; //rdy
            if (updateCsbMotorAddr) {
                updateCsbMotorAddr = false;
                eg3000_send_word_packet(ftHandle, ADR_CPU, CPU_MOTOR_CSB_ADDR, csb_motor_addr);
            }
            if (updateCsbMotorCount) {
                updateCsbMotorCount = false;
                eg3000_send_word_packet(ftHandle, ADR_CPU, CPU_MOTOR_CSB_COUNT, csb_motor_count);
            }
        }
        if (motorRst == 2) { //ready
            if (RawMotor_pwm) {
                RawMotor_pwm = false;
                eg3000_send_word_packet(ftHandle, ADR_STEP_MOTORS, SMD_MOTOR_ITOA(1) + REG_PWM_DC, motor_pwm); //head
                eg3000_send_word_packet(ftHandle, ADR_STEP_MOTORS, SMD_MOTOR_ITOA(8) + REG_PWM_DC, motor_pwm); //second controller i2c-motor board
            }
            if (RawMotor_bl_speed) {
                RawMotor_bl_speed = false;
                eg3000_send_word_packet(ftHandle, ADR_STEP_MOTORS, SMD_MOTOR_ITOA(1) + REG_BL_SPEED, motor_bl_speed); //head
                eg3000_send_word_packet(ftHandle, ADR_STEP_MOTORS, SMD_MOTOR_ITOA(8) + REG_BL_SPEED, motor_bl_speed); //second controller i2c-motor board
            }
            if (RawMotorLandSafeCalib) {
                RawMotorLandSafeCalib = false;
                eg3000_send_packet(ftHandle, ADR_CPU, CPU_LANDING_MKM, &motor_land_safe_calib, sizeof(DWORD));
            }
            for (int i = 0; i < 8; i++) {
                if (updateRemapMotor[i]) {
                    updateRemapMotor[i] = false;
                    eg3000_send_word_packet(ftHandle, ADR_CPU, CPU_MOTOR_REMAP, (i << 8) + remap_motor_to[i]);
                }
                if (RawMotor_inv_end[i]) {
                    RawMotor_inv_end[i] = false;
                    eg3000_send_word_packet(ftHandle, ADR_STEP_MOTORS,
                                            SMD_MOTOR_ITOA(i + 1) + REG_INV_END, motor_inv_byte[i]);
                }
                if (RawMotor_bl_steps[i]) {
                    RawMotor_bl_steps[i] = false;
                    eg3000_send_word_packet(ftHandle, ADR_STEP_MOTORS, SMD_MOTOR_ITOA(i + 1) + REG_BL_STEPS_H, (motor_bl_steps[i] >> 16) & 0xFFFF);
                    eg3000_send_word_packet(ftHandle, ADR_STEP_MOTORS, SMD_MOTOR_ITOA(i + 1) + REG_BL_STEPS_L, motor_bl_steps[i] & 0xFFFF);
                }
            }
            for (int i = 0; i < 8+1; i++) {
                if (updMotor[i]) {
                    updMotor[i] = false;
                    eg3000_set_motor_leveling(ftHandle, (i == 0 ? 250 : i), motor_cont_byte[i], motor_step[i], motor_speed[i]);
                }
                if (RawMotor_control[i]) {
                    RawMotor_control[i] = false;
                    eg3000_send_word_packet(ftHandle, ADR_STEP_MOTORS,
                                            SMD_MOTOR_ITOA(i) + REG_CONTROL, motor_cont_byte[i]);
                }
            }
            if (RawMotor) {
                eg3000_send_word_packet(ftHandle, ADR_STEP_MOTORS, rawmotoraddr, rawmotordata);
                RawMotor = false;
            }
        }
        if (Raw_Thermo) {
            eg3000_send_packet(ftHandle, ADR_THERMO, 0, thermo_data, 6);
            Raw_Thermo = false;
        }
        if (doGetXY) {
            eg3000_send_word_packet(ftHandle, ADR_CPU, CPU_GET_DATA, srcDACX);
            eg3000_send_word_packet(ftHandle, ADR_CPU, CPU_GET_DATA, srcDACY);
            eg3000_send_word_packet(ftHandle, ADR_CPU, CPU_GET_DATA, srcDACX2);
            eg3000_send_word_packet(ftHandle, ADR_CPU, CPU_GET_DATA, srcDACY2);
            doGetXY = false;
        }
        if (doGetCZ) {
            eg3000_send_word_packet(ftHandle, ADR_CPU, CPU_GET_DATA, srcCZ);
            eg3000_send_word_packet(ftHandle, ADR_CPU, CPU_GET_DATA, srcCZ2);
            doGetCZ = false;
        }
        if (doGetData) {
            eg3000_send_word_packet(ftHandle, ADR_CPU, CPU_GET_DATA, getDataCh);
            doGetData = false;
        }
        if (updateLandingM4) {
            eg3000_send_word_packet(ftHandle, ADR_CPU, CPU_LAND_M4_Z, landing_m4_mode);
            updateLandingM4 = false;
        }
        if (updLanding) {
            eg3000_set_landing(ftHandle, Landing_mode, Landing_direction, Landing_speed, Landing_speed_close, Landing_close_perc);
            updLanding = false;
        }
        if (Action_17) {
            eg3000_set_spectroscopy_params(ftHandle, spect_onn_off, Z1spectr, Z2spectr, SpeedSpectr, SpectrChannel, SpectrPoints);
            Action_17 = false;
        }
        if (Action_18) {
            eg3000_head_params(ftHandle, LaserOnOf, AC_DC, Preamp);
            Action_18 = false;
        }
        if (Action_19) {
            eg3000_hv_zero(ftHandle);
            Action_19 = false;
        }
        if (Action_20) {
            eg3000_litho_start(ftHandle, &LithoPar);
            Action_20 = false;
        }
        if (Action_21) {
            eg3000_litho_stop(ftHandle);
            Action_21 = false;
        }
        if (Action_22) {
            eg3000_set_extra_calib(ftHandle, adcCalib, dacCalib);
            Action_22 = false;
        }
        if (Action_23) {
            if (!csb_speed_extended) csb_speed_idx = qMin(3, csb_speed_idx);
            BYTE data = csb_speed_idx;
            eg3000_send_packet(ftHandle, ADR_CPU, CPU_SET_CSB_RATE, &data, 1);
            Action_23 = false;
        }

        st = eg3000_is_ok(ftHandle);
        write_lock.unlock();
        msleep(1);
    }

    motorRst = 0;
}
//----------------------Read Thread USB-----------------------------------------------------
USB_READ_THREAD::~USB_READ_THREAD()
{
    for (int i = 0; i < 8; i++) {
        if (ScanDataPart[i])
            free(ScanDataPart[i]);
        if (ScanDataPart_lsb[i])
            free(ScanDataPart_lsb[i]);
    }
}

USB_READ_THREAD::USB_READ_THREAD()
{
    _rThread = this;

    for (int i = 0; i < 8; i++) {
        ScanDataPart[i] = 0;
        ScanDataPart_lsb[i] = 0;
    }

    scanError = false;
    headConnected = true;
    stageConnected = true;
    scannersConnectedUpdated = false;

    Nx = 200;
    Ny = 200;
    Scan_ChA = 0;
    Scan_ChB = 0;
    Scan_ChC = 0;
    Scan_ChD = 0;

    thread_started = false;
    thread_closed = false;
    is_Osc_test = false;
    is_Spectr_test = false;

    Scan_test = false;

    ccdTest = false;
    ccdReady = false;
    ccdCounter = 0;
    Lithotest = false;

    direction = 0;
    up_down_left_right = 1;
    scan_wait = 1;
}

IProcessPacket::~IProcessPacket()
{
    if (master)
        ((USB_READ_THREAD *)master)->removeReader(this);
}

void USB_READ_THREAD::addReader(IProcessPacket *iface)
{
    readers.append(iface);
    iface->master = this;
}
void USB_READ_THREAD::removeReader(IProcessPacket *iface)
{
    if (iface->master != this)
        return;

    iface->master = 0;
    readers.remove(readers.indexOf(iface));
}

void USB_READ_THREAD::SetTestMode(bool value)
{
    testMode=value;
}
void USB_READ_THREAD::sl_set_Litho_test()
{
    Lithotest=true;
}
void USB_READ_THREAD::start_stop(bool st)
{
    if (st) {
        if (!thread_started) {
            scanError = false;
            thread_started = true;
            headConnected = true; //here for test mode
            stageConnected = true; //here for test mode
            scannersConnectedUpdated = false;

            emit capBits(0, 0); //eg3k
            emit capBits(2, 0); //head
            emit capBits(3, 0); //stage
            emit capBits(4, 0); //z3
            emit capBits(6, 0); //motors

            start(QThread::HighestPriority);
        }
    } else {
        thread_started = false;
    }
}
void USB_READ_THREAD::thread_close()
{
    thread_closed = true;

    while (isRunning()) {
        qApp->processEvents();
        msleep(3);
    }
}

//---for test mode--------------------
void USB_READ_THREAD::sl_set_osc_test(bool value)
{
    if (testMode) is_Osc_test=value;
}
void USB_READ_THREAD::sl_set_lockin_test()
{
    if (testMode) is_Lock_in_test=true;
}
void USB_READ_THREAD::sl_set_Spectr_test()
{
    if (testMode) is_Spectr_test=true;
}

void USB_READ_THREAD::sl_set_scan_test(bool value)
{
    if (value) {
        for (int i = 0; i < 8; i++) {
            if (ScanDataPart[i])
                ScanDataPart[i]->Line = 65000;
            if (ScanDataPart_lsb[i])
                ScanDataPart_lsb[i]->Line = 65000;
        }
    }

    if (testMode) {
        if (value) {
            ccdswitch = ccdTest;
            line_test = 0;
            sc_ch = 0;
        }
        Scan_test = value;
    }
}
void USB_READ_THREAD::set_scan_test_ccd(bool value)
{
    if (value) {
        ccdswitch = true;
        ccdCounter = 0;
        ccdReady = true;
    }
    ccdTest = value;
}
void USB_READ_THREAD::sl_set_scan_test_ccd_ready()
{
    ccdReady = true;
}
double USB_READ_THREAD::test_plane_func(double xp, double yp)
{
    double a0=1000, a1=200, a2=200;

    return a0 + a1*xp + a2*yp;
}
//------------------------------

void USB_READ_THREAD::run()
{
    ccdswitch=false;

    bool Packet_Recieved = false;
    DWORD PacketSize = 0;
    BYTE Address = 0;
    BYTE SubAddress = 0;

    if (testMode) {
        deflec_ind = 0;
        switch_osc_ch = 0;
        is_Osc_test = false;
        is_Lock_in_test = false;
        Scan_test = false;

        eg3000_close_read_handle(&read_handle);
        read_handle.buffer = malloc(20480);
    }

    DWORD last_bytes_cnt = 0;

    bool st_read = true;

    if (motor_driver) { //motor command preprocessing
        while (thread_started && (!thread_closed) && st_read) {
            if (!read_lock.tryLock(10)) {
                msleep(1);
                continue;
            }

            static BYTE buffer[100];
            DWORD read;
            st_read = eg3000_read(ftHandle, buffer, 100, &read);

            if (read) {
                WORD SMStatus = buffer[read-1];
                int m0idx = 0;
                int mcount = 4;
                BYTE Status;

                for (int i = 0; i < mcount; i++) {
                    BYTE shift = (i % 8) * 2;
                    BYTE SMState = (SMStatus >> shift) & 0x03;

                    switch (SMState) {
                        case SMD_STOPPED:
                            Status = smStopped;
                            break;
                        case SMD_LIMIT_1:
                            Status = smLimitUp;
                            break;
                        case SMD_LIMIT_2:
                            Status = smLimitDown;
                            break;
                        case SMD_RUNNING:
                            Status = (StepMotorDirs[m0idx + i]) ? smRunningDown : smRunningUp;
                            break;
                    }

                    for (int k = 0; k < 4; k++) {
                        BYTE target = StepMotorRemap[k];
                        if (!target) target = k+1;
                        if (m0idx + i + 1 == target)
                            emit Motors_status(k+1, Status);
                    }
                }
            }

            read_lock.unlock();
            if (!read) msleep(1);
        }
        return;
    }

    while (thread_started && (!thread_closed) && st_read) {
        if (!read_lock.tryLock(10)) {
            msleep(1);
            continue;
        }

        bool maybe_have_more_data = false;

        //emulator
        if (testMode) //TODO unlikely()
            Packet_Recieved = makeTestData(&SubAddress);
        else {
            st_read = eg3000_update_parse(ftHandle, &read_handle);
            Packet_Recieved = read_handle.packet_recieved;
            PacketSize = read_handle.packet_size;
            Address = read_handle.address;
            SubAddress = read_handle.sub_address;
            maybe_have_more_data = (last_bytes_cnt != read_handle.buf_data_length) || read_handle.header_read;
            last_bytes_cnt = read_handle.buf_data_length;
        }

        //packet parsing
        if (Packet_Recieved) {
            for (int i = 0; i < readers.size(); i++)
                if (readers.at(i)->processPacket(Address, SubAddress, read_handle.buffer, PacketSize))
                    break; //TODO check if break by packet_listeners_count() or iterate hash_listeners_by_id[SubAddress]

            if (SubAddress == CPU_DEVICE_CAP_BITS &&
                    (Address - ADR_HEAD) >= 0 &&
                    (Address - ADR_HEAD) <= 4) {
                emit capBits(2 + (Address - ADR_HEAD), *((DWORD *)read_handle.buffer));
            } else
            if (Address == ADR_THERMO) {
                emit thermoData(QByteArray((char *)read_handle.buffer, 6));
            } else
                if (Address == ADR_CPU)
            switch(SubAddress) {
            case CPU_READ_CALDATA: {
                EG_CalData calib_data;
                memset(&calib_data, 0, sizeof (EG_CalData));
                memcpy(&calib_data, read_handle.buffer, qMin((DWORD)sizeof (EG_CalData), PacketSize));

                //first 48 floats in arrays
                memcpy((void *)Scanner_data_pid_cal_range, (void *)&calib_data, 48*sizeof(float));

                Scanner_data_scale[0] = calib_data.CSB_Mult[0];
                Scanner_data_scale[1] = calib_data.CSB_Mult[1];
                Scanner_data_scale[2] = calib_data.CSB_Mult[2];

                Scanner_data_Shift[0] = calib_data.CSB_Shift[0]-32768;
                Scanner_data_Shift[1] = calib_data.CSB_Shift[1]-32768;
                Scanner_data_Shift[2] = calib_data.CSB_Shift[2]-32768;

                *Nlc_status_cl_xy = (bool)calib_data.NLCStatus;
                *CL_status_xy  = (bool)calib_data.CLStatus;
                *Nlc_status_cl_z = (bool)calib_data.NLCStatusZ;
                *CL_status_z  = (bool)calib_data.CLStatusZ;
                *Nlc_status_ol_xy = (bool)calib_data.NLCStatusOL;
                *Nlc_status_ol_z = (bool)calib_data.NLCStatusOLZ;
                *free_move_speed = calib_data.DefaultRateOL;

                //always in pair with next packet, update on full info
                //emit Update_EEP_Data();
            }
                break;
            case CPU_READ_CALDATA_STAGE: {
                EG_CalData calib_data;
                memset(&calib_data, 0, sizeof (EG_CalData));
                memcpy(&calib_data, read_handle.buffer, qMin((DWORD)sizeof (EG_CalData), PacketSize));

                //first 48 floats in arrays
                memcpy((void *)Scanner2_data_pid_cal_range, (void *)&calib_data, 48*sizeof(float));

                Scanner2_data_scale[0] = calib_data.CSB_Mult[0];
                Scanner2_data_scale[1] = calib_data.CSB_Mult[1];
                Scanner2_data_scale[2] = calib_data.CSB_Mult[2];

                Scanner2_data_Shift[0] = calib_data.CSB_Shift[0]-32768;
                Scanner2_data_Shift[1] = calib_data.CSB_Shift[1]-32768;
                Scanner2_data_Shift[2] = calib_data.CSB_Shift[2]-32768;

                *Nlc_status_cl_xy2 = (bool)calib_data.NLCStatus;
                *CL_status_xy2  = (bool)calib_data.CLStatus;
                *Nlc_status_cl_z2 = (bool)calib_data.NLCStatusZ;
                *CL_status_z2  = (bool)calib_data.CLStatusZ;
                *Nlc_status_ol_xy2 = (bool)calib_data.NLCStatusOL;
                *Nlc_status_ol_z2 = (bool)calib_data.NLCStatusOLZ;
                *free_move_speed2 = calib_data.DefaultRateOL;

                emit Update_EEP_Data();
            }
                break;
                //--------------Osciloscope----------------------------
            case CPU_OSCDATA: {
                EG_OscData *Osc_data = (EG_OscData *)read_handle.buffer;

                int dataSize = sizeof(EG_OscData) - (OSC_LINE_SIZE - Osc_data->Size)*sizeof(WORD);
                emit OscCurveData(QByteArray((char *)Osc_data, dataSize));
            }
                break;
            case CPU_GET_DATA: {
                EG_GetData *Get_data = (EG_GetData *)read_handle.buffer;
                BYTE src = Get_data->Source;
                DWORD data = Get_data->Data;

                if (src == srcDACX)
                    emit getDataRecvX(data / 65535. * _wThread->Scanner_data_pid_cal_range[45]);
                if (src == srcDACY)
                    emit getDataRecvY(data / 65535. * _wThread->Scanner_data_pid_cal_range[46]);
                if (src == srcDACX2)
                    emit getDataRecvX2(data / 65535. * _wThread->Scanner2_data_pid_cal_range[45]);
                if (src == srcDACY2)
                    emit getDataRecvY2(data / 65535. * _wThread->Scanner2_data_pid_cal_range[46]);

                emit getDataRecv(src, data);
            }
                break;
            case CPU_CAP_BITS:
                emit capBits(0, *((DWORD *)read_handle.buffer));
                break;
            case CPU_STATUS: {
                EG_Status *Deflectomter = (EG_Status *)read_handle.buffer;

                if (!scanError && (Deflectomter->ErrorCode & ecScanErrorPID))
                    emit sigScanError();

                scanError = (Deflectomter->ErrorCode & ecScanErrorPID);
                headConnected = !(Deflectomter->ErrorCode & ecHeadNotFound);
                stageConnected = !(Deflectomter->ErrorCode & ecStageNotFound);
                scannersConnectedUpdated = true;

                if (!scanError && (Deflectomter->ErrorCode & ecAxisErrorPID))
                    emit sigSetpointError(0);
                if (!scanError && (Deflectomter->ErrorCode & ecAxisErrorPID2))
                    emit sigSetpointError(1);

                emit sigSPMStatus(Deflectomter->Zpos / 65535.,
                                  Deflectomter->DFL,
                                  Deflectomter->LFL,
                                  Deflectomter->MAG,
                                  Deflectomter->DInt / 32767. - 1);
                emit status(Deflectomter->Scanning != mdIdle ? EDplay : EDok);
            }
                break;
            case CPU_FREQSCAN_DATA: {
                EG_LockInScanData *FreqData = (EG_LockInScanData *)read_handle.buffer;

                int dataSize = sizeof(EG_LockInScanData) - (LOCKIN_LINE_SIZE - FreqData->Size)*sizeof(WORD);
                emit FreqScanData(QByteArray((char *)FreqData, dataSize));
            }
                break;
            case CPU_FORCE_CURVE_DATA: {
                EG_ForceCurveData *ForceCurve = (EG_ForceCurveData *)read_handle.buffer;

                int dataSize = sizeof(EG_ForceCurveData) - (FORCE_CURVE_LINE_SIZE - ForceCurve->Size)*sizeof(WORD);
                emit ForceCurveData(QByteArray((char *)ForceCurve, dataSize));
            }
                break;
            case CPU_SCANDATA:
            case CPU_SCANDATA_PART: {
                EG_ScanData *Scan_Data = (EG_ScanData *)read_handle.buffer;

                if (SubAddress == CPU_SCANDATA_PART) {
                    BYTE ch = Scan_Data->Channel;
                    if (ch >= 8) {
                        qDebug() << "scan part unexpected channel";
                        break;
                    }

                    EG_ScanData **bufarr = Scan_Data->Scanner & 0x80 ? ScanDataPart_lsb : ScanDataPart;
                    EG_ScanData *buf = bufarr[ch];

                    if (!buf) {
                        buf = bufarr[ch] = (EG_ScanData *)malloc(132000);
                        if (!buf) {
                            qDebug() << "scan part out of memory";
                            break;
                        }

                        buf->Line = 65000;
                    }

                    if (buf->Line != Scan_Data->Line ||
                            buf->Direction != Scan_Data->Direction) { //TODO we can clear size after send, always sync other fields
                        buf->Scanner = Scan_Data->Scanner;
                        buf->Channel = Scan_Data->Channel;
                        buf->Axis = Scan_Data->Axis;
                        buf->Direction = Scan_Data->Direction;
                        buf->Line = Scan_Data->Line;
                        buf->ScanSize = Scan_Data->ScanSize;
                        buf->Source = Scan_Data->Source;
                        buf->Size = 0;
                    }

                    int sz = buf->Axis == AxisX ? Nx : Ny;

                    for (int i = 0; i < Scan_Data->Size; i++) {
                        if (buf->Direction == dirFW)
                            buf->Data[buf->Size + i] = Scan_Data->Data[i];
                        else
                            buf->Data[sz - buf->Size - i - 1] = Scan_Data->Data[i];
                    }
                    buf->Size += Scan_Data->Size;

                    if (buf->Size != sz)
                        break;

                    Scan_Data = buf;
                }

                int dataSize = sizeof(EG_ScanData) - (MAX_LINE_SIZE - Scan_Data->Size)*sizeof(WORD);
                emit ScanLineReady(QByteArray((char *)Scan_Data, dataSize));
            }
                break;
            case CPU_CCD_START:
                emit CCD_Start();
                break;
            case CPU_STEP_MOTOR_STATUS: {
                EG_StepMotorStatus *motor_satus = (EG_StepMotorStatus *)read_handle.buffer;
                emit Motors_status(motor_satus->MotorN, motor_satus->Status);
            }
                break;
            case CPU_LANDING_STATUS: {
                WORD landing_status;
                memcpy(&landing_status, read_handle.buffer, sizeof(WORD));
                emit Landing_status((int)landing_status);
            }
                break;
            case CPU_LITHO_DONE:
                emit LithoReady();
                break;
            }

            if (testMode)
                Packet_Recieved = false;
            else
                eg3000_discard_packet(&read_handle);
        }
        else if (!maybe_have_more_data) msleep(1);

        read_lock.unlock();
    }

    if (testMode) {
        free(read_handle.buffer);
        eg3000_init_read_handle(&read_handle);
    }
}

//----------------------

bool USB_READ_THREAD::makeTestData(BYTE *subaddr)
{
    bool Packet_Recieved = false;
    BYTE SubAddress = 0;

    if (ccdswitch || !Scan_test)
        msleep(20);

    if (!Scan_test && !ccdTest && !is_Osc_test && !is_Lock_in_test && !is_Spectr_test) {
        Packet_Recieved = true;
        SubAddress = CPU_STATUS;
        EG_Status *Deflectomter = (EG_Status *)read_handle.buffer;
        Deflectomter->Scanning = mdIdle;
        Deflectomter->ErrorCode = ecOK;
        Deflectomter->DFL  = (WORD)(32000+20000*sin(0.01*deflec_ind));
        Deflectomter->LFL  = (WORD)(32000+20000*cos(0.01*deflec_ind));
        Deflectomter->DInt = (WORD)(32768+30000*cos(0.01*deflec_ind)*cos(0.01*deflec_ind));
        Deflectomter->MAG  = (WORD)(30000*cos(0.01*deflec_ind)*cos(0.01*deflec_ind))+32768;
        Deflectomter->Zpos = (WORD)(65000*cos(0.01*deflec_ind)*cos(0.01*deflec_ind));
        if (deflec_ind < 100000) deflec_ind += 3;
        else deflec_ind = 0;
        msleep(100);
    }

    if (is_Lock_in_test) {
        msleep(500);
        Packet_Recieved = true;
        SubAddress = CPU_FREQSCAN_DATA;
        int resfr = 250 + rand() % 500;
        EG_LockInScanData *FreqData = (EG_LockInScanData *)read_handle.buffer;
        FreqData->Size = 1000;
        FreqData->Source = srcMag1;
        FreqData->res_point = (325 + resfr / 100) * 1000;
        for (int i = 0; i < 1000; i++)
            FreqData->Data[i] = (WORD)(30000*exp(-0.0001*(i-resfr)*(i-resfr))+32768);
        is_Lock_in_test = false;
    }
    else {
        if (!ccdswitch && Scan_test) {
            msleep(scan_wait);
            if (line_test < (direction ? Nx : Ny)) {
                WORD line_shift = (WORD)(rand() % 5000);

                EG_ScanData *Scan_Data = (EG_ScanData *)read_handle.buffer;
                for (int i = 0; i < Nx; i++)
                {
                    double cur_x_point = St_point_Xn;
                    double cur_y_point = St_point_Yn;
                    if (direction) {
                        cur_y_point += sign * i * step_y;
                        cur_x_point += (up_down_left_right ? 1 : -1) * line_test * step_x;
                    } else {
                        cur_x_point += sign * i * step_x;
                        cur_y_point += (up_down_left_right ? 1 : -1) * line_test * step_y;
                    }

                    WORD func = 0;
                    switch (sc_ch) {
                    case 0 : func = (WORD)(10000.*sin(1*cur_x_point)*sin(1*cur_x_point)*sin(1*cur_y_point)*sin(1*cur_y_point)); break;
                    case 1 : func = (WORD)(10000.*sin(0.5*cur_x_point)*sin(0.5*cur_x_point)*sin(0.5*cur_y_point)*sin(0.5*cur_y_point)); break;
                    case 2 : func = (WORD)(10000.*sin(0.25*cur_x_point)*sin(0.25*cur_x_point)*sin(0.25*cur_y_point)*sin(0.25*cur_y_point)); break;
                    case 3 : func = (WORD)(10000.*sin(0.125*cur_x_point)*sin(0.125*cur_x_point)*sin(0.125*cur_y_point)*sin(0.125*cur_y_point)); break;
                    }

                    func += test_plane_func(cur_x_point, cur_y_point) + line_shift;
                    Scan_Data->Data[i] = func;
                }

                SubAddress = CPU_SCANDATA;
                Scan_Data->Axis = direction;
                Scan_Data->Direction = dirFW;
                Scan_Data->Channel = sc_ch;
                Scan_Data->Scanner = scHead; //or scStage
                Scan_Data->Line = line_test;
                Scan_Data->Size = direction ? Ny : Nx;
                switch (Scan_Data->Channel) {
                case 0: Scan_Data->Source = Scan_ChA; break;
                case 1: Scan_Data->Source = Scan_ChB; break;
                case 2: Scan_Data->Source = Scan_ChC; break;
                case 3: Scan_Data->Source = Scan_ChD; break;
                }
                Packet_Recieved = Scan_Data->Source != srcNone;
            } else
                Scan_test = false;

            sc_ch++;
            if (sc_ch > 3) {
                sc_ch = 0;
                line_test++;
                if (ccdTest)
                    ccdswitch = true;
            }
        }

        if (ccdTest && ccdReady && ccdswitch) {
            ccdReady = false;
            if (ccdCounter < ccdNX*ccdNY) {
                Packet_Recieved = true;
                SubAddress = CPU_CCD_START;
                if (ccdCounter > 0 && ccdCounter % (direction ? ccdNY : ccdNX) == 0)
                    ccdswitch = false;
                ccdCounter++;
            } else {
                ccdswitch = false;
                ccdTest = false;
            }
        }

        if (is_Osc_test && !Scan_test) {
            EG_OscData *Osc_data = (EG_OscData *)read_handle.buffer;
            Packet_Recieved = true;
            SubAddress = CPU_OSCDATA;
            Osc_data->Channel = switch_osc_ch;
            Osc_data->Size = 256;
            Osc_data->Source = _wThread->Active_Ch[switch_osc_ch]; //srcDFL
            for (int i = 0; i < 256; i++) {
                if (switch_osc_ch == 0) {
                    Osc_data->Data[i] = 30000;
                } else if (switch_osc_ch == 1) {
                    Osc_data->Data[i] = 40000;
                } else if (switch_osc_ch == 2) {
                    Osc_data->Data[i] = 50000;
                } else if (switch_osc_ch == 3) {
                    Osc_data->Data[i] = 60000;
                }
                Osc_data->Data[i] += rand() % (5 + i);
            }

            switch_osc_ch++;
            if (switch_osc_ch > 3) switch_osc_ch = 0;
        }
        else
        {
            if (is_Spectr_test) {
                msleep(500);
                Packet_Recieved = true;
                SubAddress = CPU_FORCE_CURVE_DATA;
                EG_ForceCurveData *ForceCurve = (EG_ForceCurveData *)read_handle.buffer;
                ForceCurve->Size = 1000;
                ForceCurve->Source = srcMag1;
                for (int i = 0; i < 1000; i++) {
                    ForceCurve->Data[i] = 65535 * exp(-0.05 * i);
                    ForceCurve->DataBack[i] = 65535 * (1 - exp(-0.05 * (1000 - i - 1)));
                }
                is_Spectr_test = false;
            }
            else if (Lithotest) {
                Packet_Recieved = true;
                SubAddress = CPU_LITHO_DONE;
                Lithotest = false;
            }
        }
    }

    *subaddr = SubAddress;
    return Packet_Recieved;
}

//--------------LOCK-IN-----------------------
void USB_READ_THREAD::Set_Scanner_Data_out(float* scanner_data_pid_cal_range, DWORD *scanner_data_scale, long* scanner_data_Time,
                                           bool* nlc_c_xy, bool* ncl_xy, bool* nlc_c_z, bool* ncl_z,
                                           bool* nlc_o_xy, bool* nlc_o_z, float *_free_move_speed,
                                           float* scanner2_data_pid_cal_range, DWORD *scanner2_data_scale, long* scanner2_data_Time,
                                           bool* nlc2_c_xy, bool* ncl2_xy, bool* nlc2_c_z, bool* ncl2_z,
                                           bool* nlc2_o_xy, bool* nlc2_o_z, float *_free_move_speed2)
{
    Scanner_data_pid_cal_range = scanner_data_pid_cal_range;
    Scanner_data_scale = scanner_data_scale;
    Scanner_data_Shift = scanner_data_Time;

    Nlc_status_cl_xy = nlc_c_xy;
    CL_status_xy = ncl_xy;
    Nlc_status_cl_z = nlc_c_z;
    CL_status_z = ncl_z;
    Nlc_status_ol_xy = nlc_o_xy;
    Nlc_status_ol_z = nlc_o_z;

    free_move_speed = _free_move_speed;

    Scanner2_data_pid_cal_range = scanner2_data_pid_cal_range;
    Scanner2_data_scale = scanner2_data_scale;
    Scanner2_data_Shift = scanner2_data_Time;

    Nlc_status_cl_xy2 = nlc2_c_xy;
    CL_status_xy2 = ncl2_xy;
    Nlc_status_cl_z2 = nlc2_c_z;
    CL_status_z2 = ncl2_z;
    Nlc_status_ol_xy2 = nlc2_o_xy;
    Nlc_status_ol_z2 = nlc2_o_z;

    free_move_speed2 = _free_move_speed2;
}
//-------SCANNING-----------------------------------------
void USB_READ_THREAD::SetScanPoints(unsigned short int nx, unsigned short int ny, double Vscn, double step_xn, double step_yn, double st_point_xn, double st_point_yn, unsigned char scan_dir_st_point_n, unsigned char scan_an, unsigned char scan_bn, unsigned char scan_cn, unsigned char scan_dn,
                                    bool , bool , bool , bool , bool , double )
{
    Scan_ChA = scan_an;
    Scan_ChB = scan_bn;
    Scan_ChC = scan_cn;
    Scan_ChD = scan_dn;

    scan_dir_st_point = scan_dir_st_point_n;
    scan_wait = qMax<int>(1, 1000/4/Vscn);

    St_point_Xn = st_point_xn;
    St_point_Yn = st_point_yn;
    step_x = step_xn;
    step_y = step_yn;

    switch(scan_dir_st_point)
    {
    case spTopLeftAxisX:
        direction = AxisX;
        up_down_left_right = dirFW;
        sign=1;
        break;
    case spTopRightAxisX:
        direction = AxisX;
        up_down_left_right = dirFW;
        sign=-1;
        break;
    case spBottomRightAxisX:
        direction = AxisX;
        up_down_left_right = dirBW;
        sign=-1;
        break;
    case spBottomLeftAxisX:
        direction = AxisX;
        up_down_left_right = dirBW;
        sign=1;
        break;
    case spTopLeftAxisY:
        direction = AxisY;
        up_down_left_right = dirBW;
        sign=-1;
        break;
    case spTopRightAxisY:
        direction = AxisY;
        up_down_left_right = dirFW;
        sign=-1;
        break;
    case spBottomRightAxisY:
        direction = AxisY;
        up_down_left_right = dirFW;
        sign=1;
        break;
    case spBottomLeftAxisY:
        direction = AxisY;
        up_down_left_right = dirBW;
        sign=1;
        break;
    }

    Nx = nx;
    Ny = ny;
}
void USB_READ_THREAD::SetccdScanMode(int , int , int nx, int ny, int , double , double , double , double , bool)
{
    ccdNX = nx;
    ccdNY = ny;
}
