#include "videoplayer.h"
#include "mainwindow.h"

#include "EG3000.h"
#include "oscilloscope.h"
#include "gui/viewer3d.hpp"
#include "mobile/remote_mobile.hpp"
#include <QCheckBox>
#include <QBuffer>
#include <QApplication>
#include <QMessageBox>
#include <QProgressBar>


void VideoPlayer::on_Camera_ImgReceived(QImage Img)
{
    if (refresh_limit_timer.isValid() && refresh_limit_timer.elapsed() < 20 + refresh_limit_last_elapsed) //~50 refreshes per second max
        return;
    refresh_limit_timer.start();

    Img.detach();
    QRectF cropp(Img.width()*(1 - cw) / 2,
                 Img.height()*(1 - ch) / 2,
                 Img.width()*cw,
                 Img.height()*ch);

    if (!Img.rect().isValid() || !cropp.isValid()) {
        qDebug() << "invalid frame";
        return;
    }

    if (!rotxy->isChecked())
        ritem->image = Img.copy(cropp.toRect()).mirrored(flipx->isChecked(), flipy->isChecked());
    else {
        QTransform m;
        m.rotate(90);
        m.scale(flipx->isChecked() ? -1 : 1, flipy->isChecked() ? -1 : 1);
        ritem->image = Img.copy(cropp.toRect()).transformed(m);
    }

    QPlot->setMinimumSize(ritem->image.size() / 32);
    QPlot->replot();

    if (DoAvi)
        VideoSave.AddImage(ritem->image);

    frames++;
    VideoSave.SetFps(qRound(1000.0 / fps_time.elapsed() * frames));

    if (inst_idx == 0)
    if (I_MODULE *net = ModuleManager::instance->find1("net")) {
        MobileServer *m = (MobileServer *)net->basecast();
        if (m->any_need_data("video_need")) {
            QByteArray vdata;
            QBuffer buf(&vdata);
            buf.open(QBuffer::WriteOnly);
            ritem->image.save(&buf, "bmp");
            buf.close();

            m->setPropIdVar(DataID::VideoData, vdata, 1);
        }
    }

    refresh_limit_last_elapsed = refresh_limit_timer.elapsed() / 0.2; //time budget, %
}



void VideoPlayer::calcOscContrast(QByteArray data)
{
    EG_OscData *sdata = (EG_OscData *)data.constData();

    if (sdata->Channel >= 4)
        return;

    osc_contrast = 0;
    for (int i = 0; i < sdata->Size; i++)
        osc_contrast += sdata->Data[i];
    osc_contrast /= sdata->Size;

    new_osc_contrast = true;
}

void VideoPlayer::on_BAutoFocusOsc_clicked()
{
    double speed = 1.0; //mkm/s
    double z1 = 0;
    double z2 = get_max_objective_z();//mkm

    double contrast = 0;
    double old = 0;
    double max = 0;
    double zmax = 0;
    int counter = 0;

    QList<I_MODULE *> l = ModuleManager::instance->find("oscilloscope");
    I_MODULE *oscm = l.isEmpty() ? 0 : l.last();
    I_MODULE *eg3km = ModuleManager::instance->find1("eg3000");
    if (!oscm || !eg3km)
        return;

    Oscilloscope *osc = (Oscilloscope *)oscm->basecast();
    USB_WRITE_THREAD *_wThread = &((USB_THREAD_GROUP *)eg3km->basecast())->wThread;

    setEnabled(false);
    osc->setEnabled(false);

    int old_ch[4];
    for (int i = 0; i < 4; i++) {
        old_ch[i] = osc->CBChannel[i]->currentIndex();
        osc->CBChannel[i]->setCurrentIndex(srcNone);
    }

    osc->CBChannel[0]->setCurrentIndex(osc->CBChannel[0]->findData(srcADC4));

    osc->start->change_state_man_with_sig(true);

    _wThread->set_Ex_dac1(0);
    osc->DacsWidget->Extra_Dac_1->setValue(_wThread->Ex_dac1);

    isInterrupted = false;

    QPointer<StatusBar> bar = new StatusBar(tr("PMT Autofocus"), 4);
    connect(bar, SIGNAL(canceled()), SLOT(cancelOperation()));
    bar->add();

    go_objective_z(0);
    SleepExport::msleep_gui(50);

    int time = 50;
    QElapsedTimer timer;

    const int daclow = 35000;
    const int dachigh = 55000;
    double dacspeed = 0.1;
    int dacsign = 1;

    counter = 0;
    do {
        _wThread->set_Ex_dac1(_wThread->Ex_dac1 + dacsign*dacspeed);
        osc->DacsWidget->Extra_Dac_1->setValue(_wThread->Ex_dac1);
        SleepExport::msleep_gui(50);

        new_osc_contrast = false;
        connect(osc, SIGNAL(forwardedOscData(QByteArray)), this, SLOT(calcOscContrast(QByteArray)));

        time = 500;
        timer.start();
        while (!new_osc_contrast && !isInterrupted && !timer.hasExpired(time)) {
            qApp->processEvents();
            SleepExport::msleep2(10);
        }
        if (timer.hasExpired(time)) {
            qDebug() << "AFO: no osc signal";
            new Notification(tr("PMT Autofocus") + ": " + tr("No data from controller"), Notification::Warn, 10);
            doBeep();

            destructStatus(bar);
            isInterrupted = true;
        }

        disconnect(osc, SIGNAL(forwardedOscData(QByteArray)), this, SLOT(calcOscContrast(QByteArray)));
        if (isInterrupted)
            continue;

        old = contrast;
        contrast = osc_contrast;

        if ((old > daclow && contrast < daclow) ||
                (old < dachigh && contrast > dachigh)) {
            dacsign *= -1;
            dacspeed /= 2;
            qDebug() << "dac overflow" << old << contrast << dacspeed;
        }
        counter++;
    } while ((contrast < daclow || contrast > dachigh) && counter < 50 && _wThread->Ex_dac1 < 0.9 && !isInterrupted);

    if (counter >= 50 || _wThread->Ex_dac1 >= 0.9 || contrast != contrast || isInterrupted) {
        destructStatus(bar);

        setEnabled(true);
        osc->setEnabled(true);

        for (int i = 0; i < 4; i++)
            osc->CBChannel[i]->setCurrentIndex(old_ch[i]);

        osc->start->change_state_man_with_sig(false);

        _wThread->set_Ex_dac1(0);
        osc->DacsWidget->Extra_Dac_1->setValue(_wThread->Ex_dac1);

        if (!isInterrupted) {
            qDebug() << "AFO: no signal";
            new Notification(tr("PMT Autofocus") + ": " + tr("No PMT signal"), Notification::Warn, 10);
            doBeep();
        }

        return;
    }

    qDebug() << "AFO dac" << contrast << _wThread->Ex_dac1;

    double z; //for loop clause

    counter = 0;
    do {
        bar->setValue(1+counter);
        qDebug() << "AFO pass" << counter << z1 << z2;

        go_objective_z(z1);
        SleepExport::msleep_gui(50);

        do {
            if (contrast > max) {
                max = contrast;
                zmax = get_objective_z();
            }

            new_osc_contrast = false;
            connect(osc, SIGNAL(forwardedOscData(QByteArray)), this, SLOT(calcOscContrast(QByteArray)));

            time = 500;
            timer.start();
            while (!new_osc_contrast && !isInterrupted && !timer.hasExpired(time)) {
                qApp->processEvents();
                SleepExport::msleep2(10);
            }
            if (timer.hasExpired(time)) {
                qDebug() << "AFO: no osc signal2";
                new Notification(tr("PMT Autofocus") + ": " + tr("No data from controller"), Notification::Warn, 10);
                doBeep();

                destructStatus(bar);
                isInterrupted = true;
            }

            disconnect(osc, SIGNAL(forwardedOscData(QByteArray)), this, SLOT(calcOscContrast(QByteArray)));
            if (isInterrupted)
                continue;

            old = contrast;
            contrast = osc_contrast;

            if (contrast > dachigh) {
                dacsign = -1;
                dacspeed = 0.05;
                double olddac = _wThread->Ex_dac1;

                do {
                    _wThread->set_Ex_dac1(_wThread->Ex_dac1 + dacsign*dacspeed);
                    osc->DacsWidget->Extra_Dac_1->setValue(_wThread->Ex_dac1);
                    SleepExport::msleep_gui(50);

                    new_osc_contrast = false;
                    connect(osc, SIGNAL(forwardedOscData(QByteArray)), this, SLOT(calcOscContrast(QByteArray)));

                    time = 500;
                    timer.start();
                    while (!new_osc_contrast && !isInterrupted && !timer.hasExpired(time)) {
                        qApp->processEvents();
                        SleepExport::msleep2(10);
                    }
                    if (timer.hasExpired(time)) {
                        qDebug() << "AFO: no osc signal3";
                        new Notification(tr("PMT Autofocus") + ": " + tr("No data from controller"), Notification::Warn, 10);
                        doBeep();

                        destructStatus(bar);
                        isInterrupted = true;
                    }

                    disconnect(osc, SIGNAL(forwardedOscData(QByteArray)), this, SLOT(calcOscContrast(QByteArray)));
                    if (isInterrupted)
                        continue;

                    old = contrast;
                    contrast = osc_contrast;

                    if ((old > daclow && contrast < daclow) ||
                            (old < dachigh && contrast > dachigh)) {
                        dacsign *= -1;
                        dacspeed /= 2;
                        qDebug() << "dac overflow" << old << contrast << dacspeed;
                    }
                } while ((contrast < daclow || contrast > dachigh) && !isInterrupted);
                max *= _wThread->Ex_dac1/olddac;

                qDebug() << "AFO dac" << contrast << _wThread->Ex_dac1 << max;
            }

            double zc = get_objective_z();
            qDebug() << "AFO_0" << contrast << zc;

            go_relative_objective_z(speed);
            SleepExport::msleep_gui(50);

            z = get_objective_z();
        } while ((z + speed) <= z2 && !isInterrupted);

        double dz = (z2 - z1)/8;
        double zrange = get_max_objective_z();

        z1 = qMin(qMax(0.0, zmax - dz/2), (double)zrange);
        z2 = qMin(qMax(0.0, zmax + dz/2), (double)zrange);

        speed /= 4;
        counter++;
    } while (speed >= 0.2 && counter < 3 && !isInterrupted);

    qDebug() << "AFO" << counter << zmax;

    go_objective_z(zmax);

    destructStatus(bar);

    setEnabled(true);
    osc->setEnabled(true);

    for (int i = 0; i < 4; i++)
        osc->CBChannel[i]->setCurrentIndex(old_ch[i]);

    osc->start->change_state_man_with_sig(false);

    _wThread->set_Ex_dac1(0);
    osc->DacsWidget->Extra_Dac_1->setValue(_wThread->Ex_dac1);
}

#define DIFF(a, b) (abs(long(a & 0xFF) - long(b & 0xFF)) + abs(long((a >> 8) & 0xFF) - long((b >> 8) & 0xFF)) + abs(long((a >> 16) & 0xFF) - long((b >> 16) & 0xFF)))
void VideoPlayer::on_BAutoFocus_clicked()
{
    double speed = 2.0;//mkm/s
    double z1 = 0;
    double z2 = get_max_objective_z();//mkm

    double max = 0;
    double zmax = 0;
    int counter = 0;

    setEnabled(false);
    bool old_start = m_BStart->on_off;
    if (!m_BStart->on_off)
        m_BStart->change_state_man_with_sig(true);

    isInterrupted = false;

    QPointer<StatusBar> bar = new StatusBar(tr("Optical Autofocus"), 3);
    connect(bar, SIGNAL(canceled()), SLOT(cancelOperation()));
    bar->add();

    int time;
    QElapsedTimer timer;

    do {
        bar->setValue(counter);
        qDebug() << "AF pass" << counter << z1 << z2;

        go_objective_z(z1);
        SleepExport::msleep_gui(200);

        double z = get_objective_z(); //for loop clause
        double contrast = 0;

        do {
            if (contrast > max) {
                max = contrast;
                zmax = z;
            }

            int frm = frames;
            time = 2000;
            timer.start();
            while (frm == frames && !isInterrupted && !timer.hasExpired(time)) {
                qApp->processEvents();
                SleepExport::msleep2(10);
            }
            if (timer.hasExpired(time)) {
                new Notification(tr("Autofocus") + ": " + tr("No frames from camera"), Notification::Warn, 10);
                doBeep();

                destructStatus(bar);
                isInterrupted = true;
            }

            if (isInterrupted)
                continue;

            z = get_objective_z();

            //async  move & calc
            go_relative_objective_z(speed);

            contrast = 0;
            QImage img = ritem->image;
            img.detach();
            for (int i = 1; i < img.width(); i++)
                for (int j = 1; j < img.height(); j++) {
                    int diff1 = DIFF(img.pixel(i, j), img.pixel(i-1, j));
                    if (abs(diff1) > 32)
                        contrast += abs(diff1)/128.0;

                    int diff2 = DIFF(img.pixel(i, j), img.pixel(i, j-1));
                    if (abs(diff2) > 32)
                        contrast += abs(diff2)/128.0;
                }

            qDebug() << "AF_0" << contrast << z;
        } while ((z + speed) <= z2 && !isInterrupted);

        double dz = (z2 - z1)/16;
        double zrange = get_max_objective_z();

        z1 = qMin(qMax(0.0, zmax - dz/2), (double)zrange);
        z2 = qMin(qMax(0.0, zmax + dz/2), (double)zrange);

        speed /= 8;
        counter++;
    } while (speed >= 0.05 && counter < 3 && !isInterrupted);

    qDebug() << "AF" << counter << zmax;

    go_objective_z(zmax);

    destructStatus(bar);

    if (!old_start)
        m_BStart->change_state_man_with_sig(false);

    setEnabled(true);
}

long POINT_DIFF(const QImage &img, int i, int j) {
    long _sum_ = 0;
    const int _R_ = 2;
    for (int _i = -_R_; _i <= _R_; _i++) {
        for (int _j = -_R_; _j <= _R_; _j++) {
            _sum_ += DIFF(img.pixel(i, j), img.pixel(i+_i, j+_j));
        }
    }
    return _sum_;
}
void VideoPlayer::on_Z_Reconstruct()
{
    static QString name;
    name = QFileDialog::getOpenFileName(this, tr("Open"), name,
                                        QString("index files(*.txt)"),
                                        0, QFileDialog::DontUseNativeDialog);

    if (name.isEmpty())
        return;

    //check for cache
    while (1) { //to break; instead of goto;
        QString nm = name;
        nm.chop(4);

        if (!QFile::exists(nm + "-heigth_map.png") || !QFile::exists(nm + "-reconstructed.png"))
            break;

        if (QMessageBox::question(this, tr("3D Z Reconstruct"), tr("Use previous result?"),
                                  QMessageBox::Yes, QMessageBox::No) != QMessageBox::Yes)
            break;

        QImage heigthMap(nm + "-heigth_map.png");
        QImage texture(nm + "-reconstructed.png");

        if (heigthMap.width() == 0 || heigthMap.height() == 0) {
            new Notification(tr("3D Z Reconstruct") + ": " + tr("File open failed"), Notification::Bad, 10);
            doBeep();
            return;
        }

        //show
#ifndef NO_3D
        int nx = heigthMap.width();
        int ny = heigthMap.height();

        float *val = new float[nx*ny];
        for (int i = 0; i < nx; i++)
            for (int j = 0; j < ny; j++)
                val[i*ny + j] = QColor(heigthMap.pixel(i, j)).blueF() - 0.5;

        T3dViewer *vw = new T3dViewer;
        vw->x_rot = 30;
        vw->z_scale = 0.2;
        vw->updateData(val, nx, ny, double(nx) / ny);
        delete [] val;
        vw->updateTexture(texture);
        vw->updateGL();
        vw->show();
#endif

        return;
    }

    isInterrupted = false;

    QPointer<StatusBar> bar = new StatusBar(tr("3D Z Reconstruct"), 1);
    connect(bar, SIGNAL(canceled()), SLOT(cancelOperation()));
    bar->add();

    //open index
    QFile fl(name);
    fl.open(QIODevice::ReadOnly | QIODevice::Text);

    QTextStream str(&fl);
    if (str.readLine() != "NST_index_file") {
        qDebug() << "3d_reconstruct: not nst index file";
        fl.close();
        return;
    }

    double start, step, count;
    str >> count >> step >> start;
    str.readLine(); // skip endl
    double end = start + step*(count-1);

    QVector<QString> lst;
    while (!str.atEnd()) {
        lst.push_back(str.readLine());
    }

    fl.close();
    if (lst.isEmpty())
        return;

    //open imgs
    QString path = name;
    if (path.lastIndexOf("/") != -1)
        path.truncate(path.lastIndexOf("/")+1);
    if (path.lastIndexOf("\\") != -1)
        path.truncate(path.lastIndexOf("\\")+1);
    QVector<double> zpos;
    QVector<QImage> imgs;
    for (int i = 0; i < lst.size(); i++) {
        QImage img;
        img.load(path + lst.at(i));

        if (!img.isNull()) {
            zpos.push_back(start + step*i);
            imgs.push_back(img);
        }
    }

    lst.clear();
    if (imgs.isEmpty()) {
        qDebug() << "3d_reconstruct: no images opened";
        return;
    }

    //calc 3d img max per point
    QSize min = imgs.first().size();
    for (int i = 1; i < imgs.size(); i++) {
        min = min.boundedTo(imgs.at(i).size());
    }

    if (!isInterrupted)
        bar->bar->setMaximum(imgs.size());

    QImage texture(min, QImage::Format_ARGB32);
    texture.fill(0xFFFFFF);

    float *max = new float[min.width()*min.height()];
    memset(max, 0, sizeof(float)*min.width()*min.height());


    //processing vvv //TODO iterative algo with PSF

    float *cmax = new float[min.width()*min.height()];
    memset(cmax, 0, sizeof(float)*min.width()*min.height());

    float contrast;
    for (int j = 2; j < min.width()-2 && !isInterrupted; j++) {
        for (int k = 2; k < min.height()-2; k++) {
            contrast = POINT_DIFF(imgs.last(), j, k);
            cmax[k*min.width() + j] = contrast;
            max[k*min.width() + j] = zpos.last();
            texture.setPixel(j, k, imgs.last().pixel(j, k));
        }
    }

    qApp->processEvents();

    imgs.pop_back();
    zpos.pop_back();

    if (!isInterrupted)
        bar->setValue(bar->bar->maximum() - imgs.size());

    while (!imgs.isEmpty() && !isInterrupted) {
        for (int j = 2; j < min.width()-2 && !isInterrupted; j++) {
            for (int k = 2; k < min.height()-2; k++) {
                contrast = POINT_DIFF(imgs.last(), j, k);

                if (contrast > cmax[k*min.width() + j]) {
                    cmax[k*min.width() + j] = contrast;
                    max[k*min.width() + j] = zpos.last();
                    texture.setPixel(j, k, imgs.last().pixel(j, k));
                }
            }
        }

        imgs.pop_back();
        zpos.pop_back();

        if (!isInterrupted)
            bar->setValue(bar->bar->maximum() - imgs.size());

        qApp->processEvents();
    }
    delete [] cmax;

    //processing ^^^


    //spikes filter x10
    if (!isInterrupted)
        bar->bar->setMaximum(10);

    float *max3 = new float[min.width()*min.height()];
    bool updated = true;
    for (int i = 0; i < 10 && !isInterrupted && updated; i++) {
        memcpy(max3, max, sizeof(float)*min.width()*min.height());
        updated = false;
        for (int j = 3; j < min.width()-3; j++) {
            for (int k = 3; k < min.height()-3; k++) {
                double avg = 0;
                avg += max3[(k-1)*min.width() + (j-1)];
                avg += max3[(k-1)*min.width() + (j)];
                avg += max3[(k-1)*min.width() + (j+1)];

                avg += max3[(k)*min.width() + (j-1)];
                avg += max3[(k)*min.width() + (j+1)];

                avg += max3[(k+1)*min.width() + (j-1)];
                avg += max3[(k+1)*min.width() + (j)];
                avg += max3[(k+1)*min.width() + (j+1)];
                avg /= 8;

                if (max[k*min.width() + j] - avg > fabs(step)) {
                    max[k*min.width() + j] = avg + fabs(step);
                    updated = true;
                } else if (max[k*min.width() + j] - avg < -fabs(step)) {
                    max[k*min.width() + j] = avg - fabs(step);
                    updated = true;
                }
            }
        }

        if (!isInterrupted)
            bar->setValue(i);

        qApp->processEvents();

        for (int i = 0; i < min.width(); i++) max[2*min.width() + i] = max[3*min.width() + i];
        for (int i = 0; i < min.width(); i++) max[(min.height()-3)*min.width() + i] = max[(min.height()-4)*min.width() + i];
        for (int i = 0; i < min.height(); i++) max[i*min.width() + 2] = max[i*min.width() + 3];
        for (int i = 0; i < min.height(); i++) max[i*min.width() + min.width()-3] = max[i*min.width() + min.width()-4];
    }
    delete [] max3;

    //average filter x10
    if (!isInterrupted)
        bar->bar->setMaximum(10);

    float *max2 = new float[min.width()*min.height()];
    for (int i = 0; i < 10 && !isInterrupted; i++) {
        memcpy(max2, max, sizeof(float)*min.width()*min.height());
        for (int j = 3; j < min.width()-3; j++) {
            for (int k = 3; k < min.height()-3; k++) {
                max[k*min.width() + j] += max2[(k-1)*min.width() + (j-1)];
                max[k*min.width() + j] += max2[(k-1)*min.width() + (j)];
                max[k*min.width() + j] += max2[(k-1)*min.width() + (j+1)];

                max[k*min.width() + j] += max2[(k)*min.width() + (j-1)];
                max[k*min.width() + j] += max2[(k)*min.width() + (j+1)];

                max[k*min.width() + j] += max2[(k+1)*min.width() + (j-1)];
                max[k*min.width() + j] += max2[(k+1)*min.width() + (j)];
                max[k*min.width() + j] += max2[(k+1)*min.width() + (j+1)];
                max[k*min.width() + j] /= 9;
            }
        }

        if (!isInterrupted)
            bar->setValue(i);

        qApp->processEvents();

        for (int i = 0; i < min.width(); i++) max[2*min.width() + i] = max[3*min.width() + i];
        for (int i = 0; i < min.width(); i++) max[(min.height()-3)*min.width() + i] = max[(min.height()-4)*min.width() + i];
        for (int i = 0; i < min.height(); i++) max[i*min.width() + 2] = max[i*min.width() + 3];
        for (int i = 0; i < min.height(); i++) max[i*min.width() + min.width()-3] = max[i*min.width() + min.width()-4];
    }
    delete [] max2;

    if (isInterrupted) {
        delete [] max;
        return;
    }

    //show
#ifndef NO_3D
    int nx = min.width()-4;
    int ny = min.height()-4;
    double sign = 1;
    if (start < end)
        sign = -1;

    QImage heigthMap(nx, ny, QImage::Format_ARGB32);
    heigthMap.fill(0xFFFFFF);

    float *val = new float[nx*ny];
    for (int i = 0; i < nx; i++)
        for (int j = 0; j < ny; j++) {
            val[i*ny + j] = sign*(0.5 - (start - max[(j+2)*min.width() + (i+2)]) / (start - end));
            QColor clr;
            double h = val[i*ny + j] + 0.5;
            clr.setRgbF(h, h, h);
            heigthMap.setPixel(i, j, clr.rgb());
        }

    T3dViewer *vw = new T3dViewer;
    vw->x_rot = 30;
    vw->z_scale = 0.2;
    vw->updateData(val, nx, ny, double(nx) / ny);
    delete [] val;
    vw->updateTexture(texture = texture.copy(2, 2, nx, ny));
    vw->updateGL();
    vw->show();

    name.chop(4);
    heigthMap.save(name + "-heigth_map.png");
    texture.save(name + "-reconstructed.png");
#endif

    destructStatus(bar);

    delete [] max;
}

void VideoPlayer::cancelOperation()
{
    isInterrupted = true;
}

void VideoPlayer::on_Z_Scan()
{
    m_ZScanWindow->layout()->activate();
    m_ZScanWindow->move(mapToGlobal(rect().center() - m_ZScanWindow->rect().center()));
    if (m_ZScanWindow->exec() != QDialog::Accepted)
        return;

    setEnabled(false);
    bool old_start = m_BStart->on_off;
    if (!m_BStart->on_off)
        m_BStart->change_state_man_with_sig(true);

    int steps = m_ZScanWindow->StepsCount->value();
    double start = m_ZScanWindow->EStartPos->value();
    double step = m_ZScanWindow->EStep->value();
    QString path = m_ZScanWindow->path;

    isInterrupted = false;

    QPointer<StatusBar> bar = new StatusBar(tr("Video") + " " + tr("Z Scan"), steps);
    connect(bar, SIGNAL(canceled()), SLOT(cancelOperation()));
    bar->add();

    QString name = path;
    if (name.lastIndexOf("/") != -1)
        name.remove(0, name.lastIndexOf("/") + 1);
    if (name.lastIndexOf("\\") != -1)
        name.remove(0, name.lastIndexOf("\\") + 1);

    QFile index(path + ".txt");
    index.open(QIODevice::WriteOnly | QIODevice::Truncate | QIODevice::Text);
    QTextStream str(&index);

    str << "NST_index_file" << endl;
    str << steps << ' ' << step << ' ' << start << endl;

    int time;
    QElapsedTimer timer;
    bool first = true;

    QDir dir(path + "_files/");
    dir.mkdir(path + "_files/");

    for (int i = 0; i <= steps && !isInterrupted; i++) {
        bar->setValue(i);

        go_objective_z(start + i*step);
        SleepExport::msleep_gui(first ? 2000 : 200);
        first = false;

        int frm = frames;
        time = 2000;
        timer.start();
        while (frm == frames && !isInterrupted && !timer.hasExpired(time)) {
            qApp->processEvents();
            SleepExport::msleep2(10);
        }
        if (timer.hasExpired(time)) {
            new Notification(tr("Video") + " " + tr("Z Scan") + ": " + tr("No frames from camera"), Notification::Warn, 10);

            destructStatus(bar);
            isInterrupted = true;
        }

        if (isInterrupted)
            continue;

        str << name + "_files/" + QString::number(i).rightJustified(5, '0') + ".png" << endl;
        ritem->image.save(path + "_files/" + QString::number(i).rightJustified(5, '0') + ".png", "png", 100);
    }

    index.close();

    destructStatus(bar);
    doBeep();

    if (!old_start)
        m_BStart->change_state_man_with_sig(false);

    setEnabled(true);
}
