#include "autocalib.hpp"
#include "dataplot.h"
#include "spmtunning.h"
#include "oscilloscope.h"
#include "mainwindow.h"
#include "EG3000.h"
#include "math/plot_SubtPlane.h"
#include <flowlayout.hpp>
#include <qwt_plot_marker.h>
#include <QSettings>
#include <QtEvents>
#include <QTableWidget>
#include <QApplication>
#include <QCheckBox>
#include <QDesktopWidget>
#include <QInputDialog>
#include <QFormLayout>
#include <QTimer>

AutoCalibPlot::AutoCalibPlot(QWidget *parent) : DelayedInitWidget(parent)
{
    isInterrupted = false;
    use_direct_get_data = false;
    sensorNoZ = false;
    for (int i = 0; i < 3; i++)
        is_optical_scanner[i] = false;


    scanplot = 0;
    dataplot = 0;
    spm = 0;
    activemarker = 0;
    activepoint = -1;

    calib_min = 0;
    calib_coeff = 1;

    CBScanner = new QComboBox();
    CBScanner->addItem(makeHugeIcon(":Head_page.png"), tr("Head"));
    CBScanner->addItem(makeHugeIcon(":stage_page.png"), tr("Stage"));

    BReturn = new QPushButton(makeHugeIcon(":/MainToolbar/Osc.png"), QString());
    BReturn->setToolTip(tr("Switch to oscilloscope"));

    QSettings settings("Param/gui.conf", QSettings::IniFormat);
    bool head_on = settings.value("display/oscilloscope/head", true).toBool();
    bool stage_on = settings.value("display/oscilloscope/stage", true).toBool();
    if (!head_on || !stage_on)
        CBScanner->setCurrentIndex(head_on ? 0 : 1);
    CBScanner->setEnabled(head_on && stage_on);
}

#include <limits>
bool AutoCalibPlot::eventFilter(QObject *obj, QEvent *Event)
{
    if (Event->type() == QEvent::KeyPress && ((QKeyEvent *)Event)->key() == Qt::Key_Delete) {
        QVector<int> rm;
        if (activepoint != -1) {
            rm.append(activepoint);
            activepoint = -1;
        } else if (activemarker) {
            rm.append(markers.indexOf(activemarker));
            activemarker = 0;
        } else if (TWvalues->hasFocus() && TWvalues->selectedItems().size()) {
            for (int i = TWvalues->rowCount()-1; i >= 0; i--)
                foreach (QTableWidgetItem *it, TWvalues->selectedItems())
                    if (TWvalues->item(i, 1) == it)
                        rm.append(i);
        }

        foreach (int i, rm) {
            TWvalues->removeRow(i);
            markers.at(i)->detach();
            delete markers.at(i);
            markers.remove(i);
            axisvalues.remove(i);
        }
        if (rm.size()) {
            bool from_source = CBSource->isChecked() || CBValue->currentIndex() == 3;
            if (!from_source)
                for (int i = 0; i < axisvalues.size(); i++)
                    axisvalues[i] = i * LEPeriod->value();

            ReCalcFromTable();
        }
    }

    if (obj == Source->Pure2dPlot) {
        if (Event->type() == QEvent::Resize) {
            //why we need it here?!! -- because plot core does not syncs font size for this axis, and not resets axis layout for different size
            Source->Pure2dPlot->QPlot->axisScaleDiv(QwtPlot::xTop)->invalidate();

            QFont font = qApp->font(); font.setPointSize(Source->Pure2dPlot->FontSize);
            Source->Pure2dPlot->QPlot->setAxisFont(QwtPlot::xTop, font);
        }
        if (Event->type() == QEvent::MouseButtonPress) {
            QPoint Pos = Source->Pure2dPlot->mapToCanvas(((QMouseEvent *)Event)->pos());
            Source->Pure2dPlot->SetClickParams(Pos);

            if (markers.size() == 0)
                return false;

            double mindist = std::numeric_limits<double>::max();
            int minidx = -1;
            for (int i = 0; i < markers.size(); i++) {
                double dist = (Pos - Source->Pure2dPlot->PlotToScreen(markers[i]->value())).manhattanLength();
                if (dist < mindist) {
                    mindist = dist;
                    minidx = i;
                }
            }

#if (QT_VERSION >= QT_VERSION_CHECK(5, 0, 0)) && !defined(_ADROID)
            double scale = qApp->desktop()->logicalDpiX();
#else
            double scale = qApp->desktop()->physicalDpiX();
#endif

            int Range = 1.5*qMax<int>(20, 0.25 * scale);
            activemarker = 0;
            if (mindist < Range)
                activemarker = markers[minidx];

            return activemarker != 0;
        }
        else if (Event->type() == QEvent::MouseButtonRelease) activemarker = 0;
        else if (Event->type() == QEvent::MouseMove && activemarker) {
            QPoint Pos = Source->Pure2dPlot->mapToCanvas(((QMouseEvent *)Event)->pos());
            activemarker->setValue(Source->Pure2dPlot->ScreenToPlot(Pos));
            Source->Pure2dPlot->Replot();
            ReCalcFromSourceMan();
            return true;
        }

        return false;
    }
    else
    if (obj == CalibCurve->Pure2dPlot && CBValue->currentIndex() != 3) {
        if (Event->type() == QEvent::MouseButtonPress) {
            QPoint Pos = CalibCurve->Pure2dPlot->mapToCanvas(((QMouseEvent *)Event)->pos());
            CalibCurve->Pure2dPlot->SetClickParams(Pos);

            if (curveDataCalib.size() == 0 || curveDataCalib.at(0)->size() == 0)
                return false;

            double mindist = std::numeric_limits<double>::max();
            int minidx = -1;
            for (int i = 0; i < curveDataCalib.at(0)->size(); i++) {
                double dist = (Pos - CalibCurve->Pure2dPlot->PlotToScreen(QPointF(curveDataCalib.at(0)->XSeq.at(i), curveDataCalib.at(0)->YSeq.at(i)))).manhattanLength();
                if (dist < mindist) {
                    mindist = dist;
                    minidx = i;
                }
            }

#if (QT_VERSION >= QT_VERSION_CHECK(5, 0, 0)) && !defined(_ADROID)
            double scale = qApp->desktop()->logicalDpiX();
#else
            double scale = qApp->desktop()->physicalDpiX();
#endif

            int Range = 1.5*qMax<int>(20, 0.25 * scale);
            activepoint = -1;
            if (mindist < Range)
                activepoint = minidx;

            return activepoint != -1;
        }
        else if (Event->type() == QEvent::MouseButtonRelease) activepoint = -1;
        else if (Event->type() == QEvent::MouseMove && activepoint != -1) {
            QPoint Pos = CalibCurve->Pure2dPlot->mapToCanvas(((QMouseEvent *)Event)->pos());
            bool from_source = CBSource->isChecked() || CBValue->currentIndex() == 3;
            if (from_source)
                curveDataCalib[0]->XSeq[activepoint] = CalibCurve->Pure2dPlot->ScreenXToPlotX(Pos.x());

            curveDataCalib[0]->YSeq[activepoint] = CalibCurve->Pure2dPlot->ScreenYToPlotY(Pos.y());

            CalibCurve->Refresh();
            ReCalcFromCalibCurve();
            return true;
        }

        return false;
    }
    else
        return QWidget::eventFilter(obj, Event);
}

#include <QScriptEngine>
void AutoCalibPlot::customFilter(bool Checked)
{
    QString filter = Source->Pure2dPlot->CustomFilter;
    if (Checked) {
        filter = QInputDialog::getText(this, tr("Custom filter"),
                                       "<a href=http://www.javascriptkit.com/javatutors/math.shtml>"
                                       "http://www.javascriptkit.com/javatutors/math.shtml" "</a><br />" "\n" "f(x) = ",
                                       QLineEdit::Normal, filter, &Checked);

        if (!Checked)
            filter = Source->Pure2dPlot->CustomFilter;
        else {
            QString flt = "function filter(x) { with(Math) { var result = ( %1 ); } return result; }";
            flt = flt.arg(filter);

            QScriptEngine myEngine;
            myEngine.evaluate(flt);
            QScriptValue fun = myEngine.globalObject().property("filter");
            QScriptValue res = fun.call(QScriptValue(), QScriptValueList() << 0);
            if (!fun.isFunction() || fun.isError() || res.isError()) {
                Checked = false;
                new Notification(tr("Custom filter") + ": " + tr("Wrong filter"), Notification::Warn, 10);
                doBeep();
            }
        }
    }
    Source->Pure2dPlot->SetCustomFilter(Checked, filter);
}
void AutoCalibPlot::BGetClicked()
{
    plot::TPure2dPlot *plot = 0;
    if (CBInput->currentIndex() == 0 && scanplot)
        plot = scanplot;
    else if (CBInput->currentIndex() == 1 && dataplot)
        plot = dataplot->MyPl->ThreeDPlot->TwoDPlot->Pure2dPlot;
    else if (CBInput->currentIndex() == 2 && dataplot)
        plot = dataplot->MyPl->TwoDPlot->Pure2dPlot;
    else if (CBInput->currentIndex() == 3 && spm)
        plot = spm->MyPl->Pure2dPlot;
    if (!plot)
        return;

    curveData = plot::getLock(plot->SourceData);
    plot::TSIUnit xunit = plot->XAxisUnit;
    plot::TSIUnit yunit = plot->YAxisUnit;

    int active = qBound(0, plot->ActiveCurve, curveData.size() - 1);
    while (curveData.size() && !(curveData.at(active) && curveData.at(active)->size())) {
        curveData.remove(active);
        active = 0;
    }
    if (!curveData.size()) {
        curveData.clear();
        new Notification(tr("Auto calibration") + ": " + tr("Source is empty!"), Notification::Warn, 5);
        doBeep();
        return;
    }
    curveData = (plot::TCurvesDataOwner() << curveData.at(active));
    curveDataCalib.clear();

    Source->Pure2dPlot->XAxisUnit = xunit;
    Source->Pure2dPlot->YAxisUnit = yunit;
    Source->Pure2dPlot->SourceData = plot::getUnLock(curveData);
    Source->Refresh();

    ReCalcFromSource();
}

void AutoCalibPlot::BSendClicked()
{
    new Notification(tr("Autocalib") + ": " + tr("Coefficients set succesfully"), Notification::Good, 5);
    doBeep();

    if (CBValue->currentIndex() == 2) { //pid nlc
        double *edit = &osc->CalibWidget->nlcPID[CBScanner->currentIndex()][CBAxis->currentIndex()][0];
        for (int i = 0; i < 6; i++)
            edit[i] = lineC[i]->text().toDouble();

        osc->CalibWidget->Scanner_type->setCurrentIndex(CBScanner->currentIndex());
        osc->CalibWidget->nlcPIDSend();
        return;
    }

    if (CBValue->currentIndex() == 3) { //SPM PID
        if (!spm) return;
        QSettings settings("Param/controller.conf", QSettings::IniFormat);
        double ifactor = settings.value("pi_autocalib_spm_i_scale", 100).toDouble();
        double pfactor = settings.value("pi_autocalib_spm_p_scale", 0.25).toDouble();
        spm->DCfbI->setValue(qMax(0.001, lineC[2]->text().toDouble() / ifactor / 20000 * 10000));
        spm->DCfbP->setValue(lineC[2]->text().toDouble() *pfactor);
        spm->DCfbP->setNewValue();
        spm->DCfbI->setNewValue();

        return;
    }

    void *calib_arr = 0;
    NLCWidget **calib_wid = 0;
    if (CBScanner->currentIndex() == 0) {
        switch (CBValue->currentIndex()) {
        case 0:
            calib_arr = osc->CalibWidget->linePiezoCalibration;
            calib_wid = osc->CalibWidget->viewPiezoCalibration;
            break;
        case 1:
            calib_arr = osc->CalibWidget->lineSensorCalibration;
            calib_wid = osc->CalibWidget->viewSensorCalibration;
            break;
        }
    } else {
        switch (CBValue->currentIndex()) {
        case 0:
            calib_arr = osc->CalibWidget->linePiezoCalibration2;
            calib_wid = osc->CalibWidget->viewPiezoCalibration2;
            break;
        case 1:
            calib_arr = osc->CalibWidget->lineSensorCalibration2;
            calib_wid = osc->CalibWidget->viewSensorCalibration2;
            break;
        }
    }
    if (calib_arr) {
        double *calib_arr2 = &((double *)calib_arr)[CBAxis->currentIndex()*6];
        for (int i = 0; i < 6; i++)
            calib_arr2[i] = lineC[5-i]->text().toDouble();
        calib_wid[CBAxis->currentIndex()]->SetPic(buildNlcPic(calib_arr2));
    }

    osc->CalibWidget->Scanner_type->setCurrentIndex(CBScanner->currentIndex());
    osc->CalibWidget->s_sendData();
}

void AutoCalibPlot::cbSrcMode()
{
    bool on = CBSource->isChecked();
    CBLevel->setEnabled(!on);
    LELevel->setEnabled(!on);
    LEPeriod->setEnabled(!on);
}

void AutoCalibPlot::channelChanged()
{
    bool recalc = CBValue->currentIndex() == 3 || !CBAxis->isEnabled();
    CBAxis->setEnabled(CBValue->currentIndex() != 3);
    CBSource->setEnabled(CBValue->currentIndex() != 3);
    if (!CBSource->isEnabled() && CBSource->isChecked())
        CBSource->setChecked(false);
    LEPeriod->setEnabled(CBValue->currentIndex() != 3 && !CBSource->isChecked());

    CBDerivative->setVisible(CBValue->currentIndex() == 2);
    CBLevel->setVisible(CBValue->currentIndex() != 2);
    if (!CBDerivative->isVisible() && CBDerivative->isChecked()) {
        CBDerivative->setChecked(false);
        recalc = true;
    }
    if (recalc) {
        UpdateLayout();
        ReCalcFromSource();
    }
}

void AutoCalibPlot::UpdateLayout()
{
    Source->setVisible(!CBSource->isChecked());
    CalibCurve->setVisible(CBValue->currentIndex() != 3);

    lineC[0]->setVisible(CBDerivative->isChecked());
    coeffOk->setVisible(!CBDerivative->isChecked());
}

void AutoCalibPlot::ReCalcCoefficent()
{
    bool from_source = CBSource->isChecked() || CBValue->currentIndex() == 3;
    bool fit_derivative = CBDerivative->isChecked();
    int npower = qMin(fit_derivative ? 6 : 4, adcvalues.size()-1);
    if (npower < 0)
        return;

    double coeffs[7];
    memset(coeffs, 0, sizeof(double)*7);
    plot::math::dataLinePartFitPolynom(adcvalues.data(), npower, coeffs, 0, adcvalues.size());

    double scale = LEPeriod->value();
    if (from_source && axisvalues.size() > 1)
        scale = axisvalues.at(1) - axisvalues.at(0);

    if (fit_derivative) {
        coeffs[6] /= scale*scale*scale*scale*scale*scale;
        coeffs[5] /= scale*scale*scale*scale*scale;
    }
    coeffs[4] /= scale*scale*scale*scale;
    coeffs[3] /= scale*scale*scale;
    coeffs[2] /= scale*scale;
    coeffs[1] /= scale;
    //coeffs[0] /= 1;

    if (CBValue->currentIndex() == 3 && npower == 1) { //SPM PID
        double z_range = (CBScanner->currentIndex() == 0) ? osc->CalibWidget->Scanner_data_pid_cal_range[47] : osc->CalibWidget->Scanner2_data_pid_cal_range[47];//mkm
        coeffs[1] = qFabs(coeffs[1]) * z_range * 1e-6 / 65535. / 20. * 65535.; //z->16b, spm_sig->16b
        coeffs[0] = 0;
    }

    if (!fit_derivative) {
        for (int i = 0; i < 5; i++)
            lineC[i+1]->setText(QString::number(coeffs[i]));
        lineC[0]->setText(0);
    } else {
        for (int i = 0; i < 6; i++)
            lineC[i]->setText(QString::number(coeffs[i+1]));
        c0 = coeffs[0];
    }
}

void AutoCalibPlot::ReCalcFromCalibCurve()
{
    bool from_source = CBSource->isChecked() || CBValue->currentIndex() == 3;
    if (from_source)
        axisvalues.clear();
    adcvalues.clear();

    for (int i = 0; i < curveDataCalib.at(0)->size(); i++) {
        if (!from_source) {
            //axisvalues.push_back(curveDataCalib.at(0)->XSeq.at(i)*1e6);
            adcvalues.push_back(curveDataCalib.at(0)->YSeq.at(i));
            markers[i]->setXValue((adcvalues.back() / calib_coeff) + calib_min);
        } else {
            axisvalues.push_back(curveDataCalib.at(0)->XSeq.at(i));
            adcvalues.push_back(curveDataCalib.at(0)->YSeq.at(i));
            markers[i]->setYValue(adcvalues.back());
            markers[i]->setXValue((axisvalues.back() / calib_coeff) + calib_min);
        }
    }

    ValuesUpdate();
}

void AutoCalibPlot::ReCalcFromSourceMan()
{
    bool from_source = CBSource->isChecked() || CBValue->currentIndex() == 3;
    if (from_source)
        axisvalues.clear();
    adcvalues.clear();

    for (int i = 0; i < markers.size(); i++) {
        if (!from_source)
            adcvalues.push_back((markers[i]->xValue() - calib_min)*calib_coeff);
        else {
            axisvalues.push_back((markers[i]->xValue() - calib_min)*calib_coeff);
            adcvalues.push_back(markers[i]->yValue());
        }
    }

    ValuesUpdate();
}

void AutoCalibPlot::ReCalcFromTable()
{
    bool from_source = CBSource->isChecked() || CBValue->currentIndex() == 3;
    if (from_source)
        axisvalues.clear();
    adcvalues.clear();

    for (int i = 0; i < TWvalues->rowCount(); i++) {
        adcvalues.push_back(TWvalues->item(i, 1)->text().toDouble());
        if (!from_source)
            markers[i]->setXValue((adcvalues.back() / calib_coeff) + calib_min);
        else {
            axisvalues.push_back(TWvalues->item(i, 0)->text().toDouble());
            markers[i]->setYValue(adcvalues.back());
            markers[i]->setXValue((axisvalues.back() / calib_coeff) + calib_min);
        }
    }

    Source->Refresh();
    ValuesUpdate();
}

void AutoCalibPlot::ReCalcFromSource()
{
    while (!Source->Pure2dPlot->isMathReady()) {
        qApp->processEvents();
        SleepExport::msleep2(3);
    }
    Source->Pure2dPlot->s_Refresh();

    if (!Source->Pure2dPlot->QwtCurves.size() || !Source->Pure2dPlot->QwtCurves[0]->Curve.dataSize()) {
        //new Notification(tr("Auto calibration") + ": " + tr("Source is empty!"), Notification::Warn, 5);
        //doBeep();
        return;
    }

    QwtPlotCurve &curve = Source->Pure2dPlot->QwtCurves[0]->Curve;
    bool from_source = CBSource->isChecked();

    bool falling_edge = CBLevel->isChecked();
    double lvl = LELevel->value();
    double level = (1-lvl)*curve.minYValue() + lvl*curve.maxYValue();
    double prev = curve.sample(0).y();

    calib_min = curve.sample(0).x();
    calib_coeff = (CBValue->currentIndex() == 0 && CBAxis->currentIndex() < 2 ? setpoint_limit_top(65535) : 65535) / ((curve.sample(curve.dataSize() - 1).x()) - calib_min);
    if (CBValue->currentIndex() == 3) { //SPM PID
        calib_min = 0;
        calib_coeff = 1;
    }

    adcvalues.clear();
    axisvalues.clear();
    axisvalues.push_back(0);
    foreach (QwtPlotMarker *marker, markers) {
        marker->detach();
        delete marker;
    }
    markers.clear();

    levelMarker->setYValue(level);

    if (CBValue->currentIndex() == 3) { //SPM PID
        int imin = 0;
        double minv = std::numeric_limits<double>::max();
        for (unsigned int i = 1; i < curve.dataSize(); i++) {
            if (curve.sample(i).y() < minv) {
                minv = curve.sample(imin).y();
                imin = i;
            }
        }
        QwtPlotMarker *marker = new QwtPlotMarker;
        marker->attach(Source->Pure2dPlot->QPlot);
        marker->setSymbol(
                new QwtSymbol(QwtSymbol::Diamond, QColor(255,255,255), QColor(255,255,255), QSize(9, 9)));
        marker->setValue(curve.sample(imin));
        marker->setLineStyle(QwtPlotMarker::VLine);
        marker->setLinePen(QPen(Qt::white, 1, Qt::DashLine));
        markers.push_back(marker);

        axisvalues.clear();
        adcvalues.push_back(curve.sample(imin).y());
        axisvalues.push_back(curve.sample(imin).x());
    }

    bool need0 = true;
    for (unsigned int i = 1; i < curve.dataSize(); i++) {
        if (!from_source) {
            if ((falling_edge && prev >= level && curve.sample(i).y() < level) ||
                    (!falling_edge && prev <= level && curve.sample(i).y() > level)) {
                double k = (level - curve.sample(i).y()) / (prev - curve.sample(i).y());
                double x = QLineF(curve.sample(i), curve.sample(i-1)).pointAt(k).x();

                QwtPlotMarker *marker = new QwtPlotMarker;
                marker->attach(Source->Pure2dPlot->QPlot);
                marker->setSymbol(
                            new QwtSymbol(QwtSymbol::Diamond, QColor(255,255,255), QColor(255,255,255), QSize(9, 9)));
                marker->setXValue(x);
                marker->setYValue(level);
                marker->setLineStyle(QwtPlotMarker::VLine);
                marker->setLinePen(QPen(Qt::white, 1, Qt::DashLine));
                markers.push_back(marker);

                if (CBValue->currentIndex() == 3) { //SPM PID
                    adcvalues.push_back(level);
                    axisvalues.push_back(x);
                } else {
                    adcvalues.push_back((x - calib_min)*calib_coeff);
                    axisvalues.push_back(axisvalues.back() + LEPeriod->value());
                }
            }

            prev = curve.sample(i).y();
        } else {
            if (need0) {
                i = 0;
                axisvalues.clear();
                need0 = false;
            }
            double x = curve.sample(i).x();

            QwtPlotMarker *marker = new QwtPlotMarker;
            marker->attach(Source->Pure2dPlot->QPlot);
            marker->setSymbol(
                        new QwtSymbol(QwtSymbol::Diamond, QColor(255,255,255), QColor(255,255,255), QSize(9, 9)));
            marker->setXValue(x);
            marker->setYValue(curve.sample(i).y());
            marker->setLineStyle(QwtPlotMarker::NoLine);
            marker->setLinePen(QPen(Qt::white, 1, Qt::DashLine));
            markers.push_back(marker);

            adcvalues.push_back(curve.sample(i).y());
            axisvalues.push_back((x - calib_min)*calib_coeff);
        }
    }

    ValuesUpdate();
}

template<class T>
double NewtonResult(T arr[5], double x0, double value, double ac, DWORD max = 100)
{
    double x_new = x0;
    WORD N = 0;

    do {
        x0 = x_new;
        x_new = x0 -
                (((((arr[0] * x0 + arr[1]) * x0 + arr[2]) * x0 + arr[3]) * x0 + arr[4]) - value) /
                (((4*arr[0] * x0 + 3*arr[1]) * x0 + 2*arr[2]) * x0 + arr[3]);
        N++;
    } while ((fabs(x0 - x_new) > ac) && (N < max));

    return x_new;
}

inline
double polynom_at_pos(double pos, double coeff[], int count = 4) {
    double acc = coeff[0];
    for (int i = 1; i <= count; i++)
        acc = acc * pos + coeff[i];

    return acc;
}

void AutoCalibPlot::ValuesUpdate()
{
    bool from_source = CBSource->isChecked() || CBValue->currentIndex() == 3;

    TWvalues->blockSignals(true);

    TWvalues->clear();
    QList<QString> header;
    header.push_back(from_source ? tr("Value") : tr("Pos, ")+tr("micro")+tr("m"));
    header.push_back(from_source ? tr("Pos") : tr("Counts"));
    TWvalues->setHorizontalHeaderLabels(header);

    QTableWidgetItem *newItem;
    TWvalues->setRowCount(adcvalues.size());

    if (!from_source) {
        CalibCurve->Pure2dPlot->XAxisUnit.SetDescAndSIUnit(QString(), tr("m"));
        CalibCurve->Pure2dPlot->YAxisUnit.SetDescAndSIUnit(tr("Counts"));
    } else {
        CalibCurve->Pure2dPlot->XAxisUnit.SetDescAndSIUnit(tr("Counts"));
        CalibCurve->Pure2dPlot->YAxisUnit = Source->Pure2dPlot->YAxisUnit;
    }
    CalibCurve->Pure2dPlot->Title = tr("Calibration curve");
    Source->Pure2dPlot->Title = tr("Source curve");

    curveDataCalib.clear();
    plot::ResizeOwner(curveDataCalib, 2);
    curveDataCalib[0]->Pen.setColor(Qt::black);
    curveDataCalib[1]->Pen.setColor(Qt::red);

    for (int i = 0; i < adcvalues.size(); i++) {
        newItem = new QTableWidgetItem(QString::number(axisvalues[i]));
        if (!from_source)
            newItem->setFlags(Qt::NoItemFlags);
        TWvalues->setItem(i, 0, newItem);

        newItem = new QTableWidgetItem(QString::number(adcvalues[i]));
        TWvalues->setItem(i, 1, newItem);

        curveDataCalib[0]->AddPoint(axisvalues[i]*(from_source ? 1 : 1e-6), adcvalues[i]);
    }

    ReCalcCoefficent();

    double coeff[6];
    for (int i = 0; i < 6; i++)
        coeff[i] = lineC[5 - i]->text().toDouble();
    bool fit_derivative = CBDerivative->isChecked();
    double val = -1;

    if (!from_source) {
        curveDataCalib[1]->setXCalib(-5e-6, 1e-6);
        for (int i = -5; i < axisvalues.last(); i++) {
            if (!fit_derivative)
                val = polynom_at_pos(i, coeff);
            else
                val = polynom_at_pos(i, coeff, 5)*i + c0;
            curveDataCalib[1]->AddY(val);
        }
    } else {
        curveDataCalib[1]->setXCalib(-(axisvalues.at(1) - axisvalues.at(0)), (axisvalues.at(1) - axisvalues.at(0))/10.);
        for (int x = -10; x < axisvalues.size()*10; x++) {
            double i = x*(axisvalues.at(1) - axisvalues.at(0))/10.;
            if (!fit_derivative)
                val = polynom_at_pos(i, coeff);
            else
                val = polynom_at_pos(i, coeff, 5)*i + c0;

            curveDataCalib[1]->AddY(val);
        }
    }

    Source->Refresh();
    CalibCurve->Pure2dPlot->SourceData = plot::getUnLock(curveDataCalib);
    CalibCurve->Refresh();
    CalibCurve->Pure2dPlot->QwtCurves[0]->Curve.setSymbol(
            new QwtSymbol(QwtSymbol::Ellipse, QBrush(Qt::gray), QPen(Qt::blue), QSize(4, 4)));

    TWvalues->blockSignals(false);

    if (fit_derivative)
        return;

    //check newton converges
    double CalibRange[5];
    double Shift[5];
    const double acc = 0.001;

    bool ok = true;
    int x[] = {1000, 2000, 5000, 10000, 50000};
    for (int i = 0; i < 5; i++) {
        int maxx = x[i];
        Shift[i] = NewtonResult(coeff, 10, 0, acc, maxx);
        CalibRange[i] = NewtonResult(coeff, 50, 65535, acc, maxx);
        if (i > 0) if (fabs(Shift[i-1] - Shift[i]) > acc || fabs(CalibRange[i-1] - CalibRange[i]) > acc) {
            ok = false;
            break;
        }
    }

    if (!ok && CBValue->currentIndex() != 2) { // !PID NLC
        coeffOk->setText(tr("Invalid curve"));
        coeffOk->setIcon(makeHugeIcon(":/SPMTuning_window/off.png"));
        disconnect(coeffOk, SIGNAL(clicked()), this, SLOT(BSendClicked()));
    } else {
        coeffOk->setText(tr("Valid curve") + ", " + tr("click to send"));
        coeffOk->setIcon(makeHugeIcon(":/SPMTuning_window/onn.png"));
        connect(coeffOk, SIGNAL(clicked()), this, SLOT(BSendClicked()), Qt::UniqueConnection);
    }
}

void AutoCalibPlot::GuiCreate()
{
    QHBoxLayout *mainL = new QHBoxLayout(this);
    mainL->setContentsMargins(0, 0, 0, 0);

    QVBoxLayout *rightL1 = new QVBoxLayout;
    QVBoxLayout *plotSplit = new QVBoxLayout;
    plotSplit->setSpacing(9);

    mainL->addLayout(rightL1);
    mainL->addLayout(plotSplit, 1);


    Source = new plot::T2dPlot(0, false, this);
    Source->layout()->setContentsMargins(0, 0, 6, 0);

    CalibCurve = new plot::T2dPlot(0, false, this);
    CalibCurve->layout()->setContentsMargins(0, 0, 6, 0);

    plotSplit->addWidget(Source);
    plotSplit->addWidget(CalibCurve);


    QHBoxLayout *highL11 = new QHBoxLayout();
    highL11->setSpacing(6);

    QGroupBox *L2grBox = new QGroupBox(this);
    QVBoxLayout *rightL2 = new QVBoxLayout(L2grBox);
    QGridLayout *highD = new QGridLayout;
    QFormLayout *highLopt = makeFormLayout();
    rightL2->addLayout(highD);
    rightL2->addLayout(highLopt);

    TWvalues = new QTableWidget(0, 2, this);
    QList<QString> header;
    header.push_back(tr("Pos, ")+tr("micro")+tr("m"));
    header.push_back(tr("Counts"));
    TWvalues->setHorizontalHeaderLabels(header);
    //TWvalues->horizontalHeader()->setResizeMode(QHeaderView::Fixed);
    //TWvalues->horizontalHeader()->resizeSection(0, 60);
    //TWvalues->horizontalHeader()->resizeSection(1, 80);
    connect(TWvalues, SIGNAL(itemChanged(QTableWidgetItem*)), this, SLOT(ReCalcFromTable()));

    QGroupBox *CoeffGB = new QGroupBox(tr("Coefficients"), this);
    QFormLayout *CoeffL = makeFormLayout(CoeffGB);

    rightL1->addLayout(highL11);
    rightL1->addWidget(L2grBox);
    rightL1->addWidget(TWvalues);
    rightL1->addWidget(CoeffGB);

    { //fixed scale for source data
        plot::TScale *sc;
        Source->Pure2dPlot->QPlot->setAxisScaleDraw(QwtPlot::xTop, sc = new plot::TScale);
        Source->Pure2dPlot->QPlot->setAxisScaleEngine(QwtPlot::xTop, new plot::TScaleDiv(sc));
        Source->Pure2dPlot->QPlot->enableAxis(QwtPlot::xTop);

        QFont font = qApp->font(); font.setPointSize(Source->Pure2dPlot->FontSize);
        QwtText tt("Counts"); tt.setFont(font);
        Source->Pure2dPlot->QPlot->setAxisTitle(QwtPlot::xTop, tt);
        Source->Pure2dPlot->QPlot->setAxisFont(QwtPlot::xTop, font);
        ((plot::TScale *)Source->Pure2dPlot->QPlot->axisScaleDraw(QwtPlot::xTop))->font = font;
        Source->Pure2dPlot->QPlot->setAxisScale(QwtPlot::xTop, 0, 65535);
    }

    Source->Pure2dPlot->XAxisUnit.SetDescAndSIUnit(tr("Counts"));
    Source->Pure2dPlot->YAxisUnit.SetDescAndSIUnit(tr("Z"));
    Source->Pure2dPlot->Title = tr("Source curve");

    CalibCurve->Pure2dPlot->XAxisUnit.SetDescAndSIUnit(QString(), tr("m"));
    CalibCurve->Pure2dPlot->YAxisUnit.SetDescAndSIUnit(tr("Counts"));
    CalibCurve->Pure2dPlot->Title = tr("Calibration curve");

    Source->PlotControl->hide();
    CalibCurve->PlotControl->hide();

    Source->Refresh();
    CalibCurve->Refresh();

    coeffOk = new QPushButton;
    coeffOk->setFlat(true);

    CBInput = new QComboBox();
    CBInput->addItem(tr("Scan Section"));
    CBInput->addItem(tr("Data 3d Section"));
    CBInput->addItem(tr("Data 2d"));
    CBInput->addItem(tr("SPM"));

    CBValue = new QComboBox();
    CBValue->addItem(tr("Voltage"));
    CBValue->addItem(tr("Sensor"));
    CBValue->addItem(tr("PID") + " " + tr("dV/dC"));
    CBValue->addItem(tr("SPM") + " " + tr("PID"));
    connect(CBValue, SIGNAL(currentIndexChanged(int)), this, SLOT(channelChanged()));

    CBAxis = new QComboBox();
    CBAxis->addItem("X");
    CBAxis->addItem("Y");
    CBAxis->addItem("Z");

    TwoStateButton *BAutoZoom = new TwoStateButton(":/plot/allow_deny_autoscale.png",
                                                   ":/plot/allow_deny_autoscale_on.png");
    BAutoZoom->setToolTip(QObject::tr("Enable autoscaling"));
    BAutoZoom->setCheckable(true);
    BAutoZoom->change_state_man(true);
    connect(BAutoZoom, SIGNAL(clicked()), Source->PlotControl->BAutoScale, SLOT(animateClick()));
    connect(BAutoZoom, SIGNAL(clicked()), CalibCurve->PlotControl->BAutoScale, SLOT(animateClick()));

    QPushButton *BGet = new QPushButton(makeHugeIcon(":/AutoCalib/CalImport.png"), QString());
    BGet->setToolTip(tr("Get data"));

    BSend = new QPushButton(makeHugeIcon(":/AutoCalib/MoveToTable.png"), QString());
    BSend->setToolTip(tr("Save to settings"));

    BSensorsAutoAdj = new QPushButton(makeHugeIcon(":/AutoCalib/SensorsAutoAdjOf.png"), QString());
    BSensorsAutoAdj->setToolTip(tr("Run Capacitance sensors Auto Adjustment"));

    BPIDCalib = new QPushButton(makeHugeIcon(":/AutoCalib/PIDtune.png"), QString());
    BPIDCalib->setToolTip(tr("Find PI for selected scanner"));

    LEPeriod = new Edit(0.1, 100, 2);
    LEPeriod->setValue(3);
    LEPeriod->setImmediateMode();
    LELevel = new Edit(0.1, 0.9, 1);
    LELevel->setValue(0.5);
    LELevel->setImmediateMode();

    QCheckBox *CBcustom3dfilter = new QCheckBox(tr("Custom source filter"));
    connect(CBcustom3dfilter, SIGNAL(toggled(bool)), this, SLOT(customFilter(bool)));
    if (!show_advanced_gui())
        CBcustom3dfilter->hide();

    CBDerivative = new QCheckBox(tr("Find 6th order derivative")); CBDerivative->hide();
    CBLevel = new QCheckBox(tr("Falling edge"));
    CBSource = new QCheckBox(tr("Fit source curve"));
    connect(CBSource, SIGNAL(toggled(bool)), this, SLOT(cbSrcMode()));

    for (int i = 0; i < 6; i++) {
        lineC[i] = new QLineEdit();
        lineC[i]->setReadOnly(true);
    }
    lineC[0]->hide(); //hidden, only for derivative


    highL11->addWidget(BReturn);
    highL11->addWidget(Source->PlotControl->TBMathProcessing);
    highL11->addWidget(BAutoZoom);
    highL11->addWidget(Source->PlotControl->BResetMCCurve);
    highL11->addWidget(Source->PlotControl->BApply);
    highL11->addWidget(Source->PlotControl->BCancel);
    addDelimeterV(highL11, BReturn->sizeHint().height());
    highL11->addWidget(BSensorsAutoAdj);
    highL11->addWidget(BPIDCalib);
    highL11->setAlignment(Qt::AlignLeft);

    highD->addWidget(new QLabel(tr("Scanner:")),0,0);
    highD->addWidget(new QLabel(tr("Source:")),1,0);
    highD->addWidget(new QLabel(tr("Channel:")),2,0);
    highD->addWidget(new QLabel(tr("Axis:")),3,0);
    highD->addWidget(CBScanner,0,1);
    highD->addWidget(CBInput,1,1);
    highD->addWidget(CBValue,2,1);
    highD->addWidget(CBAxis,3,1);
    highD->addWidget(BGet,0,2, 2, 1, Qt::AlignBottom);
    highD->addWidget(BSend,2,2, 2, 1, Qt::AlignBottom);

    highLopt->addRow(CBcustom3dfilter);
    highLopt->addRow(CBSource);
    highLopt->addRow(CBDerivative);
    highLopt->addRow(CBLevel);
    highLopt->addRow(tr("Level")+", 0..1", LELevel);
    highLopt->addRow(tr("Period, ")+ tr("micro")+("m"), LEPeriod);

    CoeffL->addRow(tr("A = "), lineC[5]);
    CoeffL->addRow(tr("B = "), lineC[4]);
    CoeffL->addRow(tr("C = "), lineC[3]);
    CoeffL->addRow(tr("D = "), lineC[2]);
    CoeffL->addRow(tr("E = "), lineC[1]);
    CoeffL->addRow(0, lineC[0]);
    CoeffL->addRow(coeffOk);


    QTimer *delayUpdate = new QTimer(this);
    delayUpdate->setInterval(100);
    delayUpdate->setSingleShot(true);

    levelMarker = new QwtPlotMarker;
    levelMarker->setYValue(0);
    levelMarker->setLinePen(QPen(QColor(0, 255, 255), 0, Qt::SolidLine));
    levelMarker->setLineStyle(QwtPlotMarker::HLine);
    levelMarker->attach(Source->Pure2dPlot->QPlot);

    connect(BGet, SIGNAL(clicked()), SLOT(BGetClicked()));
    connect(BSend, SIGNAL(clicked()), SLOT(BSendClicked()));
    connect(BSensorsAutoAdj, SIGNAL(clicked()), this, SLOT(SensorReCalc()));
    connect(BPIDCalib, SIGNAL(clicked()), this, SLOT(PICalcAuto()));

    //connect(TWvalues, SIGNAL(itemChanged(QTableWidgetItem*)), this, SLOT(ReCalcFromTable()));

    connect(CBDerivative, SIGNAL(clicked()), this, SLOT(UpdateLayout()));
    connect(CBSource, SIGNAL(clicked()), this, SLOT(UpdateLayout()));

    connect(delayUpdate, SIGNAL(timeout()), this, SLOT(ReCalcFromSource()));
    connect(LEPeriod, SIGNAL(newValue()), this, SLOT(ReCalcFromSource()));
    connect(LELevel, SIGNAL(newValue()), this, SLOT(ReCalcFromSource()));
    connect(CBDerivative, SIGNAL(clicked()), this, SLOT(ReCalcFromSource()));
    connect(CBLevel, SIGNAL(clicked()), this, SLOT(ReCalcFromSource()));
    connect(CBSource, SIGNAL(clicked()), this, SLOT(ReCalcFromSource()));
    connect(CBcustom3dfilter, SIGNAL(clicked()), delayUpdate, SLOT(start()));
    connect(Source->PlotControl->BApply, SIGNAL(clicked()), delayUpdate, SLOT(start()));
    for (int i = 0; i < Source->PlotControl->AGFitOrders->actions().size()-1; i++)
        connect(Source->PlotControl->AGFitOrders->actions()[i], SIGNAL(triggered()), delayUpdate, SLOT(start()));

    Source->Pure2dPlot->installEventFilter(this);
    CalibCurve->Pure2dPlot->installEventFilter(this);
    TWvalues->installEventFilter(this);
}
