#include "mainwindow.h"
#include "pms400a.h"
#include <QTimer>
#include <QApplication>

#ifndef USE_PMS_EMULATOR

PmsController::PmsController()
{
    start_pt = 0;
    stop_pt = 32767;
    block_length = stop_pt - start_pt + 1;

    dataA = new unsigned long[block_length];
    dataB = new unsigned long[block_length];
    timeA = new float[block_length];
    timeB = new float[block_length];

    checkTimer = new QTimer(this);
    checkTimer->setInterval(10);
    connect(checkTimer, SIGNAL(timeout()), this, SLOT(checkStatus()));
}

PmsController::~PmsController()
{
    Deinit();
    delete dataA;
    delete dataB;
    delete timeA;
    delete timeB;
}

short PmsController::StartScan()
{
    PMS_fill_memory(0, -1, 0, 32767, 0);
    PMS_test_if_busy(&pms_state);

    ret = PMS_start_measure();
    if (ret == 0) {
        emit status(EDplay);
        checkTimer->start();
    }

    return ret;
}

short PmsController::StopScan()
{
    checkTimer->stop();
    emit scan_finished();
    emit status(EDok);
    return PMS_stop_measure();
}

void PmsController::checkStatus()
{
    short prev = pms_state;
    PMS_test_if_busy(&pms_state);

    if (pms_state != prev && pms_state == 0) {
        checkTimer->stop();
        emit scan_finished();
        emit status(EDok);
    }
}

short PmsController::Init()
{
    char pms_error_string[128] ;
    PMSModInfo mod_info;

    ret = PMS_init("PMS400A.INI");

    if (ret == -PMS_LICENSE_ERR)
      return ret;              // DLL not registered - Wrong or missing license key

    if (ret < 0 && -ret < PMS_NO_ACT_MOD)
      return ret;              // fatal errors

    PMS_get_module_info(0, &mod_info);
    if (mod_info.module_type == M_WRONG_TYPE)
        return -1;

    if (mod_info.init != INIT_OK)
        if (mod_info.init != INIT_MOD_IN_USE)
            return -2;

    memset(in_use, 0, MAX_NO_OF_PMS*sizeof(short));
    in_use[0] = 1;
    ret = PMS_set_mode(PMS_HARD, 1, in_use);
    PMS_get_module_info(0, &mod_info);

    if (ret < 0) {
      PMS_get_test_error_string( pms_error_string );
      qDebug() << pms_error_string;
      return -3;
    }

    SetParam(ACTIVE, 1);
    SetParam(TRIGGER, 0);
    SetParam(ENABLE_MEAS, 1);
    SetParam(MEAS_MODE, EVENT_MODE);
    SetParam(ACCUMULATE, 0);
    SetParam(START_PTR, start_pt);
    SetParam(END_PTR, stop_pt);
    SetParam(SWEEPS, 0);

    SetParam(COLLECT_TIME, 0.1);
    SetParam(MACRO_TIME, 0.5);
    SetParam(INP_HOLDOFF, 1);

    SetParam(GATE_LEVEL_A, 0.5);
    SetParam(GATE_LEVEL_B, 0.5);
    SetParam(EVENT_THRESHOLD_A, 1);
    SetParam(EVENT_THRESHOLD_B, 1);
    SetParam(INP_THRESHOLD_A, -0.05);
    SetParam(INP_THRESHOLD_B, 0.05);

    return 0;
}

void PmsController::GetResult(quint32 *chA, quint32 *chB)
{
    PMS_read_events(0, &events_read, start_pt, stop_pt,
                      dataA, dataB, timeA, timeB);

    double time;
    PMS_get_macro_time(0, &time);

    unsigned long coll = 0;
    for (unsigned int i = 0; i < events_read; i++)
        coll += dataA[i];
    if (chA)
        *chA = coll;

    coll = 0;
    for (unsigned int i = 0; i < events_read; i++)
        coll += dataB[i];
    if (chB)
        *chB = coll;
}

double PmsController::GetParam(short par_id)
{
    float var;
    PMS_get_parameter(0, par_id, &var);
    return var;
}

void PmsController::SetParam(short par_id, float val)
{
    PMS_set_parameter(0, par_id, val);
}

short PmsController::Deinit()
{
    PMS_stop_measure();

    memset(in_use, 0, MAX_NO_OF_PMS*sizeof(short));
    ret = PMS_set_mode(PMS_HARD, 1, in_use);

    return PMS_close();
}

#else //USE_PMS_EMULATOR - true
static double __x = 0;
static double __y = 0;
static double __delay = 100;

PmsController::PmsController()
{
}
PmsController::~PmsController()
{
}
short PmsController::StartScan()
{
    emit status(EDplay);

    SleepExport::msleep_gui(__delay);

    emit scan_finished();
    emit status(EDok);
    return 0;
}
short PmsController::StopScan()
{
    emit status(EDok);
    return 0;
}
void PmsController::checkStatus()
{
}
short PmsController::Init()
{
    return 0;
}
void PmsController::GetResult(quint32 *chA, quint32 *chB)
{
    quint32 coll = 0;
    coll = 2e5 + 1e5*sin(__x)*cos(__y);

    if (chA)
        *chA = coll;

    if (chB)
        *chB = coll + 1e3;
}

double PmsController::GetParam(short /*par_id*/)
{
    return 0;
}
void PmsController::SetParam(short par_id, float val)
{
    if (par_id == MACRO_TIME)
        __delay = val;
    if (par_id == GATE_LEVEL_A)
        __x = val;
    if (par_id == GATE_LEVEL_B)
        __y = val;
    if (par_id == INP_HOLDOFF) {
        __x+=0.1;
        __y+=0.1;
    }

}
short PmsController::Deinit()
{
    return 0;
}

#endif //USE_PMS_EMULATOR
