#include "pmsplot.h"
#include "flowlayout.hpp"
#include "mainwindow.h"
#include "pms400a.h"
#include "scanplot.h"
#include "ccd/ccdmatrix.h"
#include "ccd/i_spectr.hpp"
#include <QApplication>
#include <QFormLayout>
#include <QSettings>


#undef _X
#undef _Y
class IPMSScan : public IExternalScan
{
    Q_OBJECT

public slots:
    void completed() {
        if (!started)
            return;

        quint32 chA, chB;
        pmsControl->GetResult(&chA, &chB);

        quint32 value = chA;
        if (CBChannel && CBChannel->currentIndex() == 1)
            value = chB;

        double xx = sx[0] + (sx[1] - sx[0]) * _X / nx;
        if (dir == ccd::TInfoForSetOfSpectrums::Vert)
            xx = sy[0] + (sy[1] - sy[0]) * _Y / ny;

        Output3d.at(0)->Array[_Y][_X] = value;
        Section.at(0)->AddPoint(xx, value);

        if (dir == ccd::TInfoForSetOfSpectrums::Horiz) {
            switch (start_point) {
            case ccd::TInfoForSetOfSpectrums::TopLeft:
            case ccd::TInfoForSetOfSpectrums::BottomLeft:
                _X++;
                if (_X >= nx) {
                    _X = 0;
                    _Y += (start_point == ccd::TInfoForSetOfSpectrums::BottomLeft) ? 1 : -1;
                    Section.at(0)->clear();
                }
                break;
            case ccd::TInfoForSetOfSpectrums::TopRight:
            case ccd::TInfoForSetOfSpectrums::BottomRight:
                _X--;
                if (_X < 0) {
                    _X = nx - 1;
                    _Y += (start_point == ccd::TInfoForSetOfSpectrums::BottomRight) ? 1 : -1;
                    Section.at(0)->clear();
                }
                break;
            }
        } else {
            switch (start_point) {
            case ccd::TInfoForSetOfSpectrums::TopLeft:
            case ccd::TInfoForSetOfSpectrums::TopRight:
                _Y--;
                if (_Y < 0) {
                    _Y = ny - 1;
                    _X += (start_point == ccd::TInfoForSetOfSpectrums::TopLeft) ? 1 : -1;
                    Section.at(0)->clear();
                }
                break;
            case ccd::TInfoForSetOfSpectrums::BottomRight:
            case ccd::TInfoForSetOfSpectrums::BottomLeft:
                _Y++;
                if (_Y >= ny) {
                    _Y = 0;
                    _X += (start_point == ccd::TInfoForSetOfSpectrums::BottomLeft) ? 1 : -1;
                    Section.at(0)->clear();
                }
                break;
            }
        }

        int x0, x1, y0, y1;
        GetMathLimits(x0, x1, y0, y1);
        Output3d.at(0)->LimsForMathX[0] = x0;
        Output3d.at(0)->LimsForMathX[1] = x1;
        Output3d.at(0)->LimsForMathY[0] = y0;
        Output3d.at(0)->LimsForMathY[1] = y1;

        emit acquired();
    }

public:
    bool init(int _nx, int _ny, int _dir_stpoint, Range _sx, Range _sy) {
        IExternalScan::init(_nx, _ny, _dir_stpoint, _sx, _sy);

        _X = 0;
        _Y = 0;
        switch (start_point) {
        case ccd::TInfoForSetOfSpectrums::TopLeft:
            _Y = ny-1;
            break;
        case ccd::TInfoForSetOfSpectrums::TopRight:
            _X = nx-1;
            _Y = ny-1;
            break;
        case ccd::TInfoForSetOfSpectrums::BottomRight:
            _X = nx-1;
            break;
        }

        attributes.insert("XLabel", QString());
        attributes.insert("YLabel", tr("Counts"));
        attributes.insert("XUnit", tr("m"));
        attributes.insert("YUnit", QString());
        attributes.insert("Name", tr("PMS scan profile"));
        attributes.insert("PMS channel", CBChannel->currentText());

        return true;
    }
    void finish() {
        pmsControl->StopScan();

        IExternalScan::finish();
    }

    QWidget *createInterfaceWidget() {
        QGroupBox *wid = new QGroupBox(tr("PMS"));
        QFormLayout *lay = makeFormLayout(wid);

        CBChannel = new QComboBox;
        disableWidgets.append(CBChannel);
        CBChannel->addItem("A");
        CBChannel->addItem("B");

        lay->addRow(tr("Channel"), CBChannel);

        return wid;
    }

    bool acquireNextPoint() {
#ifdef USE_PMS_EMULATOR
        double xx = sx[0] + (sx[1] - sx[0]) * _X / nx;
        double yy = sy[0] + (sy[1] - sy[0]) * _Y / ny;

        pmsControl->SetParam(GATE_LEVEL_A, 1e6*xx);
        pmsControl->SetParam(GATE_LEVEL_B, 1e6*yy);
#endif
        if (pmsControl->StartScan() != 0) {
            pmsControl->StopScan();
            lastError = tr("PMS scan: can't start.");
            return false;
        }

        return true;
    }

    void GetMathLimits(int &x0, int &x1, int &y0, int &y1) {
        x0 = y0 = x1 = y1 = 0;

        int curLine = _Y;
        if (dir == ccd::TInfoForSetOfSpectrums::Vert)
            curLine = _X;

        switch(dir_stpoint) {
        case spTopLeftAxisX:
        case spTopRightAxisX:
            x1 = nx;
            y0 = curLine;
            y1 = ny;
            break;
        case spBottomRightAxisX:
        case spBottomLeftAxisX:
            x1 = nx;
            y1 = curLine+1;
            break;
        case spTopLeftAxisY:
        case spBottomLeftAxisY:
            x1 = curLine+1;
            y1 = ny;
            break;
        case spBottomRightAxisY:
        case spTopRightAxisY:
            x0 = curLine;
            x1 = nx;
            y1 = ny;
            break;
        }
    }

    int _X, _Y;

    PmsController *pmsControl;
    QPointer<QComboBox> CBChannel;
};
#include <pmsplot.moc>

IExternalScan *getNewPMSIExtScan(PMSPlot *pms)
{
    IPMSScan *sc = new IPMSScan;
    sc->pmsControl = pms->pmsControl;
    QObject::connect(sc->pmsControl, SIGNAL(scan_finished()), sc, SLOT(completed()), Qt::QueuedConnection);
    sc->disableWidgets << pms->BStartStop << pms->zPos <<
                          pms->GBOptions << pms->GBOptions2;
    return sc;
}

void PMSPlot::sl_set_X_scale(int sc)
{
    if (!curveCalib.contains("Laser Wavelength, nm") ||
            !curveCalib.contains("Central Wavelength, nm") ||
            !curveCalib.contains("Dispersion") ||
            !curveCalib.contains("pixelsize") ||
            !curveCalib.contains("Nx") ||
            !curveCalib.contains("Central Pixel"))
        sc = 0;

    switch (sc) {
    case 0:
        MyPl->Pure2dPlot->XAxisUnit.SetDescAndSIUnit(QObject::tr("pixels"));
        MyPl->Pure2dPlot->SetAllXScales(0,10000);
        break;
    case 1:
        MyPl->Pure2dPlot->XAxisUnit.SetDescAndSIUnit(QObject::tr("WL"), QObject::tr("m"));
        MyPl->Pure2dPlot->SetAllXScales(0,10000);
        break;
    case 2:
        MyPl->Pure2dPlot->XAxisUnit.SetDescAndSIUnit(QObject::tr("Raman Shift, 1/cm"));
        MyPl->Pure2dPlot->SetAllXScales(-10000,10000);
        break;
    }

    for (int i = 0; i < curveData.size(); i++)
        set_spectra_calib(*curveData[i], curveCalib, sc);

    MyPl->Refresh();
}

void PMSPlot::setDeviceStatus(int idx)
{
    if (!deviceStatus)
        return;

    setEnabled(idx >= EDok);
    switch (idx) {
    case EDno:
        deviceStatus->status->set(StatusWidget::Cross, Qt::red);
        deviceStatus->status->setToolTip(tr("No device"));
        break;
    case EDok:
        deviceStatus->status->set(StatusWidget::Circle, Qt::green);
        deviceStatus->status->setToolTip(tr("Connected"));
        break;
    case EDplay:
        deviceStatus->status->set(StatusWidget::Play, Qt::green);
        deviceStatus->status->setToolTip(tr("Running"));
        break;
    }
}

void PMSPlot::StartStop(bool st)
{
    if (st) {
        _start = true;
        _time = 0;
        GBspec->setEnabled(false);
        GBOptions->setEnabled(false);
        GBOptions2->setEnabled(false);

        curveData.at(0)->clear();
        curveData.at(1)->clear();
        if (mode == 3) {
            if (Ewmin->value() >= Ewmax->value())
                Ewmax->setValue(Ewmin->value() + 10);
            _time = 0;
            _currwl = Ewmin->value();
            _dwl = (Ewmax->value() - _currwl)/(CBpoints->currentText().toInt() - 1);
            double pxsz = (Ewmax->value() - _currwl)/(CBpoints->currentText().toInt());

            curveCalib.insert("Laser Wavelength, nm", ccdmatrix->Laser_WL);
            curveCalib.insert("Central Wavelength, nm", _currwl + pxsz*CBpoints->currentText().toInt()/2.0);
            curveCalib.insert("Dispersion", 1.0);
            curveCalib.insert("pixelsize", pxsz);
            curveCalib.insert("Nx", CBpoints->currentText().toInt());
            curveCalib.insert("Central Pixel", CBpoints->currentText().toInt()/2.0);

            I_SPECTR *spectr = ccdmatrix->m_spectrInterface;
            if (!ccdmatrix->testMode && spectr) {
                while (!spectr->isReady()) {
                    qApp->processEvents();
                    SleepExport::msleep2(2);
                }

                spectr->setWl(_currwl);

                while (!spectr->isReady()) {
                    qApp->processEvents();
                    SleepExport::msleep2(2);
                }
            }
        } else if (mode == 1) {
            curveData.at(0)->setXCalib(0, 0.5);
            curveData.at(1)->setXCalib(0, 0.5);
            MyPl->Pure2dPlot->SetAllXScales(0, _xtime);
            MyPl->Refresh();
        }

        if (pmsControl->StartScan() != 0) {
            new Notification(tr("PMS Error!") + ": " + tr("Can't start PMS scan"), Notification::Bad, 10);
            doBeep();
            _start = false;
            pmsControl->StopScan();
            BStartStop->change_state_man_with_sig(false);
        }
    } else {
        _start = false;
        pmsControl->StopScan();
        BStartStop->change_state_man(false);
        GBspec->setEnabled(true);
        GBOptions->setEnabled(true);
        GBOptions2->setEnabled(true);
    }
}

void PMSPlot::ScanStop()
{
    if (!_start)
        return;

    quint32 chA, chB;
    pmsControl->GetResult(&chA, &chB);

    if (mode == 3) {
        double _xx = _time;
        switch (CBunits->currentIndex()) {
        case 0:
            break;
        case 1:
            _xx = 1e-9*wavelength(curveCalib.value("Central Wavelength, nm").toDouble(),
                                  curveCalib.value("Dispersion").toDouble(),
                                  curveCalib.value("pixelsize").toDouble(),
                                  _time,
                                  curveCalib.value("Central Pixel").toInt());
            break;
        case 2:
            _xx = raman_shift(curveCalib.value("Laser Wavelength, nm").toDouble(),
                              wavelength(curveCalib.value("Central Wavelength, nm").toDouble(),
                                         curveCalib.value("Dispersion").toDouble(),
                                         curveCalib.value("pixelsize").toDouble(),
                                         _time,
                                         curveCalib.value("Central Pixel").toInt()));
            break;
        }

        curveData.at(0)->AddPoint(_xx, chA);
        curveData.at(1)->AddPoint(_xx, chB);
        _time += 1;

        MyPl->Refresh();

#ifdef USE_PMS_EMULATOR
        pmsControl->SetParam(INP_HOLDOFF, 0);
#endif

        if (_time >= CBpoints->currentText().toInt()) {
            BStartStop->change_state_man_with_sig(false);
            return;
        }

        _currwl += _dwl;
        I_SPECTR *spectr = ccdmatrix->m_spectrInterface;
        if (!ccdmatrix->testMode && spectr) {
            while (!spectr->isReady()) {
                qApp->processEvents();
                SleepExport::msleep2(2);
            }

            spectr->setWl(_currwl);

            while (!spectr->isReady()) {
                qApp->processEvents();
                SleepExport::msleep2(2);
            }
        }
    }
    if (mode == 1) {
        //osc mode
        if (_time >= _xtime) {
            _time = 0;
            curveData.at(0)->YSeq.clear();
            curveData.at(1)->YSeq.clear();
        }
        _time += 0.5;

        curveData.at(0)->AddY(chA);
        curveData.at(1)->AddY(chB);

        avgA->setText(QString::number(avg(curveData[0]->YSeq), 'f', 4));
        rmsA->setText(QString::number(rms(curveData[0]->YSeq), 'f', 4));

        avgB->setText(QString::number(avg(curveData[1]->YSeq), 'f', 4));
        rmsA->setText(QString::number(rms(curveData[1]->YSeq), 'f', 4));

        MyPl->Pure2dPlot->RefreshInc();

#ifdef USE_PMS_EMULATOR
        pmsControl->SetParam(INP_HOLDOFF, 0);
#endif
    }

    if (pmsControl->StartScan() != 0) {
        new Notification(tr("PMS Error!") + ": " + tr("Can't start PMS scan"), Notification::Bad, 10);
        doBeep();
        _start = false;
        pmsControl->StopScan();
        BStartStop->change_state_man_with_sig(false);
    }
}

void PMSPlot::setTime(int time)
{
    _xtime = time;
    if (mode == 1) {
        MyPl->Pure2dPlot->SetAllXScales(0, time);
        MyPl->Refresh();
    }
}
void PMSPlot::setMacroTime(double time)
{
    pmsControl->SetParam(MACRO_TIME, time);
    Emacrotime->setValue(pmsControl->GetParam(MACRO_TIME));
}
void PMSPlot::setCollectTime(double time)
{
    pmsControl->SetParam(COLLECT_TIME, time);
    Ecollecttime->setValue(pmsControl->GetParam(COLLECT_TIME));
}
void PMSPlot::setHoldoff(double hoff)
{
    pmsControl->SetParam(INP_HOLDOFF, hoff);
    Eholdoff->setValue(pmsControl->GetParam(INP_HOLDOFF));
}


void PMSPlot::setAlvl(double lvl)
{
    pmsControl->SetParam(INP_THRESHOLD_A, lvl);
    EAlvl->setValue(pmsControl->GetParam(INP_THRESHOLD_A));
}

void PMSPlot::setBlvl(double lvl)
{
    pmsControl->SetParam(INP_THRESHOLD_B, lvl);
    EBlvl->setValue(pmsControl->GetParam(INP_THRESHOLD_B));
}
void PMSPlot::setAgate(double lvl)
{
    pmsControl->SetParam(GATE_LEVEL_A, lvl);
    EAgate->setValue(pmsControl->GetParam(GATE_LEVEL_A));
}
void PMSPlot::setBgate(double lvl)
{
    pmsControl->SetParam(GATE_LEVEL_B, lvl);
    EBgate->setValue(pmsControl->GetParam(GATE_LEVEL_B));
}
void PMSPlot::setAevnt(double lvl)
{
    pmsControl->SetParam(EVENT_THRESHOLD_A, lvl);
    EAevnt->setValue(pmsControl->GetParam(EVENT_THRESHOLD_A));
}

void PMSPlot::setBevnt(double lvl)
{
    pmsControl->SetParam(EVENT_THRESHOLD_B, lvl);
    EBevnt->setValue(pmsControl->GetParam(EVENT_THRESHOLD_B));
}

void PMSPlot::on_MyPl_CurveColourChanged(QColor c, int AC)
{
    if ((unsigned int)AC == curveData.size() - 2)
        LEdColour[0]->SetColour(c);
    else if ((unsigned int)AC == curveData.size() - 1)
        LEdColour[1]->SetColour(c);
    else if (AC == -1) {
        LEdColour[0]->SetColour(c);
        LEdColour[1]->SetColour(c);
    }
}

void PMSPlot::AddData(plot::TMetaStorage *lst)
{
    lst->insert("Exposure time", Emacrotime->value());
    lst->insert("Ch A level", EAlvl->value());
    lst->insert("Ch B level", EBlvl->value());
    lst->insert("Ch A gate level", EAgate->value());
    lst->insert("Ch B gate level", EBgate->value());
    lst->insert("Objective Z", zPos->value());
}

void PMSPlot::AddDataToContainer_slot(plot::TCurvesDataOwner src, plot::TMetaStorage attributes)
{
    attributes.unite(curveCalib);

    attributes.insert("Name", MyPl->Pure2dPlot->Title);
    attributes.insert("Source", "PMS");

    if (mode == 1) {
        attributes.insert("chA Average", avgA->text());
        attributes.insert("chA RMS", rmsA->text());

        attributes.insert("chB Average", avgB->text());
        attributes.insert("chB RMS", rmsB->text());

        attributes.insert("T, s", _xtime);
    }

    AddData(&attributes);

    emit AddDataToContainer(src, attributes);
}

PMSPlot::PMSPlot(CCDMatrix *ccd, QWidget *parent) : QWidget(parent)
{
    ccdmatrix = ccd;

    deviceStatus = new StatusWidget1(tr("PMS"));
#ifdef USE_PMS_EMULATOR
    deviceStatus->setEmulator(true);
#endif
    setDeviceStatus(EDno);

    pmsControl = new PmsController();
    connect(pmsControl, SIGNAL(status(int)), this, SLOT(setDeviceStatus(int)), Qt::QueuedConnection);
    int ret = pmsControl->Init();
    setDeviceStatus(ret == 0 ? EDok : EDno);

    _time = 0;
    _xtime = 30;
    _start = false;

    mode = 1;//osc

    QSettings settings("Param/pms.conf", QSettings::IniFormat);

    _xtime = settings.value("osc_time", 30).toDouble();
    pmsControl->SetParam(MACRO_TIME, settings.value("macro_time", 0.5).toDouble());
    pmsControl->SetParam(COLLECT_TIME, settings.value("collect_time", 0.1).toDouble());
    pmsControl->SetParam(INP_HOLDOFF, settings.value("input_holdoff", 1).toDouble());
    pmsControl->SetParam(EVENT_THRESHOLD_A, settings.value("channelA/threshold", 1).toDouble());
    pmsControl->SetParam(GATE_LEVEL_A, settings.value("channelA/gate_level", 0.5).toDouble());
    pmsControl->SetParam(INP_THRESHOLD_A, settings.value("channelA/level", 0.05).toDouble());
    pmsControl->SetParam(EVENT_THRESHOLD_B, settings.value("channelB/threshold", 1).toDouble());
    pmsControl->SetParam(GATE_LEVEL_B, settings.value("channelB/gate_level", 0.5).toDouble());
    pmsControl->SetParam(INP_THRESHOLD_B, settings.value("channelB/level", -0.05).toDouble());

    MyPl = new plot::T2dPlot(this, false, this);
    MyPl->layout()->setContentsMargins(0, 0, 6, 0);

    MyPl->Pure2dPlot->XAxisUnit.SetDescAndSIUnit(tr("Seconds"));
    MyPl->Pure2dPlot->YAxisUnit.SetDescAndSIUnit(tr("Counts"));
    MyPl->Pure2dPlot->SetAllXScales(0, _xtime);
    MyPl->Pure2dPlot->SetAllYScales(0, 4e10);
    MyPl->Pure2dPlot->SetTitle(tr("Pms oscilloscope mode"));
    connect(MyPl, SIGNAL(AddDataToContainer(plot::TCurvesDataOwner,plot::TMetaStorage)), SLOT(AddDataToContainer_slot(plot::TCurvesDataOwner,plot::TMetaStorage)));

    plot::ResizeOwner(curveData, 2);
    curveData.at(0)->Pen.setColor(Qt::red);
    curveData.at(1)->Pen.setColor(Qt::blue);

    MyPl->Pure2dPlot->SourceData = plot::getUnLock(curveData);

    PBToNewWindow =  new TwoStateButton(":/CCD/Detach.png",":/CCD/Attach.png");
    //connect(PBToNewWindow, SIGNAL(clicked()), this, SLOT(map()));

    BStartStop = new TwoStateButton(":start.png",":stop.png");
    connect(BStartStop, SIGNAL(state_changed(bool)), SLOT(StartStop(bool)));
    connect(pmsControl, SIGNAL(scan_finished()), SLOT(ScanStop()), Qt::QueuedConnection);

    BOpt2 = new TwoStateButton(":/SPMTuning_window/tools.png",":/SPMTuning_window/tools.png");

    QHBoxLayout *layoutcc = new QHBoxLayout();
    layoutcc->setSpacing(6);
    layoutcc->setAlignment(Qt::AlignLeft);
    //layoutcc->addWidget(PBToNewWindow);
    layoutcc->addWidget(BStartStop);
    layoutcc->addWidget(BOpt2);
    layoutcc->setContentsMargins(0, 0, 0, 0);

    QWidget *CControl = new QWidget();
    CControl->setLayout(layoutcc);

    Etime = new Edit; Etime->setImmediateMode();
    Etime->setValue(_xtime);
    connect(Etime, SIGNAL(newValue(int)), SLOT(setTime(int)));

    Emacrotime = new Edit(0, 600, 3);
    Emacrotime->setValue(pmsControl->GetParam(MACRO_TIME));
    connect(Emacrotime, SIGNAL(newValue(double)), SLOT(setMacroTime(double)));

    Ecollecttime = new Edit(0.00001, 10, 5);
    Ecollecttime->setValue(pmsControl->GetParam(COLLECT_TIME));
    connect(Ecollecttime, SIGNAL(newValue(double)), SLOT(setCollectTime(double)));

    Eholdoff = new Edit(1, 15, 1);
    Eholdoff->setValue(pmsControl->GetParam(INP_HOLDOFF));
    connect(Eholdoff, SIGNAL(newValue(double)), SLOT(setHoldoff(double)));

    EAlvl = new Edit(-1, 1, 3);
    EAlvl->setValue(pmsControl->GetParam(INP_THRESHOLD_A));
    connect(EAlvl, SIGNAL(newValue(double)), SLOT(setAlvl(double)));

    EBlvl = new Edit(-1, 1, 3);
    EBlvl->setValue(pmsControl->GetParam(INP_THRESHOLD_B));
    connect(EBlvl, SIGNAL(newValue(double)), SLOT(setBlvl(double)));

    EAgate = new Edit(-1, 1, 3);
    EAgate->setValue(pmsControl->GetParam(GATE_LEVEL_A));
    connect(EAgate, SIGNAL(newValue(double)), SLOT(setAgate(double)));

    EBgate = new Edit(-1, 1, 3);
    EBgate->setValue(pmsControl->GetParam(GATE_LEVEL_B));
    connect(EBgate, SIGNAL(newValue(double)), SLOT(setBgate(double)));

    EAevnt = new Edit(0, 30000);
    EAevnt->setValue(pmsControl->GetParam(EVENT_THRESHOLD_A));
    connect(EAevnt, SIGNAL(newValue(double)), SLOT(setAevnt(double)));

    EBevnt = new Edit(0, 30000);
    EBevnt->setValue(pmsControl->GetParam(EVENT_THRESHOLD_B));
    connect(EBevnt, SIGNAL(newValue(double)), SLOT(setBevnt(double)));

    GBOptions = new QGroupBox(tr("PMS Options"));
    QFormLayout *opt_ll = makeFormLayout(GBOptions);
    opt_ll->addRow(tr("Exposure time, s"), Emacrotime);
    opt_ll->addRow(tr("Ch A thresold level, V"), EAlvl);
    opt_ll->addRow(tr("Ch B thresold level, V"), EBlvl);

    GBOptions2 = new QGroupBox(tr("PMS Extra Options"));
    GBOptions2->setVisible(false);
    connect(BOpt2, SIGNAL(state_changed(bool)), GBOptions2, SLOT(setVisible(bool)));
    QFormLayout *opt_ll2 = makeFormLayout(GBOptions2);
    opt_ll2->addRow(tr("Ch A gate level, V"), EAgate);
    opt_ll2->addRow(tr("Ch B gate level, V"), EBgate);
    opt_ll2->addRow(tr("Ch A event thresold"), EAevnt);
    opt_ll2->addRow(tr("Ch B event thresold"), EBevnt);
    opt_ll2->addRow(tr("Collect time, s"), Ecollecttime);
    opt_ll2->addRow(tr("Input holdoff, 1..15\n(10ns/[1..15])"), Eholdoff);


    avgA = new QLineEdit();
    avgB = new QLineEdit();
    rmsA = new QLineEdit();
    rmsB = new QLineEdit();
    avgA->setReadOnly(true);
    avgB->setReadOnly(true);
    rmsA->setReadOnly(true);
    rmsB->setReadOnly(true);

    LEdColour[0] = new TRectWidget(Qt::red);
    LEdColour[1] = new TRectWidget(Qt::blue);
    connect(MyPl, SIGNAL(CurveColourChanged(QColor, int)), this, SLOT(on_MyPl_CurveColourChanged(QColor, int)));

    QGroupBox *GBchA = new QGroupBox(tr("Channel A"));
    QFormLayout *opt_A = makeFormLayout(GBchA);
    opt_A->addRow(tr("Average"), avgA);
    opt_A->addRow(tr("RMS"), rmsA);
    opt_A->addRow(LEdColour[0]);

    QGroupBox *GBchB = new QGroupBox(tr("Channel B"));
    QFormLayout *opt_B = makeFormLayout(GBchB);
    opt_B->addRow(tr("Average"), avgB);
    opt_B->addRow(tr("RMS"), rmsB);
    opt_B->addRow(LEdColour[1]);


    GBosc = new QGroupBox(tr("Oscilloscope mode"));
    QFormLayout *layout_osc = makeFormLayout(GBosc);
    layout_osc->addRow(tr("Period, s"), Etime);
    layout_osc->addRow(GBchA);
    layout_osc->addRow(GBchB);


    Ewmin = new Edit(200, 800);
    Ewmin->setValue(350);
    Ewmax = new Edit(200, 800);
    Ewmax->setValue(550);

    CBpoints = new QComboBox;
    CBpoints->addItem("10");
    CBpoints->addItem("25");
    CBpoints->addItem("50");
    CBpoints->addItem("100");
    CBpoints->addItem("250");
    CBpoints->addItem("500");
    CBpoints->addItem("1000");

    CBunits = new QComboBox;
    CBunits->addItem(tr("Pixels"));
    CBunits->addItem(tr("Wavelength"));
    CBunits->addItem(tr("Raman shift"));
    connect(CBunits, SIGNAL(currentIndexChanged(int)), SLOT(sl_set_X_scale(int)));


    GBspec = new QGroupBox(tr("Spectrum mode"));
    GBspec->hide();
    QFormLayout *layout_spec = makeFormLayout(GBspec);
    layout_spec->addRow(tr("From, nm"), Ewmin);
    layout_spec->addRow(tr("To, nm"), Ewmax);
    layout_spec->addRow(tr("Units"), CBunits);
    layout_spec->addRow(tr("Points"), CBpoints);


    //---------------xy
    setXYscanMode = new QComboBox;
    setXYscanMode->addItem(tr("Oscilloscope mode"));
    setXYscanMode->addItem(tr("Spectrum mode"));
    setXYscanMode->setSizePolicy(QSizePolicy::Fixed, QSizePolicy::Fixed);
    layoutcc->addWidget(setXYscanMode);
    connect(setXYscanMode, SIGNAL(currentIndexChanged(int)), SLOT(ScanTypeChanged(int)));

    zPos = new DoubleChanger(0, 10, 4);

    QHBoxLayout *layout_Zpos = new QHBoxLayout;
    layout_Zpos->addWidget(zPos);

    QGroupBox *box_Zpos = new QGroupBox(tr("Objective Z, ")+tr("micro")+tr("m"));
    box_Zpos->setLayout(layout_Zpos);

    QVBoxLayout *layout_Options = new QVBoxLayout;
    layout_Options->addWidget(CControl);
    layout_Options->addWidget(GBosc);
    layout_Options->addWidget(GBspec);
    layout_Options->addWidget(GBOptions);
    layout_Options->addWidget(GBOptions2);
    layout_Options->addWidget(box_Zpos);
    layout_Options->setAlignment(Qt::AlignTop);
    layout_Options->setSpacing(3);

    QHBoxLayout *MainLayout = new QHBoxLayout(this);
    MainLayout->setContentsMargins(6, 0, 0, 0);
    MainLayout->addLayout(layout_Options);
    MainLayout->addWidget(MyPl, 1);
}

PMSPlot::~PMSPlot()
{
    QSettings settings("Param/pms.conf", QSettings::IniFormat);

    settings.setValue("osc_time", _xtime);
    settings.setValue("macro_time", pmsControl->GetParam(MACRO_TIME));
    settings.setValue("collect_time", pmsControl->GetParam(COLLECT_TIME));
    settings.setValue("input_holdoff", pmsControl->GetParam(INP_HOLDOFF));
    settings.setValue("channelA/threshold", pmsControl->GetParam(EVENT_THRESHOLD_A));
    settings.setValue("channelA/gate_level", pmsControl->GetParam(GATE_LEVEL_A));
    settings.setValue("channelA/level", pmsControl->GetParam(INP_THRESHOLD_A));
    settings.setValue("channelB/threshold", pmsControl->GetParam(EVENT_THRESHOLD_B));
    settings.setValue("channelB/gate_level", pmsControl->GetParam(GATE_LEVEL_B));
    settings.setValue("channelB/level", pmsControl->GetParam(INP_THRESHOLD_B));

    pmsControl->Deinit();
    delete pmsControl;
}

void PMSPlot::ScanTypeChanged(int St)
{
    if (St == 1) {
        GBosc->hide();
        GBspec->show();
    } else {
        GBosc->show();
        GBspec->hide();
    }

    if (St == 1) {
        mode = 3;

        switch (CBunits->currentIndex()) {
        case 0:
            MyPl->Pure2dPlot->XAxisUnit.SetDescAndSIUnit(QObject::tr("pixels"));
            MyPl->Pure2dPlot->SetAllXScales(0,10000);
            break;
        case 1:
            MyPl->Pure2dPlot->XAxisUnit.SetDescAndSIUnit(QObject::tr("WL"), QObject::tr("m"));
            MyPl->Pure2dPlot->SetAllXScales(0,10000);
            break;
        case 2:
            MyPl->Pure2dPlot->XAxisUnit.SetDescAndSIUnit(QObject::tr("Raman Shift, 1/cm"));
            MyPl->Pure2dPlot->SetAllXScales(-10000,10000);
            break;
        }
    } else {
        mode = 1;

        MyPl->Pure2dPlot->XAxisUnit.SetDescAndSIUnit(tr("Seconds"));
    }
}

