#include "modules.hpp"
#include <QDebug>

class EG3000Module;
class EG3030Module;
class DataModule;
class CCDMatrixModule;
class CCDPlotModule;
class ScanModule;
class SPMModule;
class MotorsModule;
class OscilloscopeModule;
class VideoModule;
class LithoModule;


ModuleManager *ModuleManager::instance = 0;

I_MODULE::~I_MODULE()
{
    if (ModuleManager::instance)
        ModuleManager::instance->unload(this);
}

ModuleManager::ModuleManager()
{
    if (instance)
        throw instance;
    instance = this;
    isRunning = false;

    ModuleManager::instance->addToFactory<EG3000Module>();
    ModuleManager::instance->addToFactory<EG3030Module>();
    ModuleManager::instance->addToFactoryGui<DataModule>();
    ModuleManager::instance->addToFactory<CCDMatrixModule>();
    ModuleManager::instance->addToFactoryGui<CCDPlotModule>();
    ModuleManager::instance->addToFactoryGui<ScanModule>();
    ModuleManager::instance->addToFactoryGui<SPMModule>();
    ModuleManager::instance->addToFactoryGui<MotorsModule>();
    ModuleManager::instance->addToFactoryGui<OscilloscopeModule>();
    ModuleManager::instance->addToFactoryGui<VideoModule>();
    ModuleManager::instance->addToFactoryGui<LithoModule>();
}

ModuleManager::~ModuleManager()
{
    instance = 0;

    if (isRunning)
        foreach (I_MODULE *m, modules)
            m->setRunning(false);

    foreach (I_MODULE *m, modules)
        delete m;

    modules.clear();
}

I_MODULE *ModuleManager::load(const QLatin1String name, QWidget *parent)
{
    if (!factory.contains(name))
        return 0;

    QList<QLatin1String> l;
    if (factory[name].requires)
        l = factory[name].requires();
    for (int i = 0; i < l.size(); i++)
        if (!modules.contains(l.at(i)))
            if (!load(l.at(i), parent))
                qCritical() << "can't create module" << QString(l.at(i)) << "required for" << QString(name);

    I_MODULE *m;
    if (factory[name].creategui)
        m = factory[name].creategui(parent);
    else
        m = factory[name].create();
    add(name, m);
    return m;
}

void ModuleManager::add(const QLatin1String name, I_MODULE *module)
{
    QMapIterator<QLatin1String, I_MODULE *> i(modules);
    while (i.hasNext()) {
        i.next();
        i.value()->connected(name, module);
        module->connected(i.key(), i.value());
    }

    if (isRunning)
        module->setRunning(true); //TODO fix order for dependencies

    modules.insert(name, module);
}

I_MODULE *ModuleManager::find1(const QLatin1String name) const
{
    return modules.value(name);
}
const QList<I_MODULE *> ModuleManager::find(const QLatin1String name) const
{
    return modules.values(name);
}

const QLatin1String ModuleManager::find(I_MODULE *module) const
{
    foreach (QLatin1String name, modules.keys(module))
        return name;

    return QLatin1String(0);
}

void ModuleManager::unload(const QLatin1String name)
{
    QList<I_MODULE *> l = modules.values(name);
    if (isRunning)
        foreach (I_MODULE *m0, l)
            m0->setRunning(false);

    foreach (I_MODULE *m0, l)
        foreach (I_MODULE *m, modules)
            if (m != m0)
                m->disconnected(name, m0);

    modules.remove(name);

    foreach (I_MODULE *m0, l)
        delete m0;
}

void ModuleManager::unload(I_MODULE *module)
{
    QLatin1String name = find(module);
    if (!name.latin1())
        return;

    if (isRunning)
        module->setRunning(false);

    foreach (I_MODULE *m, modules)
        if (m != module)
            m->disconnected(name, module);

    modules.remove(name, module);
}

void ModuleManager::setRunning(bool running)
{
    if (isRunning == running)
        return;

    isRunning = running;

    foreach (I_MODULE *m, modules)
        m->setRunning(isRunning);
}

//---------------------------------------------------------------------------------------
#include <QSettings>
#include "mainwindow.h"
#include "mobile/remote_mobile.hpp"

#include "EG3000.h"
USB_WRITE_THREAD *_wThread = 0;
class EG3000Module : public USB_THREAD_GROUP, public I_MODULE
{
public:
    EG3000Module() {
        _wThread = &wThread;

        QSettings settings("Param/Emulator.conf", QSettings::IniFormat);
        bool USB_Emulator = settings.value("usb", false).toBool();

        cThread.SetTestMode(USB_Emulator);
        wThread.SetTestMode(USB_Emulator);
        rThread.SetTestMode(USB_Emulator);
        if (USB_Emulator) {
            QObject::connect(&wThread, SIGNAL(sig_osc_test(bool)), &rThread, SLOT(sl_set_osc_test(bool)));
            QObject::connect(&wThread, SIGNAL(sig_lock_test()), &rThread, SLOT(sl_set_lockin_test()));
            QObject::connect(&wThread, SIGNAL(sig_Spectr_test()), &rThread, SLOT(sl_set_Spectr_test()));
            QObject::connect(&wThread, SIGNAL(sig_Lit_test()), &rThread, SLOT(sl_set_Litho_test()));
            QObject::connect(&wThread, SIGNAL(sig_scan_test(bool)), &rThread, SLOT(sl_set_scan_test(bool)));
            QObject::connect(&wThread, SIGNAL(sig_scan_test_ccd(bool)), &rThread, SLOT(set_scan_test_ccd(bool)));
            QObject::connect(&wThread, SIGNAL(sig_scan_test_ccd_ready()), &rThread, SLOT(sl_set_scan_test_ccd_ready()));
        }

        QObject::connect(&cThread, SIGNAL(updateUsbStatus(bool)), &rThread, SLOT(start_stop(bool)));
        QObject::connect(&cThread, SIGNAL(updateUsbStatus(bool)), &wThread, SLOT(start_stop(bool)));

        cThread.start();
    }
    static const QLatin1String name() { return QLatin1String("eg3000"); }
    void *basecast(const QLatin1String name) { return name.latin1() != 0 ? 0 : static_cast<USB_THREAD_GROUP *>(this); }
    void setRunning(bool on) {
        if (on)
            cThread.m_waitMainInit = false;
        else
            cThread.thread_close(&wThread, &rThread);
    }
};

#include "hardware/control3030.hpp"
Control3030 *_ControllerStub3030 = 0;
class EG3030Module : public Control3030, public I_MODULE
{
public:
    EG3030Module() {
        _ControllerStub3030 = this;

        QSettings settings("Param/Emulator.conf", QSettings::IniFormat);
        SetTestMode(settings.value("usb", false).toBool());

        start();
    }
    static const QLatin1String name() { return QLatin1String("eg3030"); }
    void *basecast(const QLatin1String name) { return name.latin1() != 0 ? 0 : static_cast<EG3030Module *>(this); }
    void setRunning(bool on) {
        if (on)
            m_waitMainInit = false;
        else
            thread_close();
    }
};

#include "dataplot.h"
class DataModule : public I_GUI_MODULE<DataPlot>, public I_NET_MODULE_BASE
{
public:
    GUI_MODULE_CTOR(DataModule) {
        QSettings settings("Param/gui.conf", QSettings::IniFormat);
        Container->SetAutoSaveTime(settings.value("autosave_time", 30).toInt() * 60 * 1000); //to msec
    }
    static const QLatin1String name() { return QLatin1String("data"); }
    void *basecast(const QLatin1String name) { return name == QLatin1String("net") ? static_cast<I_NET_MODULE_BASE *>(this) : I_GUI_MODULE<DataPlot>::basecast(name); }
    void connected(const QLatin1String name, I_MODULE *module);
    void setRunning(bool on) { if (!on) Container->thread_close(); }

    //gui
    QPushButton *createPanelButton() { return new QPushButton(makeHugeIcon(":/MainToolbar/Data_plot.png"), MainWindow::tr("Data")); }

    //net
    void initialSend(MobileInstance *client);
    void setValue(int id, QVariant val, QObject *sender);
};

#include "ccd/ccdmatrix.h"
#include "ccd/i_spectr.hpp"
class CCDMatrixModule : public CCDMatrix, public I_MODULE
{
public:
    CCDMatrixModule() {
        QSettings settings("Param/Emulator.conf", QSettings::IniFormat);
        SetTestMode(settings.value("ccd", false).toBool());
        start();
    }
    static const QLatin1String name() { return QLatin1String("ccdmatrix"); }
    void *basecast(const QLatin1String name) { return name.latin1() != 0 ? 0 : static_cast<CCDMatrix *>(this); }
    void setRunning(bool on) {
        if (on)
            m_waitMainInit = false;
        else {
            thread_close();

            if (m_spectrInterface)
                m_spectrInterface->shutdown();
        }
    }
};

#include "ccd/ccdplot.h"
#include "hardware/pmsplot.h"
class CCDPlotModule : public I_GUI_MODULE<CCDPlot>
{
public:
    GUI_MODULE_CTOR(CCDPlotModule) {
        QSettings settings("Param/gui.conf", QSettings::IniFormat);
        AutoSend = settings.value("scan/autosend_data", true).toBool();
    }
    static const QLatin1String name() { return QLatin1String("ccdplot"); }
    static const QList<QLatin1String> requires() { return QList<QLatin1String>() << QLatin1String("ccdmatrix"); }
    void connected(const QLatin1String name, I_MODULE *module);
    void setRunning(bool on) { if (on) m_BStart->click(); }

    QPushButton *detachButton() { return PBToNewWindow; }
    QPushButton *createPanelButton() { return new QPushButton(makeHugeIcon(":/MainToolbar/ccd.png"), MainWindow::tr("Spectroscopy")); }
    QList<QWidget *> getStatusWidgets() { return QList<QWidget *>() << getDeviceStatusList(); }
};

#include "scanplot.h"
class ScanModule : public I_GUI_MODULE<ScanPlot>, public I_NET_MODULE_BASE
{
public:
    GUI_MODULE_CTOR(ScanModule) {
        QSettings settings("Param/gui.conf", QSettings::IniFormat);
        Autosend = settings.value("scan/autosend_data", true).toBool();
    }
    static const QLatin1String name() { return QLatin1String("scan"); }
    void *basecast(const QLatin1String name) { return name == QLatin1String("net") ? static_cast<I_NET_MODULE_BASE *>(this) : I_GUI_MODULE<ScanPlot>::basecast(name); }
    void connected(const QLatin1String name, I_MODULE *module);

    //gui
    QPushButton *detachButton() { return PBToNewWindow; }
    QPushButton *createPanelButton() { return new QPushButton(makeHugeIcon(":/MainToolbar/Scan.png"), MainWindow::tr("Scan")); }

    //net
    void initialSend(MobileInstance *client);
    void setValue(int id, QVariant val, QObject *sender);
};

#include "spmtunning.h"
#include "shuttle.h"
#include "hardware/temp_control.hpp"
class SPMModule : public I_GUI_MODULE<SPMTunning>, public I_NET_MODULE_BASE
{
public:
    GUI_MODULE_CTOR(SPMModule) {
        QSettings settings("Param/gui.conf", QSettings::IniFormat);
        AutoSend = settings.value("scan/autosend_data", true).toBool();
    }
    static const QLatin1String name() { return QLatin1String("spm"); }
    void *basecast(const QLatin1String name) { return name == QLatin1String("net") ? static_cast<I_NET_MODULE_BASE *>(this) : I_GUI_MODULE<SPMTunning>::basecast(name); }
    void connected(const QLatin1String name, I_MODULE *module);
    void setRunning(bool on) {
        QSettings settings("Param/gui.conf", QSettings::IniFormat);
        if (on) {
            BLaserOnOff->change_state_man_with_sig(
                        settings.value("state/spm/laser_on", true).toBool());

            bool head, stage;
            {
                QSettings settings("Param/gui.conf", QSettings::IniFormat);
                head = settings.value("display/oscilloscope/head", true).toBool();
                stage = settings.value("display/oscilloscope/stage", true).toBool();
            }
            int fb_en = 0;
            if (head && stage) {
                QSettings settings("Param/controller.conf", QSettings::IniFormat);
                fb_en = settings_double(settings, "limit_spm_z", 0);
                if (fb_en != 0 && fb_en != 1 && fb_en != 2) fb_en = 0;
            } else {
                if (!head) fb_en = 2;
                if (!stage) fb_en = 1;
            }
            if (fb_en != 0) CBFBScanner->setEnabled(false);
            if (fb_en == 2) CBFBScanner->setCurrentIndex(1);
            if (fb_en == 0) {
                QSettings settings("Param/gui.conf", QSettings::IniFormat);
                CBFBScanner->setCurrentIndex(settings.value("state/spm/fb_scanner", 0).toInt());
            }
        } else {
            settings.setValue("state/spm/laser_on", BLaserOnOff->on_off);
            settings.setValue("state/spm/fb_scanner", CBFBScanner->currentIndex());
        }
    }

    //gui
    QPushButton *detachButton() { return PBToNewWindow; }
    QPushButton *createPanelButton() { return new QPushButton(makeHugeIcon(":/MainToolbar/Spm.png"), MainWindow::tr("SPM")); }
    QList<QWidget *> getStatusWidgets() {
        QList<QWidget *> l;
        l << (QWidget *)0 << SWspmfb;
        QSettings settings("Param/gui.conf", QSettings::IniFormat);
        if (shuttle_gui && settings.value("display/shuttle_status", true).toBool())
            l << (QWidget *)1 << shuttle_gui->deviceStatus;
        if (tcontrol)
            l << (QWidget *)4 << tcontrol->deviceStatus;
        return l;
    }

    //net
    void initialSend(MobileInstance *client);
    void setValue(int id, QVariant val, QObject *sender);
};

#include "motors.hpp"
class MotorsModule : public I_GUI_MODULE<Motors>, public I_NET_MODULE_BASE
{
    Q_OBJECT
public slots:
    void on_Motors_clicked();
public:
    GUI_MODULE_CTOR(MotorsModule) { }
    static const QLatin1String name() { return QLatin1String("motors"); }
    void *basecast(const QLatin1String name) { return name == QLatin1String("net") ? static_cast<I_NET_MODULE_BASE *>(this) : I_GUI_MODULE<Motors>::basecast(name); }
    void connected(const QLatin1String name, I_MODULE *module);

    //gui
    /*QPushButton *createPanelButton() {
        QSettings settings("Param/gui.conf", QSettings::IniFormat);
        if (!settings.value("display/motors", false).toBool())
            return 0;

        return new QPushButton(makeHugeIcon(":/SPMTuning_window/Motors_leveling.png"), MainWindow::tr("Motors"));
    }*/
    QList<QWidget *> getStatusWidgets() { return QList<QWidget *>() << (QWidget *)0 << new QLabel(MainWindow::tr("Motors:")) << SWmotor1; }

    //net
    void setValue(int id, QVariant val, QObject *sender);
};
#include <modules.moc>

#include "oscilloscope.h"
#include "autocalib.hpp"
class OscilloscopeModule : public I_GUI_MODULE<Oscilloscope>
{
public:
    GUI_MODULE_CTOR(OscilloscopeModule) { }
    static const QLatin1String name() { return QLatin1String("oscilloscope"); }
    void connected(const QLatin1String name, I_MODULE *module);
    void setRunning(bool on) {
        if (second_3030)
            return;

        QSettings settings("Param/gui.conf", QSettings::IniFormat);
        if (on) {
            for (int i = 0; i < 4; i++)
                CBChannel[i]->setCurrentIndex(
                            qMax(0, CBChannel[i]->findData(
                                     settings.value("state/osc/ch" + QString::number(i), 0))));

            DacsWidget->DacsChannel->setCurrentIndex(
                        qMax(0, DacsWidget->DacsChannel->findData(
                                 settings.value("state/osc/ch_control", 0).toInt())));
        } else {
            for (int i = 0; i < 4; i++)
                settings.setValue("state/osc/ch" + QString::number(i), CBChannel[i]->itemData(CBChannel[i]->currentIndex()));

            settings.setValue("state/osc/ch_control", DacsWidget->DacsChannel->itemData(DacsWidget->DacsChannel->currentIndex()).toInt());
        }
    }

    QPushButton *detachButton() { return PBToNewWindow; }
    QPushButton *createPanelButton() { return new QPushButton(makeHugeIcon(":/MainToolbar/Osc.png"), MainWindow::tr("Oscilloscope") + (second_3030 ? " 2" : QString())); }
};

#include "USBVideo/videoplayer.h"
class VideoModule : public I_GUI_MODULE<VideoPlayer>, public I_NET_MODULE_BASE
{
public:
    GUI_MODULE_CTOR(VideoModule) { }
    static const QLatin1String name() { return QLatin1String("video"); }
    static const QList<QLatin1String> requires() { return QList<QLatin1String>() << QLatin1String("managedthreadedcleaner"); }
    void *basecast(const QLatin1String name) { return name == QLatin1String("net") ? static_cast<I_NET_MODULE_BASE *>(this) : I_GUI_MODULE<VideoPlayer>::basecast(name); }
    void connected(const QLatin1String name, I_MODULE *module);
    void setRunning(bool on) { if (on) resetCameras(); }

    //gui
    QPushButton *detachButton() { return PBToNewWindow; }
    QPushButton *createPanelButton() { return new QPushButton(makeHugeIcon(":/SPMTuning_window/Video.png"), MainWindow::tr("Video") + (inst_idx == 0 ? QString() : " 2")); }
    QList<QWidget *> getStatusWidgets() { return QList<QWidget *>() << (QWidget *)5 << deviceStatus; }

    //net
    void setValue(int id, QVariant val, QObject *sender);
};

#include "grapheditor/TGraphEditor.h"
#include <grapheditor/widgets/TCheckableButton.h>
class LithoModule : public I_GUI_MODULE<graphed::TGraphEditor>
{
public:
    GUI_MODULE_CTOR(LithoModule) { }
    static const QLatin1String name() { return QLatin1String("litho"); }
    void connected(const QLatin1String name, I_MODULE *module);

    QPushButton *detachButton() { return BToNewWindow(); }
    QPushButton *createPanelButton() { return new QPushButton(makeHugeIcon(":/MainToolbar/litho.png"), MainWindow::tr("Lithography")); }
};


//---------------------------------------------------------------------------------------
//guidelines:
//connect from user module to common service provider module
//connect from user module to hook/notify module
//be wary of possible order dependencies

void DataModule::connected(const QLatin1String name, I_MODULE *module)
{
    if (name == QLatin1String("oscilloscope")) {
        Oscilloscope *m = (Oscilloscope *)module->basecast();
        m->acp->dataplot = this;
    }
}

void CCDPlotModule::connected(const QLatin1String name, I_MODULE *module)
{
    if (name == QLatin1String("eg3000")) {
        USB_THREAD_GROUP *m = (USB_THREAD_GROUP *)module->basecast();

        connect(this, SIGNAL(XYpos(int, double, double)), &m->wThread, SLOT(Go_to_XYZ_CCD(int, double, double)));
    }
    if (name == QLatin1String("data")) {
        DataPlot *m = (DataPlot *)module->basecast();

        connect(this, SIGNAL(AddDataToContainer(plot::TCurvesDataOwner,plot::TMetaStorage)), m->Container, SLOT(AddDataToContainer(plot::TCurvesDataOwner,plot::TMetaStorage)));
        connect(this, SIGNAL(AddDataToContainer(plot::T3dDataOwner, plot::TMetaStorage)), m->Container, SLOT(AddDataToContainer(plot::T3dDataOwner, plot::TMetaStorage)));
        if (pms_plot)
            connect(pms_plot, SIGNAL(AddDataToContainer(plot::TCurvesDataOwner,plot::TMetaStorage)), m->Container, SLOT(AddDataToContainer(plot::TCurvesDataOwner,plot::TMetaStorage)));
    }
    if (name == QLatin1String("scan")) {
        ScanPlot *m = (ScanPlot *)module->basecast();

        IExternalScan *sc;
        m->addExternalScan(MainWindow::tr("CCD"), sc = getNewCCDIExtScan(this));
        if (DataModule *data = (DataModule *)ModuleManager::instance->find1("data"))
            connect(sc, SIGNAL(AddDataToContainer(ccd::TSetOfSpectrum*,plot::TMetaStorage)),
                    data->Container, SLOT(AddDataToContainer(ccd::TSetOfSpectrum*,plot::TMetaStorage)));
        if (pms_plot)
            m->addExternalScan(MainWindow::tr("PMS"), getNewPMSIExtScan(pms_plot));
    }
}

#include <grapheditor/widgets/canvas/TWView.h>
#include <QCheckBox>
#include <QTimer>
void ScanModule::connected(const QLatin1String name, I_MODULE *module)
{
    if (name == QLatin1String("eg3000")) {
        USB_THREAD_GROUP *m = (USB_THREAD_GROUP *)module->basecast();

        connect(&m->wThread, SIGNAL(started()), this, SLOT(reset()));

        connect(&m->rThread, SIGNAL(capBits(int,long)), this, SLOT(capBits(int,long)));
        connect(&m->rThread, SIGNAL(ScanLineReady(QByteArray)), this, SLOT(sl_update_progress(QByteArray)), Qt::QueuedConnection);
        connect(&m->rThread, SIGNAL(sigSPMStatus(double,double,double,double,double)), this, SLOT(slUpdZscanner(double)), Qt::QueuedConnection);

        connect(HVxyzToZero, SIGNAL(clicked()), &m->wThread, SLOT(HVXYZzero()));

        connect(this, SIGNAL(sig_XYZPosition(double, double, double)), &m->wThread, SLOT(Go_to_XYZ(double, double, double)));
        connect(this, SIGNAL(sig_XYZPosition2(double, double, double)), &m->wThread, SLOT(Go_to_XYZ2(double, double, double)));

        connect(&m->wThread, SIGNAL(xUpdate(double)), this, SLOT(sl_setCurrentX(double)));
        connect(&m->wThread, SIGNAL(yUpdate(double)), this, SLOT(sl_setCurrentY(double)));
        connect(&m->wThread, SIGNAL(x2Update(double)), this, SLOT(sl_setCurrentX2(double)));
        connect(&m->wThread, SIGNAL(y2Update(double)), this, SLOT(sl_setCurrentY2(double)));
        connect(&m->wThread, SIGNAL(zUpdate(double)), DCZPos, SLOT(setValue(double)));

        connect(this, SIGNAL(sync_scan_x(double)), &m->wThread, SIGNAL(xUpdate(double)));
        connect(this, SIGNAL(sync_scan_y(double)), &m->wThread, SIGNAL(yUpdate(double)));
        connect(this, SIGNAL(sync_scan_x2(double)), &m->wThread, SIGNAL(x2Update(double)));
        connect(this, SIGNAL(sync_scan_y2(double)), &m->wThread, SIGNAL(y2Update(double)));

        //scan params
        connect(this, SIGNAL(scan_start()), &m->wThread, SLOT(SCAN_Start()));
        connect(this, SIGNAL(scan_stop()), &m->wThread, SLOT(SCAN_Stop()));
        connect(CBScanner, SIGNAL(currentIndexChanged(int)), &m->wThread, SLOT(set_scanner(int)));
        connect(this, SIGNAL(scan_speed(double)), &m->wThread, SLOT(Set_Speed(double)));
        connect(LEd_LFmodeHeight, SIGNAL(newValue(double)), &m->wThread, SLOT(scan_lm_height(double)));

        connect(this, SIGNAL(sig_ScanCorrection(int)), &m->wThread, SLOT(ScanCorrection(int)));
        connect(BPreciseScan, SIGNAL(state_changed(bool)), &m->wThread, SLOT(scanPreciseMode(bool)));
        connect(spmPrecision, SIGNAL(newValue(double)), &m->wThread, SLOT(spm_pid_precision(double)));
        connect(firstDelay, SIGNAL(newValue(double)), &m->wThread, SLOT(scan_first_delay(double)));
        connect(lineDelay, SIGNAL(newValue(double)), &m->wThread, SLOT(scan_line_delay(double)));
        connect(pointDelay, SIGNAL(newValue(double)), &m->wThread, SLOT(scan_point_delay(double)));
        connect(fastReturn, SIGNAL(toggled(bool)), &m->wThread, SLOT(scan_fast_return(bool)));
        connect(pointSync, SIGNAL(currentIndexChanged(int)), &m->wThread, SLOT(scan_sync_mode(int)));

        connect(this, SIGNAL(sig_sendScanParam(unsigned short int, unsigned short int, double,
                                               double, double, double, double,
                                               unsigned char, unsigned char, unsigned char, unsigned char, unsigned char,
                                               bool, bool, bool, bool, bool, double)),
                &m->wThread, SLOT(Set_Scan_Parameters(unsigned short int, unsigned short int, double,
                                                      double, double, double,double,
                                                      unsigned char, unsigned char, unsigned char, unsigned char, unsigned char,
                                                      bool, bool, bool, bool, bool, double)));
        connect(this, SIGNAL(sig_sendScanParam(unsigned short int, unsigned short int, double,
                                               double, double, double, double,
                                               unsigned char, unsigned char, unsigned char, unsigned char, unsigned char,
                                               bool, bool, bool, bool, bool, double)),
                &m->rThread, SLOT(SetScanPoints(unsigned short int, unsigned short int, double,
                                                double, double, double, double,
                                                unsigned char, unsigned char, unsigned char, unsigned char, unsigned char,
                                                bool, bool, bool, bool, bool, double)));


        connect(this, SIGNAL(sig_ccd_scan_start()), &m->wThread, SLOT(CCD_SCAN_Start()));
        connect(this, SIGNAL(sig_ccd_scan_stop()), &m->wThread, SLOT(CCD_SCAN_Stop()));
        connect(this, SIGNAL(sig_CCDready()), &m->wThread, SLOT(CCD_SCAN_Ready()));
        connect(&m->rThread, SIGNAL(CCD_Start()), this, SLOT(acquireNextPoint()), Qt::QueuedConnection);

        connect(this, SIGNAL(ccdScanParam(int, int, int, int, int, double, double, double, double, bool)),
                &m->wThread, SLOT(SetccdScanMode(int, int, int, int, int, double, double, double, double, bool)));
        connect(this, SIGNAL(ccdScanParam(int, int, int, int, int, double, double, double, double, bool)),
                &m->rThread, SLOT(SetccdScanMode(int, int, int, int, int, double, double, double, double, bool)));
    }
    if (name == QLatin1String("data")) {
        DataPlot *m = (DataPlot *)module->basecast();

        //TODO cant connect externalScans here, race. need modules for extscanmodules
        //for now connect individually
        /*foreach (IExternalScan *sc, externalScans)
            if (sc->metaObject()->indexOfSignal("AddDataToContainer(ccd::TSetOfSpectrum*,plot::TMetaStorage)") != -1)
                connect(sc, SIGNAL(AddDataToContainer(ccd::TSetOfSpectrum*,plot::TMetaStorage)),
                        m->Container, SLOT(AddDataToContainer(ccd::TSetOfSpectrum*,plot::TMetaStorage)));*/

        connect(this, SIGNAL(AddDataToContainer(plot::T3dDataOwner, plot::TMetaStorage)),
                m->Container, SLOT(AddDataToContainer(plot::T3dDataOwner, plot::TMetaStorage)));
        connect(this, SIGNAL(AddDataToContainer(plot::TCurvesDataOwner,plot::TMetaStorage)),
                m->Container, SLOT(AddDataToContainer(plot::TCurvesDataOwner,plot::TMetaStorage)));
    }
    if (name == QLatin1String("spm")) {
        SPMTunning *m = (SPMTunning *)module->basecast();

        QSettings settings("Param/motor.conf", QSettings::IniFormat);
        if (settings.value("land_mode", 2).toInt() != 1)
            BMotorLevel->show();

        connect(m->DCfbPrec, SIGNAL(newValue()), this, SLOT(check_speed()));
        connect(m->DCfbHeight, SIGNAL(newValue()), this, SLOT(check_speed()));
        connect(m->BFeedbackOnOffMode, SIGNAL(state_changed(bool)), this, SLOT(check_speed()));

        connect(HVxyzToZero, SIGNAL(clicked()), m, SLOT(sl_HVtoZero()));
        connect(m->CBFBScanner, SIGNAL(currentIndexChanged(int)), this, SLOT(setZFB_status(int)));
        connect(m->BFeedbackOnOff, SIGNAL(state_changed(bool)), this, SLOT(Zpos_dis_en(bool)));

        //TODO central hub for pid values, maybe remote_server hub?
        //TODO add tab from spm::connected
        connect(m->DCfbP, SIGNAL(newValue(double)), DCfbP, SLOT(setValue(double)));
        connect(DCfbP, SIGNAL(newValue(double)), m->DCfbP, SLOT(setValue(double)));
        connect(m->DCfbI, SIGNAL(newValue(double)), DCfbI, SLOT(setValue(double)));
        connect(DCfbI, SIGNAL(newValue(double)), m->DCfbI, SLOT(setValue(double)));
        connect(m->DCfbSetPoint, SIGNAL(newValue(double)), DCfbSetPoint, SLOT(setValue(double)));
        connect(DCfbSetPoint, SIGNAL(newValue(double)), m->DCfbSetPoint, SLOT(setValue(double)));

        connect(DCfbP, SIGNAL(newValue()), m, SLOT(s_sendFBP()));
        connect(DCfbI, SIGNAL(newValue()), m, SLOT(s_sendFBI()));
        connect(DCfbSetPoint, SIGNAL(newValue()), m, SLOT(s_sendSP()));
        connect(DCfbP, SIGNAL(newValue()), m, SLOT(sl_save_to_file()));
        connect(DCfbI, SIGNAL(newValue()), m, SLOT(sl_save_to_file()));
        connect(DCfbSetPoint, SIGNAL(newValue()), m, SLOT(sl_save_to_file()));

        connect(m->DCfbSetPoint, SIGNAL(newDecimals(int)), DCfbSetPoint, SLOT(setDecimals(int)));
        connect(m->DCfbSetPoint, SIGNAL(newMinimum(double)), DCfbSetPoint, SLOT(setMinimum(double)));
        connect(m->DCfbSetPoint, SIGNAL(newMaximum(double)), DCfbSetPoint, SLOT(setMaximum(double)));
        connect(m->DCfbSetPoint->slider, SIGNAL(valueChanged(double)), DCfbSetPoint, SLOT(setValue(double))); //sync value change on gains, etc.
    }
    if (name == QLatin1String("oscilloscope")) {
        Oscilloscope *m = (Oscilloscope *)module->basecast();

        m->acp->scanplot = Plot->TwoDPlot->Pure2dPlot;

        if (!m->second_3030) {
            connect(m->CalibWidget, SIGNAL(sig_csb_speed(int)), this, SLOT(check_speed()));

            connect(m->CalibWidget, SIGNAL(sig_XYZ_range(double,double,double)), this, SLOT(sl_set_XYZ_range(double, double, double)));
            connect(m->CalibWidget, SIGNAL(sig_XYZ_range2(double,double,double)), this, SLOT(sl_set_XYZ_range2(double, double, double)));

            connect(m->CalibWidget, SIGNAL(input_calib_updated()), this, SLOT(sl_CB_update()));
            connect(m->CalibWidget, SIGNAL(input_calib_updated()), this, SLOT(SetSectionUnits()));
        }
    }
    if (name == QLatin1String("litho")) {
        graphed::TGraphEditor *m = (graphed::TGraphEditor *)module->basecast();
        connect(this, SIGNAL(SendRect(QRectF)), m->GetView(), SLOT(SetDocumentRect(QRectF)));
        connect(this, SIGNAL(SendBackground(QImage)), m->GetView()->CanvasController(), SLOT(SetBackground(QImage)));

        if (show_additional_gui())
            findChild<QPushButton *>("BSendToGraphEditor")->show();
    }
}

void SPMModule::connected(const QLatin1String name, I_MODULE *module)
{
    if (name == QLatin1String("eg3000")) {
        USB_THREAD_GROUP *m = (USB_THREAD_GROUP *)module->basecast();

        if (tcontrol)
            connect(&m->rThread, SIGNAL(thermoData(QByteArray)), tcontrol, SLOT(nativeThermoData(QByteArray)), Qt::QueuedConnection);

        connect(&m->wThread, SIGNAL(started()), this, SLOT(reset()));

        connect(&m->rThread, SIGNAL(capBits(int,long)), this, SLOT(capBits(int,long)));
        connect(&m->wThread, SIGNAL(zUpdate(double)), NS_Z_spectr, SLOT(setValue(double)));

        connect(&m->rThread, SIGNAL(sigSPMStatus(double,double,double,double,double)), this, SLOT(set_spot_data(double,double,double,double,double)), Qt::QueuedConnection);
        connect(&m->rThread, SIGNAL(Landing_status(int)), this, SLOT(slot_set_Landing_status(int)), Qt::QueuedConnection);
        connect(&m->rThread, SIGNAL(FreqScanData(QByteArray)), this, SLOT(s_setFrequencyScanData(QByteArray)), Qt::QueuedConnection);
        connect(&m->rThread, SIGNAL(ForceCurveData(QByteArray)), this, SLOT(s_setForceCurveData(QByteArray)), Qt::QueuedConnection);
        connect(&m->rThread, SIGNAL(getDataRecv(int,long)), this, SLOT(sl_get_data(int,long)), Qt::QueuedConnection);

        connect(this, SIGNAL(sig_Zpos(double)), &m->wThread, SLOT(Go_to_Z(double)));
        connect(this, SIGNAL(sig_Z2pos(double)), &m->wThread, SLOT(Go_to_Zobject(double)));

        //spm pid
        connect(this, SIGNAL(sig_setFBOnOff(unsigned char)), &m->wThread, SLOT(FBZ_send(unsigned char)));
        connect(B_PID_sign, SIGNAL(state_changed(bool)), &m->wThread, SLOT(Set_FB_sign(bool)));

        connect(this, SIGNAL(sig_Head_data(int, int, int)), &m->wThread, SLOT(sl_Head_Data(int, int, int)));
        connect(CBFBScanner, SIGNAL(currentIndexChanged(int)), &m->wThread, SLOT(Set_FB_Scanner(int)));
        connect(this, SIGNAL(sig_sendFBP(double)), &m->wThread, SLOT(SET_SPM_P(double)));
        connect(this, SIGNAL(sig_sendFBI(double)), &m->wThread, SLOT(SET_SPM_I(double)));
        connect(this, SIGNAL(sig_sendSP(double)), &m->wThread, SLOT(SET_SPM_Sp(double)));
        connect(this, SIGNAL(sig_FbChannel(int)), &m->wThread, SLOT(SET_SPM_FB_Channel(int)));

        connect(this, SIGNAL(jm_prec(WORD, WORD)), &m->wThread, SLOT(spm_jm_precision(WORD, WORD)));
        connect(this, SIGNAL(jm_height(DWORD)), &m->wThread, SLOT(spm_jm_lift(DWORD)));
        connect(JMdiv, SIGNAL(newValue(int)), &m->wThread, SLOT(spm_jm_div(int)));

        connect(this, SIGNAL(Landing_Data(BYTE, BYTE, double, double, double)), &m->wThread, SLOT(set_landing(BYTE, BYTE, double, double, double)));
        connect(this, SIGNAL(m4landing(int)), &m->wThread, SLOT(set_landing_m4(int)));

        connect(Vtip, SIGNAL(newValue(double)), &m->wThread, SLOT(Set_headDAC(double)));
        connect(Vtip2, SIGNAL(newValue(double)), &m->wThread, SLOT(Set_headDAC2(double)));

        //lockin control
        connect(this, SIGNAL(sig_sendFrequencyRange(double, double, int)), &m->wThread, SLOT(freq_scan_start(double, double, int)));
        connect(this, SIGNAL(sig_spectroscopy(int, double, double, double, int, int)), &m->wThread, SLOT(set_spectroscopy(int, double, double, double, int, int)));

        connect(this, SIGNAL(sig_sendGains(int, int, int)), &m->wThread, SLOT(set_ampl_gains(int, int, int)));
        connect(this, SIGNAL(sig_sendGains2(int, int, int)), &m->wThread, SLOT(set_ampl_gains2(int, int, int)));
        connect(curFrequency, SIGNAL(newValue(double)), &m->wThread, SLOT(set_freq(double)));
        connect(lowPassFilter, SIGNAL(newValue(double)), &m->wThread, SLOT(set_cutoff_freq(double)));
        connect(phaseShift, SIGNAL(newValue(double)), &m->wThread, SLOT(set_phase(double)));
        connect(this, SIGNAL(sig_sendDriveInt(double)), &m->wThread, SLOT(set_piezodrive_volt(double)));

        connect(piezoTip, SIGNAL(toggled(bool)), &m->wThread, SLOT(setPiezoTip(bool)));
        connect(capMode, SIGNAL(toggled(bool)), &m->wThread, SLOT(setCapMode(bool)));
        connect(piezoTip2, SIGNAL(toggled(bool)), &m->wThread, SLOT(setPiezoTip2(bool)));
        connect(capMode2, SIGNAL(toggled(bool)), &m->wThread, SLOT(setCapMode2(bool)));
        connect(efmSinglePass, SIGNAL(toggled(bool)), &m->wThread, SLOT(setEFMplus(bool)));

        connect(lflCurrent, SIGNAL(toggled(bool)), &m->wThread, SLOT(setLFLCurrent(bool)));
        connect(extCurrent, SIGNAL(toggled(bool)), &m->wThread, SLOT(setExtCurrent(bool)));

        connect(curFrequency2, SIGNAL(newValue(double)), &m->wThread, SLOT(setFreq2(double)));
        connect(driveInt2, SIGNAL(newValue(double)), &m->wThread, SLOT(setDrive2(double)));

        connect(Kelvin_on, SIGNAL(toggled(bool)), &m->wThread, SLOT(setKelvinOn(bool)));
        connect(Kelvin_P, SIGNAL(newValue(double)), &m->wThread, SLOT(setKelvinP(double)));

        connect(Edfl_xtalk, SIGNAL(newValue(double)), &m->wThread, SLOT(setDFLXtalk(double)));
        connect(Emag_xtalk, SIGNAL(newValue(double)), &m->wThread, SLOT(setMAGXtalk(double)));
        connect(EQControlGain, SIGNAL(newValue(double)), &m->wThread, SLOT(setQcontrolGain(double)));
    }
    if (name == QLatin1String("data")) {
        DataPlot *m = (DataPlot *)module->basecast();
        connect(this, SIGNAL(AddDataToContainer(plot::TCurvesDataOwner,plot::TMetaStorage)),
                m->Container, SLOT(AddDataToContainer(plot::TCurvesDataOwner,plot::TMetaStorage)));
    }
    if (name == QLatin1String("scan")) {
        ScanPlot *m = (ScanPlot *)module->basecast();

        if (show_advanced_gui()) {
            IExternalScan *sc;
            m->addExternalScan(MainWindow::tr("JM force curve"), sc = getNewJMIExtScan(this));
            if (DataModule *data = (DataModule *)ModuleManager::instance->find1("data"))
                connect(sc, SIGNAL(AddDataToContainer(ccd::TSetOfSpectrum*,plot::TMetaStorage)),
                        data->Container, SLOT(AddDataToContainer(ccd::TSetOfSpectrum*,plot::TMetaStorage)));
        }
    }
    if (name == QLatin1String("motors")) {
        Motors *m = (Motors *)module->basecast();
        motors = m;

        connect(m, SIGNAL(m123stopped()), this, SLOT(slot_m123stopped()));
        connect(BMotors_leveling, SIGNAL(clicked()), m, SLOT(on_Motors_clicked()));
    }
    if (name == QLatin1String("oscilloscope")) {
        Oscilloscope *m = (Oscilloscope *)module->basecast();
        m->acp->spm = this;

        if (!m->second_3030) {
            connect(m->DacsWidget->Head_Dac, SIGNAL(newValue(double)), Vtip, SLOT(setValue(double)));
            connect(Vtip, SIGNAL(newValue(double)), m->DacsWidget->Head_Dac, SLOT(setValue(double)));

            connect(this, SIGNAL(sig_osc_start()), m, SLOT(sl_osc_start()));
            connect(this, SIGNAL(sig_osc_stop()), m, SLOT(sl_osc_stop()));
            connect(this, SIGNAL(sig_osc_channel(int)), m, SLOT(setChA(int)));
            connect(m->SBT, SIGNAL(newValue(int)), SBT, SLOT(setValue(int)));
            connect(SBT, SIGNAL(newValue(int)), m->SBT, SLOT(setValue(int)));
            connect(m, SIGNAL(oscRefreshed()), this, SLOT(osc_refresh()));

            connect(m->CalibWidget, SIGNAL(sig_XYZ_range(double,double,double)), this, SLOT(sl_set_XYZ_range(double, double, double)));
            connect(m->CalibWidget, SIGNAL(sig_XYZ_range2(double,double,double)), this, SLOT(sl_set_XYZ_range2(double, double, double)));

            connect(this, SIGNAL(input_calib_updated()), m->CalibWidget, SIGNAL(input_calib_updated()));
            connect(m->CalibWidget, SIGNAL(input_calib_updated()), this, SLOT(sl_setPlotMode()));
        }
    }
}

void MotorsModule::on_Motors_clicked()
{
    MainWindow *mainWindow = (MainWindow *)ModuleManager::instance->find1("mainwindow");
    layout()->activate();
    move(mainWindow->mapToGlobal(mainWindow->rect().center() - rect().center()));
    show();
    raise();
}
void MotorsModule::connected(const QLatin1String name, I_MODULE *module)
{
    if (name == QLatin1String("eg3000")) {
        USB_THREAD_GROUP *m = (USB_THREAD_GROUP *)module->basecast();

        connect(&m->wThread, SIGNAL(started()), this, SLOT(reset()));
        connect(&m->rThread, SIGNAL(capBits(int,long)), this, SLOT(capBits(int,long)));
        connect(&m->rThread, SIGNAL(Motors_status(int, int)),
                this, SLOT(Slot_Set_Motor_Status_handler(int, int)), Qt::QueuedConnection);

        connect(&m->wThread, SIGNAL(started()), this, SLOT(init()));
        connect(this, SIGNAL(Motor_data_Leveling(int, unsigned char, double, double)),
                &m->wThread, SLOT(set_step_motors_leveling(int, unsigned char, double, double)));
        connect(this, SIGNAL(RawMotor_inv_end(unsigned char, unsigned short)),
                &m->wThread, SLOT(set_raw_motors_inv_end(unsigned char,unsigned short)));
    }
    if (name == QLatin1String("mainwindow")) {
        MainWindow *m = (MainWindow *)module->basecast();

        QSettings settings("Param/gui.conf", QSettings::IniFormat);
        if (!settings.value("display/motors", false).toBool())
            return;

        QPushButton *pb = new QPushButton(makeHugeIcon(":/SPMTuning_window/Motors_leveling.png"), MainWindow::tr("Motors"));
#ifdef _ADROID
        pb->setText(QString());
#endif
        pb->setFlat(true);
        connect(pb, SIGNAL(clicked()), this, SLOT(on_Motors_clicked()));

        QToolBar *TBMain = m->findChild<QToolBar *>("MainTB");
        TBMain->addWidget(pb);
    }
}

void OscilloscopeModule::connected(const QLatin1String name, I_MODULE *module)
{
    if (name == QLatin1String("eg3000") && !second_3030) {
        USB_THREAD_GROUP *m = (USB_THREAD_GROUP *)module->basecast();

        connect(&m->wThread, SIGNAL(started()), this, SLOT(reset()));
        connect(&m->rThread, SIGNAL(capBits(int,long)), acp, SLOT(capBits(int,long)));
        connect(&m->rThread, SIGNAL(getDataRecv(int,long)), acp, SLOT(recv_get_data(int,long)), Qt::QueuedConnection);

        connect(&m->cThread, SIGNAL(device_status(int)), this, SLOT(device_status(int)));


        connect(this, SIGNAL(osc_start()), &m->wThread, SLOT(Osc_Start()));
        connect(this, SIGNAL(osc_stop()), &m->wThread, SLOT(Osc_Stop()));
        connect(&m->rThread, SIGNAL(OscCurveData(QByteArray)), this, SLOT(newOscData(QByteArray)), Qt::QueuedConnection);
        connect(&m->rThread, SIGNAL(OscCurveData(QByteArray)), this, SIGNAL(forwardedOscData(QByteArray)), Qt::QueuedConnection);
        connect(this, SIGNAL(sig_ch_changed(int, int)), &m->wThread, SLOT(Set_osc_Channel(int, int)));


        connect(DacsWidget->Head_Dac, SIGNAL(newValue(double)), &m->wThread, SLOT(Set_headDAC(double)));
        connect(DacsWidget->Extra_Dac_1, SIGNAL(newValue(double)), &m->wThread, SLOT(set_Ex_dac1(double)));
        connect(DacsWidget->Extra_Dac_2, SIGNAL(newValue(double)), &m->wThread, SLOT(set_Ex_dac2(double)));
        connect(DacsWidget->Extra_Dac_3, SIGNAL(newValue(double)), &m->wThread, SLOT(set_Ex_dac3(double)));
        connect(DacsWidget->Extra_Dac_4, SIGNAL(newValue(double)), &m->wThread, SLOT(set_Ex_dac4(double)));

        connect(DacsWidget->X, SIGNAL(newValue(double)), &m->wThread, SLOT(Go_to_X(double)));
        connect(DacsWidget->Y, SIGNAL(newValue(double)), &m->wThread, SLOT(Go_to_Y(double)));
        connect(DacsWidget->Z, SIGNAL(newValue(double)), &m->wThread, SLOT(Go_to_Z(double)));
        connect(DacsWidget->X2, SIGNAL(newValue(double)), &m->wThread, SLOT(Go_to_X2(double)));
        connect(DacsWidget->Y2, SIGNAL(newValue(double)), &m->wThread, SLOT(Go_to_Y2(double)));
        connect(DacsWidget->Z2, SIGNAL(newValue(double)), &m->wThread, SLOT(Go_to_Zobject(double)));
        connect(DacsWidget->Z3, SIGNAL(newValue(double)), &m->wThread, SLOT(Go_to_Z3(double)));

        connect(&m->wThread, SIGNAL(xUpdate(double)), DacsWidget->X, SLOT(setValue(double)));
        connect(&m->wThread, SIGNAL(x2Update(double)), DacsWidget->X2, SLOT(setValue(double)));
        connect(&m->wThread, SIGNAL(yUpdate(double)), DacsWidget->Y, SLOT(setValue(double)));
        connect(&m->wThread, SIGNAL(y2Update(double)), DacsWidget->Y2, SLOT(setValue(double)));
        connect(&m->wThread, SIGNAL(zUpdate(double)), DacsWidget->Z, SLOT(setValue(double)));
        connect(&m->wThread, SIGNAL(z2Update(double)), DacsWidget->Z2, SLOT(setValue(double)));
        connect(&m->wThread, SIGNAL(z3Update(double)), DacsWidget->Z3, SLOT(setValue(double)));

        //calib
        connect(&m->wThread, SIGNAL(started()), CalibWidget, SLOT(reset()));
        connect(&m->rThread, SIGNAL(capBits(int,long)), CalibWidget, SLOT(capBits(int,long)));

        connect(CalibWidget->Load_from_Contr, SIGNAL(clicked()), &m->wThread, SLOT(Usb_Load_Data()));
        connect(CalibWidget->Save_to_Contr, SIGNAL(clicked()), &m->wThread, SLOT(Save_to_EEPROM()));
        connect(CalibWidget, SIGNAL(sig_SentDataToEG3000()), &m->wThread, SLOT(Data_to_Cont()));
        connect(&m->rThread, SIGNAL(Update_EEP_Data()), CalibWidget, SLOT(eeprom_received()));
        connect(CalibWidget, SIGNAL(sigUpdExCalibData()), &m->wThread, SLOT(SendExtraCalib()));

        connect(CalibWidget, SIGNAL(program_controller(QString)), &m->cThread, SLOT(program(QString)));
        connect(CalibWidget, SIGNAL(sig_csb_speed(int)), &m->wThread, SLOT(sl_csb_speed(int)));

        m->rThread.Set_Scanner_Data_out(CalibWidget->Scanner_data_pid_cal_range,
                                        CalibWidget->Scanner_data_scale,
                                        CalibWidget->Scanner_data_Shift,
                                        &CalibWidget->Nlc_status_cl_xy,
                                        &CalibWidget->CL_status_xy,
                                        &CalibWidget->Nlc_status_cl_z,
                                        &CalibWidget->CL_status_z,
                                        &CalibWidget->Nlc_status_ol_xy,
                                        &CalibWidget->Nlc_status_ol_z,
                                        &CalibWidget->free_move_speed,
                                        CalibWidget->Scanner2_data_pid_cal_range,
                                        CalibWidget->Scanner2_data_scale,
                                        CalibWidget->Scanner2_data_Shift,
                                        &CalibWidget->Nlc2_status_cl_xy,
                                        &CalibWidget->CL2_status_xy,
                                        &CalibWidget->Nlc2_status_cl_z,
                                        &CalibWidget->CL2_status_z,
                                        &CalibWidget->Nlc2_status_ol_xy,
                                        &CalibWidget->Nlc2_status_ol_z,
                                        &CalibWidget->free_move_speed2);
        m->wThread.Set_Scanner_Data_in(CalibWidget->Scanner_data_pid_cal_range,
                                       CalibWidget->Scanner_data_scale,
                                       CalibWidget->Scanner_data_Shift,
                                       &CalibWidget->Nlc_status_cl_xy,
                                       &CalibWidget->CL_status_xy,
                                       &CalibWidget->Nlc_status_cl_z,
                                       &CalibWidget->CL_status_z,
                                       &CalibWidget->Nlc_status_ol_xy,
                                       &CalibWidget->Nlc_status_ol_z,
                                       &CalibWidget->free_move_speed,
                                       CalibWidget->Scanner2_data_pid_cal_range,
                                       CalibWidget->Scanner2_data_scale,
                                       CalibWidget->Scanner2_data_Shift,
                                       &CalibWidget->Nlc2_status_cl_xy,
                                       &CalibWidget->CL2_status_xy,
                                       &CalibWidget->Nlc2_status_cl_z,
                                       &CalibWidget->CL2_status_z,
                                       &CalibWidget->Nlc2_status_ol_xy,
                                       &CalibWidget->Nlc2_status_ol_z,
                                       &CalibWidget->free_move_speed2);

        m->wThread.Scanner3_data_pid_cal_range = CalibWidget->Scanner3_data_pid_cal_range;
        m->wThread.Scanner3_data_scale = CalibWidget->Scanner3_data_scale;
        m->wThread.Scanner3_data_Shift = CalibWidget->Scanner3_data_Shift;
        m->wThread.CL3_status_z = &CalibWidget->CL3_status_z;
        m->wThread.free_move_speed3 = &CalibWidget->free_move_speed3;

        m->wThread.SetExtraCalibOut(CalibWidget->adcCalib, CalibWidget->dacCalib);
    }
    if (name == QLatin1String("eg3030") && second_3030) {
        Control3030 *m = (Control3030 *)module->basecast();

        connect(m, SIGNAL(sig_Osc_reset()), this, SLOT(reset()));
        connect(m, SIGNAL(no_firmware()), this, SLOT(no_firmware()));
        connect(m, SIGNAL(boot_failed()), this, SLOT(boot_failed()));


        connect(this, SIGNAL(osc_start()), m, SLOT(Osc_Start()));
        connect(this, SIGNAL(osc_stop()), m, SLOT(Osc_Stop()));
        connect(m, SIGNAL(OscCurveData(QByteArray)), this, SLOT(newOscData(QByteArray)));
        connect(m, SIGNAL(OscCurveData(QByteArray)), this, SIGNAL(forwardedOscData(QByteArray)));
        connect(this, SIGNAL(sig_ch_changed(int, int)), m, SLOT(Set_osc_Channel(int, int)));


        connect(m, SIGNAL(x2Update(double)), DacsWidget->X2, SLOT(setValue(double)));
        connect(m, SIGNAL(y2Update(double)), DacsWidget->Y2, SLOT(setValue(double)));
        connect(m, SIGNAL(z2Update(double)), DacsWidget->Z2, SLOT(setValue(double)));

        connect(DacsWidget->X2, SIGNAL(newValue(double)), m, SLOT(Go_to_Xobject(double)));
        connect(DacsWidget->Y2, SIGNAL(newValue(double)), m, SLOT(Go_to_Yobject(double)));
        connect(DacsWidget->Z2, SIGNAL(newValue(double)), m, SLOT(Go_to_Zobject(double)));

        //calib
        connect(m, SIGNAL(sig_Osc_reset()), CalibWidget, SLOT(reset()));
        connect(CalibWidget->Load_from_Contr, SIGNAL(clicked()), m, SLOT(Usb_Load_Data()));
        connect(CalibWidget->Save_to_Contr, SIGNAL(clicked()), m, SLOT(Save_to_EEPROM()));
        connect(CalibWidget, SIGNAL(sig_SentDataToEG3000()), m, SLOT(Data_to_Cont()));
        connect(m, SIGNAL(Update_EEP_Data()), CalibWidget, SLOT(eeprom_received()));
        connect(CalibWidget, SIGNAL(sigUpdExCalibData()), m, SLOT(SendExtraCalib()));
        connect(CalibWidget, SIGNAL(sig_csb_speed(int)), m, SLOT(sl_csb_speed(int)));

        m->Set_Scanner_Data(CalibWidget->Scanner_data_pid_cal_range,
                            CalibWidget->Scanner_data_scale,
                            CalibWidget->Scanner_data_Shift,
                            &CalibWidget->Nlc_status_cl_xy,
                            &CalibWidget->CL_status_xy,
                            &CalibWidget->Nlc_status_cl_z,
                            &CalibWidget->CL_status_z,
                            &CalibWidget->Nlc_status_ol_xy,
                            &CalibWidget->Nlc_status_ol_z,
                            &CalibWidget->free_move_speed,
                            CalibWidget->Scanner2_data_pid_cal_range,
                            CalibWidget->Scanner2_data_scale,
                            CalibWidget->Scanner2_data_Shift,
                            &CalibWidget->Nlc2_status_cl_xy,
                            &CalibWidget->CL2_status_xy,
                            &CalibWidget->Nlc2_status_cl_z,
                            &CalibWidget->CL2_status_z,
                            &CalibWidget->Nlc2_status_ol_xy,
                            &CalibWidget->Nlc2_status_ol_z,
                            &CalibWidget->free_move_speed2);

        m->SetExtraCalibOut(CalibWidget->adcCalib, CalibWidget->dacCalib);
    }
    if (name == QLatin1String("data")) {
        DataPlot *m = (DataPlot *)module->basecast();
        connect(this, SIGNAL(AddDataToContainer(plot::TCurvesDataOwner,plot::TMetaStorage)),
                m->Container, SLOT(AddDataToContainer(plot::TCurvesDataOwner,plot::TMetaStorage)));
    }
    if (name == QLatin1String("oscilloscope") && !second_3030) {
        Oscilloscope *m = (Oscilloscope *)module->basecast(); //osc2+

        DacsWidget->DacsChannel->addItem(Oscilloscope::tr("Z objective"), 12);

        connect(CalibWidget, SIGNAL(sig_csb_speed(int)), m->CalibWidget->CBSpeedCSB, SLOT(setCurrentIndex(int)));
        connect(m->CalibWidget, SIGNAL(sig_csb_speed(int)), CalibWidget->CBSpeedCSB, SLOT(setCurrentIndex(int)));

        connect(m->CalibWidget, SIGNAL(sig_XYZ_range2(double,double,double)), this, SLOT(sl_z_osc2(double, double, double)));
    }
}

void VideoModule::connected(const QLatin1String name, I_MODULE *module)
{
    if (name == QLatin1String("oscilloscope")) {
        Oscilloscope *m = (Oscilloscope *)module->basecast();
        connect(m->CalibWidget, SIGNAL(sig_XYZ_range(double,double,double)), this, SLOT(sl_range())); //depends on module order, needs osc.calib to eg3k before
        connect(m->CalibWidget, SIGNAL(sig_XYZ_range2(double,double,double)), this, SLOT(sl_range()));
    }
}

#include "grapheditor/core/core.h"
#include "grapheditor/tools/TApplyTool.h"
#include "grapheditor/tools/ToolManager.h"
#include "grapheditor/widgets/TCheckableButton.h"
#include "grapheditor/widgets/TWMisc.h"
void LithoModule::connected(const QLatin1String name, I_MODULE *module)
{
    if (name == QLatin1String("eg3000")) {
        USB_THREAD_GROUP *m = (USB_THREAD_GROUP *)module->basecast();

        connect(this, SIGNAL(sig_XYZPosition(double, double, double)), &m->wThread, SLOT(Go_to_XYZ(double, double, double)));
        connect(this, SIGNAL(sig_XYZPosition2(double, double, double)), &m->wThread, SLOT(Go_to_XYZ2(double, double, double)));
        connect(&m->wThread, SIGNAL(zUpdate(double)), this, SLOT(updZ1(double)));
        connect(&m->wThread, SIGNAL(z2Update(double)), this, SLOT(updZ2(double)));

        graphed::TApplyTool *ApplyTool = (graphed::TApplyTool *)FindTool("TApplyTool");
        m->wThread.setLithResOut(&ApplyTool->ResultsList());
        setResListPointer(&ApplyTool->ResultsList());

        connect(BStartStop, SIGNAL(BChecked(bool)), &m->wThread, SLOT(slLitoStartStop(bool)));
        connect(TxyUpdate, SIGNAL(timeout()), &m->wThread, SLOT(sl_get_curr_xy()));

        connect(BLithoType, SIGNAL(currentIndexChanged(int)), &m->wThread, SLOT(set_Litho_Type(int)));
        connect(BlithoVout, SIGNAL(currentIndexChanged(int)), &m->wThread, SLOT(set_Litho_VOut(int)));
        connect(CBisContinuous, SIGNAL(clicked(bool)), &m->wThread, SLOT(set_Litho_continuous(bool)));
        connect(BScanner, SIGNAL(currentIndexChanged(int)), &m->wThread, SLOT(set_Litho_Scanner(int)));

        connect(&m->rThread, SIGNAL(LithoReady()),&m->wThread, SLOT(sl_LithoNextpacket()));
        connect(&m->rThread, SIGNAL(LithoReady()), this, SLOT(updatePrbar()));
        connect(&m->wThread, SIGNAL(sig_LithoTotalReady()), this, SLOT(Finish()));

        {//litho settings
            QSettings settings("Param/lithography.conf", QSettings::IniFormat);
            BlithoVout->setCurrentIndex(settings.value("VoutType", 0).toInt());
            BScanner->setCurrentIndex(settings.value("ScannerType", 0).toInt());
            BLithoType->setCurrentIndex(settings.value("LithoType", 0).toInt());
            WProperties->getSpeedSlider()->setValue(settings.value("ScanSpeed", 0).toDouble());
        }

        //soft xy sync
        connect(this, SIGNAL(syncX1pos(double)), &m->wThread, SIGNAL(xUpdate(double)));
        connect(this, SIGNAL(syncY1pos(double)), &m->wThread, SIGNAL(yUpdate(double)));
        connect(this, SIGNAL(syncX2pos(double)), &m->wThread, SIGNAL(x2Update(double)));
        connect(this, SIGNAL(syncY2pos(double)), &m->wThread, SIGNAL(y2Update(double)));

        connect(&m->wThread, SIGNAL(xUpdate(double)), this, SLOT(setX1(double)));
        connect(&m->wThread, SIGNAL(x2Update(double)), this, SLOT(setX2(double)));
        connect(&m->wThread, SIGNAL(yUpdate(double)), this, SLOT(setY1(double)));
        connect(&m->wThread, SIGNAL(y2Update(double)), this, SLOT(setY2(double)));

        connect(&m->rThread, SIGNAL(getDataRecvX(double)), this, SLOT(setX1(double)));
        connect(&m->rThread, SIGNAL(getDataRecvX2(double)), this, SLOT(setX2(double)));
        connect(&m->rThread, SIGNAL(getDataRecvY(double)), this, SLOT(setY1(double)));
        connect(&m->rThread, SIGNAL(getDataRecvY2(double)), this, SLOT(setY2(double)));
    }
    if (name == QLatin1String("oscilloscope")) {
        Oscilloscope *m = (Oscilloscope *)module->basecast();

        if (!m->second_3030) {
            connect(m->CalibWidget, SIGNAL(sig_XYZ_range(double,double,double)), this, SLOT(sl_setHeadRange(double, double, double)));
            connect(m->CalibWidget, SIGNAL(sig_XYZ_range2(double,double,double)), this, SLOT(sl_setStageRange(double, double, double)));
        }
    }
}


//---------------------------------------------------------------------------------------
#include "container/datalist.hpp"
void DataModule::initialSend(MobileInstance *client)
{
    QDir dir(TContainer::getDataDir());
    client->serv->setPropIdVarS(client->socket, DataID::DataOpenList, dir.entryList(QStringList() << "*.nstdat"));
}
void DataModule::setValue(int id, QVariant val, QObject *)
{
    if (id == DataID::DataOpenName) {
        QString fname = TContainer::getDataDir() + "/" + val.toString();
        if (val.toString().startsWith("NST_TAB_IDX/")) {
            bool ok = false;
            int idx = val.toString().split("/").last().toInt(&ok);
            if (!ok) return;
            Container->CBDataLists->setCurrentIndex(idx);
            Container->SendContainerData();
            return;
        }
        for (int i = 1; i < Container->CBDataLists->count(); i++) {
            TListWidget *w = ((TListWidget *)Container->CBDataLists->widget(i));
            if (QFileInfo(w->filename).absoluteFilePath() == QFileInfo(fname).absoluteFilePath()) {
                Container->CBDataLists->setCurrentIndex(i);
                Container->SendContainerData();
                return;
            }
        }
        Container->OpenFile(fname);
    }
    if (id == DataID::ActiveData) {
        int id = qMin(Container->CurrentDataList->findItems("*", Qt::MatchWildcard).size()-1, val.toInt());
        if (id < 0) return;
        TListItem *it = Container->CurrentDataList->FindItem(Container->CurrentDataList->findItems("*", Qt::MatchWildcard).at(id));
        if (!it) return;
        Container->CurrentDataList->on_itemPressed(it->Item);
    }
}

void ScanModule::initialSend(MobileInstance *client)
{
    QStringList chnames;
    QList<QVariant> chdata;
    for (int i = 0; i < CBChannel[1]->count(); i++) {
        chnames.append(CBChannel[1]->itemText(i));
        chdata.append(CBChannel[1]->itemData(i));
    }
    client->serv->setPropIdVarS(client->socket, DataID::ChannelList, QList<QVariant>() << chnames << QVariant(chdata));

    client->serv->setPropIdVarS(client->socket, DataID::ScanNxNy, qSqrt(Nx_points * Ny_points));
    if (!m_BStart->on_off) {
        Send2dPlotData();
        Send3dPlotData1();
        Send3dPlotData2();
    }
}
void ScanModule::setValue(int id, QVariant val, QObject *)
{
    if (id == DataID::ScanOnToggle)
        m_BStart->animateClick();

    if (id == DataID::ScanNxNy) {
        int i = val.toInt();
        if (show_additional_gui())
            i += 3;

        CBPointsOX->blockSignals(true);
        CBPointsOY->blockSignals(true);
        CBPointsOXext->blockSignals(true);
        CBPointsOYext->blockSignals(true);
        CBPointsOX->setCurrentIndex(i);
        CBPointsOY->setCurrentIndex(i);
        CBPointsOXext->setCurrentIndex(i);
        CBPointsOYext->setCurrentIndex(i);
        CBPointsOX->blockSignals(false);
        CBPointsOY->blockSignals(false);
        CBPointsOXext->blockSignals(false);
        CBPointsOYext->blockSignals(false);
        check_nx_ny();
        sl_ext_mode();
    }
    if (id == DataID::ScanSaveCh) {
        if (val.toInt() == 4) {
            Plot->TwoDPlot->PlotControl->BSave->animateClick();
        } else {
            Plot->SetActivePlot(val.toInt());
            Plot->PlotControl->BSave->animateClick();
        }
    }

    if (id == DataID::PlotScalesChanged) {
        QVariantList lst = val.toList();
        plot::TScalesChangedEvent e(Range(), Range(), (plot::TScalesChangedEvent::TReason)lst.at(0).toInt());
        e.XLimits[0] = lst.at(1).toDouble();
        e.XLimits[1] = lst.at(2).toDouble();
        e.YLimits[0] = lst.at(3).toDouble();
        e.YLimits[1] = lst.at(4).toDouble();
        Plot->SetPlotsXScales(e.XLimits[0], e.XLimits[1]);
        Plot->SetPlotsYScales(e.YLimits[0], e.YLimits[1]);
        eventFilter(this, &e);
    }
}

void SPMModule::initialSend(MobileInstance *client)
{
    QStringList chnames;
    for (int i = 0; i < fbs->count(); i++)
        chnames.append(fbs->itemText(i));
    client->serv->setPropIdVarS(client->socket, DataID::SPM_FBch_list, chnames);
}
void SPMModule::setValue(int id, QVariant val, QObject *sender)
{
    if (id == DataID::Deflect_send)
        if (sender)
            sender->setProperty("deflect_need", val);
}

void MotorsModule::setValue(int id, QVariant val, QObject *)
{
    if (id == DataID::Motor_123)
        goShuttle(val.toInt(), 250);
    if (id == DataID::Motor_1)
        goShuttle(val.toInt(), 1);
    if (id == DataID::Motor_2)
        goShuttle(val.toInt(), 2);
    if (id == DataID::Motor_3)
        goShuttle(val.toInt(), 3);
    if (id == DataID::Motor_4)
        goShuttle(val.toInt(), 4);
}

void VideoModule::setValue(int id, QVariant val, QObject *sender)
{
    if (inst_idx != 0)
        return;

    if (id == DataID::Video_send)
        if (sender)
            sender->setProperty("video_need", val);

    if (id == DataID::VideoOnToggle)
        m_BStart->animateClick();
    if (id == DataID::VideoXYjoy)
        goXYjoy(val.toPointF());
    if (id == DataID::VideoZjoy)
        goZjoy(val.toDouble());
}
