#include "xyzcalibrations.hpp"
#include "EG3000.h"
#include "mainwindow.h"
#include "autocalib.hpp"
#include "oscilloscope.h"
#include <QVBoxLayout>
#include <QSplitter>
#include <QCheckBox>
#include <QSettings>
#include <QFormLayout>
#include <QFileDialog>
#include <QApplication>
#if (QT_VERSION >= QT_VERSION_CHECK(5, 0, 0))
#include <QtMath>
#else
#include <qmath.h>
#include <qnumeric.h>
#endif

extern USB_WRITE_THREAD *_wThread; //from modules.cpp


QLabel *newLabelCentered(QString text)
{
    QLabel *l = new QLabel(text);
    l->setAlignment(Qt::AlignCenter);
    return l;
}
QLabel *newLabelNoexpand(QString text)
{
    QLabel *l = new QLabel(text);
    l->setSizePolicy(QSizePolicy::Fixed, QSizePolicy::Fixed);
    return l;
}

XYZCalibrations::XYZCalibrations(QWidget *parent, Oscilloscope *_osc, bool second3030) : QWidget(parent)
{
    osc = _osc;
    second_3030 = second3030;
    eeprom_first_load = true;
    nlcpid_loaded = false;
    skip_autocalib = false;
    lastFileName = defaultFileName = !second_3030 ? "Param/default.par" : "Param/default2.par";

    CL_status_xy = false;
    CL2_status_xy = false;
    CL_status_z = false;
    CL2_status_z = false;
    CL3_status_z = false;
    Nlc_status_cl_xy = false;
    Nlc2_status_cl_xy = false;
    Nlc_status_cl_z = false;
    Nlc2_status_cl_z = false;
    Nlc_status_ol_xy = false;
    Nlc2_status_ol_xy = false;
    Nlc_status_ol_z = false;
    Nlc2_status_ol_z = false;


    QVBoxLayout *mlay = new QVBoxLayout(this);
    mlay->setContentsMargins(0, 0, 0, 0);

    QSplitter *splRight = new QSplitter(Qt::Vertical, this);
    mlay->addWidget(splRight);

    QGroupBox *GrBCalibrations = new QGroupBox(splRight);
    QVBoxLayout *Stage_test_layout = new QVBoxLayout(GrBCalibrations);
    Stage_test_layout->setSpacing(6);
    Stage_test_layout->setContentsMargins(6, 6, 6, 6);

    QWidget *hiddenWid = new QWidget(splRight);
    QVBoxLayout *hiddenLay = new QVBoxLayout(hiddenWid);

    splRight->addWidget(GrBCalibrations);
    splRight->addWidget(hiddenWid);
    splRight->setSizes(QList<int>() << 1 << 0);


    // ----------------Stage calibration and test-------------------------
    QPushButton *Save_to_PC = new QPushButton(makeHugeIcon(":/Osc_window/save.png"), Oscilloscope::tr("Save"));
    connect(Save_to_PC, SIGNAL(clicked()), this, SLOT(save_To_File()));

    QPushButton *Load_from_PC = new QPushButton(makeHugeIcon(":/Osc_window/load.png"), Oscilloscope::tr("Load"));
    connect(Load_from_PC, SIGNAL(clicked()), this, SLOT(sl_Load_from_File()));


    Save_to_Contr = new QPushButton(makeHugeIcon(":/Osc_window/save.png"), Oscilloscope::tr("DSP Memory"));
    Load_from_Contr = new QPushButton(makeHugeIcon(":/Osc_window/load.png"), Oscilloscope::tr("DSP Memory"));

    QPushButton *ProgramController = new QPushButton(makeHugeIcon(":/Osc_window/DSPwrite.png"), Oscilloscope::tr("Program controller"));
    if (second_3030) {
        ProgramController->setEnabled(false);
        ProgramController->setToolTip(Oscilloscope::tr("Use second controller as main to program it"));
    }
    connect(ProgramController, SIGNAL(clicked()), this, SLOT(program_controller()));

    PrintDebug = new QPushButton(tr("Debug Info"));

    QHBoxLayout *Stage_Control_layout_1 = new QHBoxLayout;
    Stage_Control_layout_1->setContentsMargins(0, 0, 0, 0);

    QHBoxLayout *Stage_Control_layout_3 = new QHBoxLayout;
    Stage_Control_layout_3->setContentsMargins(0, 0, 0, 0);

    hiddenLay->addLayout(Stage_Control_layout_1);
    hiddenLay->addLayout(Stage_Control_layout_3);

    Stage_Control_layout_1->addWidget(Save_to_Contr);
    Stage_Control_layout_1->addWidget(Load_from_Contr);

    Stage_Control_layout_3->addWidget(ProgramController, 1);
    Stage_Control_layout_3->addWidget(PrintDebug);


    fb_on_off_xy = new TwoStateButton(Oscilloscope::tr("Close Loop XY"), ":/Osc_window/off.png", ":/Osc_window/onn.png");
    fb_on_off_xy->setToolTip(Oscilloscope::tr("Capasitance sensors XY feed back on/off"));
    connect(fb_on_off_xy, SIGNAL(state_changed(bool)), this, SLOT(s_setCLStatus(bool)));

    fb_on_off_z = new TwoStateButton(Oscilloscope::tr("Close Loop Z"), ":/Osc_window/off.png", ":/Osc_window/onn.png");
    fb_on_off_z->setToolTip(Oscilloscope::tr("Capasitance sensors Z feed back on/off"));
    connect(fb_on_off_z, SIGNAL(state_changed(bool)), this, SLOT(s_setCLStatusZ(bool)));

    nl_on_off_cl = new QCheckBox(Oscilloscope::tr("Sensor NLC"));
    nl_on_off_cl->setToolTip(Oscilloscope::tr("Capasitance sensors nonlinearity correction"));
    connect(nl_on_off_cl, SIGNAL(clicked(bool)), this, SLOT(set_nlc_status_cl(bool)));

    nl_on_off_ol = new QCheckBox(Oscilloscope::tr("OL NLC"));
    nl_on_off_ol->setToolTip(Oscilloscope::tr("Voltage actuator nonlinearity correction"));
    connect(nl_on_off_ol, SIGNAL(clicked(bool)), this, SLOT(set_nlc_status_ol(bool)));

    nl_on_off_cl2 = new QCheckBox(Oscilloscope::tr("Sensor NLC"));
    nl_on_off_cl2->setToolTip(Oscilloscope::tr("Capasitance sensors nonlinearity correction"));
    connect(nl_on_off_cl2, SIGNAL(clicked(bool)), this, SLOT(set_nlc_status_cl(bool)));

    nl_on_off_ol2 = new QCheckBox(Oscilloscope::tr("OL NLC"));
    nl_on_off_ol2->setToolTip(Oscilloscope::tr("Voltage actuator nonlinearity correction"));
    connect(nl_on_off_ol2, SIGNAL(clicked(bool)), this, SLOT(set_nlc_status_ol(bool)));

    //xyz
    lineFreeSpeed = new DoubleEdit(0.1,1000,1);
    connect(lineFreeSpeed, SIGNAL(newValue()), this, SLOT(s_sendData()));

    for (int i = 0; i < 3; i++) {
        for (int j = 0; j < 3; j++) {
            linePID[i][j] = new DoubleEdit(0,100,7);
            connect(linePID[i][j], SIGNAL(newValue()), this, SLOT(s_sendData()));
        }

        viewPiezoCalibration[i] = new NLCWidget;
        viewSensorCalibration[i] = new NLCWidget;
        connect(viewPiezoCalibration[i], SIGNAL(clicked()), this, SLOT(nlcEdit()));
        connect(viewSensorCalibration[i], SIGNAL(clicked()), this, SLOT(nlcEdit()));

        lineScannerRangeOL[i] = new DoubleEdit(1,1000,4);
        lineScannerRangeCL[i] = new DoubleEdit(1,1000,4);
        lineSensorScale[i] = new DoubleEdit(1,100000,0);
        lineSensorScale0[i] = new DoubleEdit(-32767,32767,0);

        connect(lineScannerRangeOL[i], SIGNAL(newValue()), this, SLOT(s_sendData()));
        connect(lineScannerRangeCL[i], SIGNAL(newValue()), this, SLOT(s_sendData()));
        connect(lineScannerRangeOL[i], SIGNAL(newValue()), this, SLOT(sl_XYZ_range()));
        connect(lineScannerRangeCL[i], SIGNAL(newValue()), this, SLOT(sl_XYZ_range()));
        connect(lineSensorScale[i], SIGNAL(newValue()), this, SLOT(s_sendData()));
        connect(lineSensorScale0[i], SIGNAL(newValue()), this, SLOT(s_sendData()));
    }

    //xyz2
    lineFreeSpeed2 = new DoubleEdit(0.1,1000,1);
    connect(lineFreeSpeed2, SIGNAL(newValue()), this, SLOT(s_sendData()));

    for (int i = 0; i < 3; i++) {
        for (int j = 0; j < 3; j++) {
            linePID2[i][j] = new DoubleEdit(0,100,7);
            connect(linePID2[i][j], SIGNAL(newValue()), this, SLOT(s_sendData()));
        }

        viewPiezoCalibration2[i] = new NLCWidget;
        viewSensorCalibration2[i] = new NLCWidget;
        connect(viewPiezoCalibration2[i], SIGNAL(clicked()), this, SLOT(nlcEdit()));
        connect(viewSensorCalibration2[i], SIGNAL(clicked()), this, SLOT(nlcEdit()));

        lineScannerRangeOL2[i] = new DoubleEdit(1,1000,4);
        lineScannerRangeCL2[i] = new DoubleEdit(1,1000,4);
        lineSensorScale2[i] = new DoubleEdit(1,100000,0);
        lineSensorScale02[i] = new DoubleEdit(-32767,32767,0);

        connect(lineScannerRangeOL2[i], SIGNAL(newValue()), this, SLOT(s_sendData()));
        connect(lineScannerRangeCL2[i], SIGNAL(newValue()), this, SLOT(s_sendData()));
        connect(lineScannerRangeOL2[i], SIGNAL(newValue()), this, SLOT(sl_XYZ_range()));
        connect(lineScannerRangeCL2[i], SIGNAL(newValue()), this, SLOT(sl_XYZ_range()));
        connect(lineSensorScale2[i], SIGNAL(newValue()), this, SLOT(s_sendData()));
        connect(lineSensorScale02[i], SIGNAL(newValue()), this, SLOT(s_sendData()));
    }

    //Z3
    linePIDProportionalZ3 = new DoubleEdit(0,100,7);
    linePIDIntegralZ3 = new DoubleEdit(0,100,7);

    connect(linePIDProportionalZ3, SIGNAL(newValue()), this, SLOT(s_sendData()));
    connect(linePIDIntegralZ3, SIGNAL(newValue()), this, SLOT(s_sendData()));

    lineScannerRangeOLZ3 = new DoubleEdit(1,1000,4);
    lineScannerRangeCLZ3 = new DoubleEdit(1,1000,4);
    lineSensorScaleZ3 = new DoubleEdit(1,100000,0);
    lineSensorScaleZ03 = new DoubleEdit(-32767,32767,0);

    connect(lineScannerRangeOLZ3, SIGNAL(newValue()), this, SLOT(s_sendData()));
    connect(lineScannerRangeCLZ3, SIGNAL(newValue()), this, SLOT(s_sendData()));
    connect(lineScannerRangeOLZ3, SIGNAL(newValue()), this, SLOT(sl_XYZ_range()));
    connect(lineScannerRangeCLZ3, SIGNAL(newValue()), this, SLOT(sl_XYZ_range()));
    connect(lineSensorScaleZ3, SIGNAL(newValue()), this, SLOT(s_sendData()));
    connect(lineSensorScaleZ03, SIGNAL(newValue()), this, SLOT(s_sendData()));

    lineFreeSpeed3 = new DoubleEdit(0.1,1000,1);
    connect(lineFreeSpeed3, SIGNAL(newValue()), this, SLOT(s_sendData()));


    //lvm
    CBlvm[0] = new QCheckBox(Oscilloscope::tr("Low voltage mode"));
    CBlvm[1] = new QCheckBox(Oscilloscope::tr("Low voltage mode"));
    connect(CBlvm[0], SIGNAL(clicked()), this, SLOT(lvmClicked()));
    connect(CBlvm[1], SIGNAL(clicked()), this, SLOT(lvmClicked()));

    //nlcPID
    CBnlcPID[0] = new QCheckBox(Oscilloscope::tr("Enable PID NLC") + " XY");
    CBnlcPID[1] = new QCheckBox(Oscilloscope::tr("Enable PID NLC") + " XY");
    connect(CBnlcPID[0], SIGNAL(clicked()), this, SLOT(nlcPIDSend()));
    connect(CBnlcPID[1], SIGNAL(clicked()), this, SLOT(nlcPIDSend()));
    CBnlcPIDz[0] = new QCheckBox("Z");
    CBnlcPIDz[1] = new QCheckBox("Z");
    connect(CBnlcPIDz[0], SIGNAL(clicked()), this, SLOT(nlcPIDSend()));
    connect(CBnlcPIDz[1], SIGNAL(clicked()), this, SLOT(nlcPIDSend()));
    if (!second_3030) {
        _wThread->nlcPID = (double *)nlcPID;
        _wThread->nlcPIDOn = (BYTE *)nlcPIDon;
    }

    //nlc edit
    QGridLayout *nlcEditLay = new QGridLayout;
    nlcEditLay->addWidget(newLabelCentered("f(x) = a* x^4"),0, 0);
    nlcEditLay->addWidget(newLabelCentered(" + b* x^3"),0,1);
    nlcEditLay->addWidget(newLabelCentered(" + c* x^2"),0,2);
    nlcEditLay->addWidget(newLabelCentered(" + d* x"),0,3);
    nlcEditLay->addWidget(newLabelCentered(" + e"),0,4);
    //nlcEditLay->addWidget(newLabelCentered(" + f"),0,5);

    for (int i = 0; i < 6; i++) {
        lineNLCEdit[i] = new DoubleEdit(-1000000, 1000000, 10);
        connect(lineNLCEdit[i], SIGNAL(newValue()), this, SLOT(nlcEditUpd()));
        nlcEditLay->addWidget(lineNLCEdit[i], 1, i);
    }
    lineNLCEdit[5]->hide();

    viewNLCEdit = new NLCWidget;
    viewNLCEdit->setFixedWidth(lineNLCEdit[0]->width());
    nlcEditLay->addWidget(viewNLCEdit,0,6, 2, 1, Qt::AlignBottom);

    NLCEditValid = new QLabel();

    //QDialogButtonBox TODO
    QPushButton *nlcApply = new QPushButton(Oscilloscope::tr("Apply"));
    QPushButton *nlcCancel = new QPushButton(Oscilloscope::tr("Cancel"));
    QPushButton *nlcRst = new QPushButton(Oscilloscope::tr("Reset"));
    QHBoxLayout *nlcBLay = new QHBoxLayout;
    nlcCancel->setShortcut(QKeySequence(Qt::Key_Escape));
    nlcBLay->addWidget(nlcRst);
    nlcBLay->addWidget(NLCEditValid);
    nlcBLay->addStretch(1);
    nlcBLay->addWidget(nlcApply);
    nlcBLay->addWidget(nlcCancel);

    NLCEditWid = new QWidget(this, Qt::Window);
    NLCEditWid->setWindowTitle(Oscilloscope::tr("Function editor"));
    NLCEditWid->setWindowModality(Qt::WindowModal);

    QVBoxLayout *nlcLay = new QVBoxLayout(NLCEditWid);
    nlcLay->setSizeConstraint(QLayout::SetFixedSize);
    nlcLay->addLayout(nlcEditLay);
    nlcLay->addLayout(nlcBLay);
    connect(nlcRst, SIGNAL(clicked()), this, SLOT(nlcEditRst()));
    connect(nlcApply, SIGNAL(clicked()), this, SLOT(nlcEditApply()));
    connect(nlcCancel, SIGNAL(clicked()), NLCEditWid, SLOT(hide()));


    //-----------------------------------------------------------------
    QHBoxLayout *Stage_Control_layout_0 = new QHBoxLayout;
    Stage_Control_layout_0->setContentsMargins(0, 0, 0, 0);
    Stage_Control_layout_0->addWidget(Save_to_PC);
    Stage_Control_layout_0->addWidget(Load_from_PC);

    QHBoxLayout *Stage_Control_layout_2 = new QHBoxLayout;
    Stage_Control_layout_2->setContentsMargins(0, 0, 0, 0);
    Stage_Control_layout_2->addWidget(fb_on_off_xy);
    Stage_Control_layout_2->addWidget(fb_on_off_z);

    QVBoxLayout *Stage_Control_layout = new QVBoxLayout;
    Stage_Control_layout->setContentsMargins(0, 0, 0, 0);
    Stage_Control_layout->addLayout(Stage_Control_layout_0);
    Stage_Control_layout->addLayout(Stage_Control_layout_2);


    //
    QScrollArea *LeftWidScroll1 = new TScrollArea;
    QScrollArea *LeftWidScroll2 = new TScrollArea;
    QScrollArea *LeftWidScroll3 = new TScrollArea;

    QSettings settings("Param/gui.conf", QSettings::IniFormat);

    Scanner_type = new QTabWidget;
    Scanner_type->addTab(LeftWidScroll1, makeHugeIcon(":Head_page.png"), Oscilloscope::tr("Head"));
    Scanner_type->addTab(LeftWidScroll2, makeHugeIcon(":stage_page.png"),Oscilloscope::tr("Stage"));
    if (!second_3030 && settings.value("display/oscilloscope/z3", false).toBool())
    Scanner_type->addTab(LeftWidScroll3, Oscilloscope::tr("Z3"));

    Scanner_type->setTabEnabled(0, settings.value("display/oscilloscope/head", true).toBool() && !second_3030);
    Scanner_type->setTabEnabled(1, settings.value("display/oscilloscope/stage", true).toBool());
    //Scanner_type->setTabEnabled(2, settings.value("display/oscilloscope/z3", false).toBool() && !second_3030);

    connect(Scanner_type, SIGNAL(currentChanged(int)), this, SLOT(CL_NLS_Lamps(int)));


    CBSpeedCSB = new QComboBox;
    CBSpeedCSB->addItem(Oscilloscope::tr("10 kHz"), 10000);
    CBSpeedCSB->addItem(Oscilloscope::tr("5 kHz"), 5000);
    CBSpeedCSB->addItem(Oscilloscope::tr("3.3 kHz"), 3333);
    CBSpeedCSB->addItem(Oscilloscope::tr("2 kHz"), 2000);
    CBSpeedCSB->addItem(Oscilloscope::tr("1 kHz"), 1000);
    CBSpeedCSB->addItem(Oscilloscope::tr("500 Hz"), 500);
    CBSpeedCSB->addItem(Oscilloscope::tr("333 Hz"), 333);
    CBSpeedCSB->addItem(Oscilloscope::tr("100 Hz"), 100);
    {
        QSettings settings("Param/controller.conf", QSettings::IniFormat);
        CBSpeedCSB->setCurrentIndex(settings.value("csb_speed_settings_idx", 1).toInt());
    }
    connect(CBSpeedCSB, SIGNAL(currentIndexChanged(int)), this, SLOT(sl_csb_speed(int)));

    QHBoxLayout *CBSpeedLay = new QHBoxLayout;
    CBSpeedLay->addWidget(new QLabel(Oscilloscope::tr("Capacitor sensors & PID rate: ")));
    CBSpeedLay->addWidget(CBSpeedCSB);


    Stage_test_layout->addLayout(Stage_Control_layout);
    Stage_test_layout->addWidget(Scanner_type);
    Stage_test_layout->addLayout(CBSpeedLay);


    //
    QWidget *headCalibWid = new QWidget;
    QVBoxLayout *Stage_data_layout = new QVBoxLayout(headCalibWid);
    Stage_data_layout->setSpacing(6);
    Stage_data_layout->setContentsMargins(6, 3, 6, 3);

    QWidget *stageCalibWid = new QWidget;
    QVBoxLayout *Stage_data_layout2 = new QVBoxLayout(stageCalibWid);
    Stage_data_layout2->setSpacing(6);
    Stage_data_layout2->setContentsMargins(6, 3, 6, 3);

    QWidget *z3CalibWid = new QWidget;
    QVBoxLayout *Stage_data_layout3 = new QVBoxLayout(z3CalibWid);
    Stage_data_layout3->setSpacing(6);
    Stage_data_layout3->setContentsMargins(6, 3, 6, 3);

    LeftWidScroll1->setWidget(headCalibWid);
    LeftWidScroll2->setWidget(stageCalibWid);
    LeftWidScroll3->setWidget(z3CalibWid);


    QGridLayout *layout_config = new QGridLayout;
    layout_config->setSpacing(6);
    layout_config->setContentsMargins(6, 6, 6, 6);
    Stage_data_layout->addLayout(layout_config);

    QGridLayout *layout_config2 = new QGridLayout;
    layout_config2->setSpacing(6);
    layout_config2->setContentsMargins(6, 6, 6, 6);
    Stage_data_layout2->addLayout(layout_config2);

    QFormLayout *layout_config3 = makeFormLayout();
    layout_config3->setSpacing(6);
    layout_config3->setContentsMargins(6, 6, 6, 6);
    Stage_data_layout3->addLayout(layout_config3);


    //xyz
    int c;

    c = 0;

    layout_config->addWidget(newLabelCentered(Oscilloscope::tr("Range")), c, 0);
    layout_config->addWidget(newLabelCentered("X"), c, 1);
    layout_config->addWidget(newLabelCentered("Y"), c, 2);
    layout_config->addWidget(newLabelCentered("Z"), c, 3);
    c++;
    for (int i = 0; i < 3; i++) {
        layout_config->addWidget(lineScannerRangeOL[i], c+0, i+1);
        layout_config->addWidget(lineScannerRangeCL[i], c+1, i+1);
    }
    layout_config->addWidget(newLabelNoexpand(Oscilloscope::tr("OL, ")+Oscilloscope::tr("micro")+Oscilloscope::tr("m")), c++, 0);
    layout_config->addWidget(newLabelNoexpand(Oscilloscope::tr("CL, ")+Oscilloscope::tr("micro")+Oscilloscope::tr("m")), c++, 0);


    layout_config->addWidget(newLabelCentered(Oscilloscope::tr("Sensor")), c, 0);
    layout_config->addWidget(newLabelCentered("X"), c, 1);
    layout_config->addWidget(newLabelCentered("Y"), c, 2);
    layout_config->addWidget(newLabelCentered("Z"), c, 3);
    c++;
    for (int i = 0; i < 3; i++) {
        layout_config->addWidget(lineSensorScale[i], c+0, i+1);
        layout_config->addWidget(lineSensorScale0[i], c+1, i+1);
    }
    layout_config->addWidget(newLabelNoexpand(Oscilloscope::tr("Gain")), c++, 0);
    layout_config->addWidget(newLabelNoexpand(Oscilloscope::tr("Shift")), c++, 0);


    layout_config->addWidget(newLabelCentered(Oscilloscope::tr("PI")), c, 0);
    layout_config->addWidget(newLabelCentered("X"), c, 1);
    layout_config->addWidget(newLabelCentered("Y"), c, 2);
    layout_config->addWidget(newLabelCentered("Z"), c, 3);
    c++;
    for (int i = 0; i < 2/*3*/; i++)
        for (int j = 0; j < 3; j++)
            layout_config->addWidget(linePID[i][j], c+i, j+1);
    layout_config->addWidget(newLabelNoexpand(Oscilloscope::tr("Proportional")), c++, 0);
    layout_config->addWidget(newLabelNoexpand(Oscilloscope::tr("Integral")), c++, 0);
    //layout_config->addWidget(newLabelNoexpand(Oscilloscope::tr("Derivative")), c++, 0);


    layout_config->addWidget(newLabelCentered("X"), c, 1);
    layout_config->addWidget(newLabelCentered("Y"), c, 2);
    layout_config->addWidget(newLabelCentered("Z"), c, 3);
    c++;
    for (int i = 0; i < 3; i++)
        layout_config->addWidget(viewPiezoCalibration[i], c, i+1);
    layout_config->addWidget(nl_on_off_ol, c, 0);
    c++;


    layout_config->addWidget(newLabelCentered("X"), c, 1);
    layout_config->addWidget(newLabelCentered("Y"), c, 2);
    layout_config->addWidget(newLabelCentered("Z"), c, 3);
    c++;
    for (int i = 0; i < 3; i++)
        layout_config->addWidget(viewSensorCalibration[i], c, i+1);
    layout_config->addWidget(nl_on_off_cl, c, 0);
    c++;

    //xyz2
    c = 0;

    layout_config2->addWidget(newLabelCentered(Oscilloscope::tr("Range")), c, 0);
    layout_config2->addWidget(newLabelCentered("X"), c, 1);
    layout_config2->addWidget(newLabelCentered("Y"), c, 2);
    layout_config2->addWidget(newLabelCentered("Z"), c, 3);
    c++;
    for (int i = 0; i < 3; i++) {
        layout_config2->addWidget(lineScannerRangeOL2[i], c+0, i+1);
        layout_config2->addWidget(lineScannerRangeCL2[i], c+1, i+1);
    }
    layout_config2->addWidget(newLabelNoexpand(Oscilloscope::tr("OL, ")+Oscilloscope::tr("micro")+Oscilloscope::tr("m")), c++, 0);
    layout_config2->addWidget(newLabelNoexpand(Oscilloscope::tr("CL, ")+Oscilloscope::tr("micro")+Oscilloscope::tr("m")), c++, 0);


    layout_config2->addWidget(newLabelCentered(Oscilloscope::tr("Sensor")), c, 0);
    layout_config2->addWidget(newLabelCentered("X"), c, 1);
    layout_config2->addWidget(newLabelCentered("Y"), c, 2);
    layout_config2->addWidget(newLabelCentered("Z"), c, 3);
    c++;
    for (int i = 0; i < 3; i++) {
        layout_config2->addWidget(lineSensorScale2[i], c+0, i+1);
        layout_config2->addWidget(lineSensorScale02[i], c+1, i+1);
    }
    layout_config2->addWidget(newLabelNoexpand(Oscilloscope::tr("Gain")), c++, 0);
    layout_config2->addWidget(newLabelNoexpand(Oscilloscope::tr("Shift")), c++, 0);


    layout_config2->addWidget(newLabelCentered(Oscilloscope::tr("PI")), c, 0);
    layout_config2->addWidget(newLabelCentered("X"), c, 1);
    layout_config2->addWidget(newLabelCentered("Y"), c, 2);
    layout_config2->addWidget(newLabelCentered("Z"), c, 3);
    c++;
    for (int i = 0; i < 2/*3*/; i++)
        for (int j = 0; j < 3; j++)
            layout_config2->addWidget(linePID2[i][j], c+i, j+1);
    layout_config2->addWidget(newLabelNoexpand(Oscilloscope::tr("Proportional")), c++, 0);
    layout_config2->addWidget(newLabelNoexpand(Oscilloscope::tr("Integral")), c++, 0);
    //layout_config2->addWidget(newLabelNoexpand(Oscilloscope::tr("Derivative")), c++, 0);


    layout_config2->addWidget(newLabelCentered("X"), c, 1);
    layout_config2->addWidget(newLabelCentered("Y"), c, 2);
    layout_config2->addWidget(newLabelCentered("Z"), c, 3);
    c++;
    for (int i = 0; i < 3; i++)
        layout_config2->addWidget(viewPiezoCalibration2[i], c, i+1);
    layout_config2->addWidget(nl_on_off_ol2, c, 0);
    c++;


    layout_config2->addWidget(newLabelCentered("X"), c, 1);
    layout_config2->addWidget(newLabelCentered("Y"), c, 2);
    layout_config2->addWidget(newLabelCentered("Z"), c, 3);
    c++;
    for (int i = 0; i < 3; i++)
        layout_config2->addWidget(viewSensorCalibration2[i], c, i+1);
    layout_config2->addWidget(nl_on_off_cl2, c, 0);
    c++;

    //z3
    layout_config3->addRow(Oscilloscope::tr("Range OL: "), lineScannerRangeOLZ3);
    layout_config3->addRow(Oscilloscope::tr("Range CL: "), lineScannerRangeCLZ3);
    layout_config3->addRow(Oscilloscope::tr("Scale: "), lineSensorScaleZ3);
    layout_config3->addRow(Oscilloscope::tr("Shift: "), lineSensorScaleZ03);
    layout_config3->addRow(Oscilloscope::tr("P: "), linePIDProportionalZ3);
    layout_config3->addRow(Oscilloscope::tr("I: "), linePIDIntegralZ3);


    //-------------------------------------------------------
    QHBoxLayout *pidnlcxyz_l = new QHBoxLayout;
    pidnlcxyz_l->addWidget(CBnlcPID[0]);
    pidnlcxyz_l->addWidget(CBnlcPIDz[0]);

    QHBoxLayout *pidnlcxyz2_l = new QHBoxLayout;
    pidnlcxyz2_l->addWidget(CBnlcPID[1]);
    pidnlcxyz2_l->addWidget(CBnlcPIDz[1]);


    QHBoxLayout *free_speed_l = new QHBoxLayout;
    free_speed_l->addWidget(new QLabel(Oscilloscope::tr("Free positioning speed: ")));
    free_speed_l->addWidget(lineFreeSpeed);

    QHBoxLayout *free_speed2_l = new QHBoxLayout;
    free_speed2_l->addWidget(new QLabel(Oscilloscope::tr("Free positioning speed: ")));
    free_speed2_l->addWidget(lineFreeSpeed2);

    QHBoxLayout *free_speed3_l = new QHBoxLayout;
    free_speed3_l->addWidget(new QLabel(Oscilloscope::tr("Free positioning speed: ")));
    free_speed3_l->addWidget(lineFreeSpeed3);


    //-------------------------------------------------------
    Stage_data_layout->addLayout(pidnlcxyz_l);
    Stage_data_layout->addLayout(free_speed_l);
    Stage_data_layout->addWidget(CBlvm[0]);
    Stage_data_layout->addStretch(1);

    //
    Stage_data_layout2->addLayout(pidnlcxyz2_l);
    Stage_data_layout2->addLayout(free_speed2_l);
    Stage_data_layout2->addWidget(CBlvm[1]);
    Stage_data_layout2->addStretch(1);

    //
    Stage_data_layout3->addLayout(free_speed3_l);
    QPushButton *pb;
    Stage_data_layout3->addWidget(pb = new QPushButton(Oscilloscope::tr("Load")));
    connect(pb, SIGNAL(clicked()), this, SLOT(z3Load()));
    connect(pb, SIGNAL(clicked()), this, SLOT(s_sendData()));
    Stage_data_layout3->addWidget(pb = new QPushButton(Oscilloscope::tr("Save")));
    connect(pb, SIGNAL(clicked()), this, SLOT(z3Save()));
    Stage_data_layout3->addStretch(1);
}


void XYZCalibrations::program_controller()
{
    QString SelectedFilter;
    QString FileName;

    FileName = QFileDialog::getOpenFileName(this,
                                            Oscilloscope::tr("Select firmware..."),
                                            QString(),
                                            "*.hex;;*.rpd",
                                            &SelectedFilter,
                                            QFileDialog::DontUseNativeDialog);

    if (FileName.isEmpty())
        return;

    if (!FileName.endsWith(SelectedFilter.remove(0, 1)))
        FileName += SelectedFilter;

    emit program_controller(FileName);
}


void XYZCalibrations::sl_csb_speed(int idx)
{
    QSettings settings("Param/controller.conf", QSettings::IniFormat);
    settings.setValue("csb_speed_settings_idx", idx);

    emit sig_csb_speed(idx);
}

void XYZCalibrations::s_setCLStatus(bool status)
{
    switch (Scanner_type->currentIndex()) {
    case scHead:
        CL_status_xy = status;
        break;
    case scStage:
        CL2_status_xy = status;
        break;
    }
    s_sendData();
    sl_XYZ_range();
}
void XYZCalibrations::s_setCLStatusZ(bool status)
{
    switch (Scanner_type->currentIndex()) {
    case scHead:
        CL_status_z = status;
        break;
    case scStage:
        CL2_status_z = status;
        break;
    case scZ3:
        CL3_status_z = status;
        break;
    }
    s_sendData();
    sl_XYZ_range();
}
void XYZCalibrations::set_nlc_status_cl(bool st)
{
    switch (Scanner_type->currentIndex()) {
    case 0:
        Nlc_status_cl_z = Nlc_status_cl_xy = st;
        break;
    case 1:
        Nlc2_status_cl_z = Nlc2_status_cl_xy = st;
        break;
    }
    s_sendData();
}
void XYZCalibrations::set_nlc_status_ol(bool st)
{
    switch (Scanner_type->currentIndex()) {
    case 0:
        Nlc_status_ol_z = Nlc_status_ol_xy = st;
        break;
    case 1:
        Nlc2_status_ol_z = Nlc2_status_ol_xy = st;
        break;
    }
    s_sendData();
}
void XYZCalibrations::CL_NLS_Lamps(int scanner)
{
    switch (scanner) {
    case scHead:
        fb_on_off_xy->setEnabled(true);
        fb_on_off_xy->change_state_man(CL_status_xy);
        fb_on_off_z->change_state_man(CL_status_z);
        nl_on_off_cl->setChecked(Nlc_status_cl_xy && Nlc_status_cl_z);
        nl_on_off_ol->setChecked(Nlc_status_ol_xy && Nlc_status_ol_z);
        break;
    case scStage:
        fb_on_off_xy->setEnabled(true);
        fb_on_off_xy->change_state_man(CL2_status_xy);
        fb_on_off_z->change_state_man(CL2_status_z);
        nl_on_off_cl2->setChecked(Nlc2_status_cl_xy && Nlc2_status_cl_z);
        nl_on_off_ol2->setChecked(Nlc2_status_ol_xy && Nlc2_status_ol_z);
        break;
    case scZ3:
        fb_on_off_xy->setEnabled(false);
        fb_on_off_xy->change_state_man(false);
        fb_on_off_z->change_state_man(CL3_status_z);
        break;
    }
}


void XYZCalibrations::nlcPIDLoad()
{
    nlcpid_loaded = false;

    QSettings settings("Param/calib.conf", QSettings::IniFormat);
    CBlvm[0]->setChecked(settings.value("lvm/0").toBool());
    CBlvm[1]->setChecked(settings.value("lvm/1").toBool());

    CBnlcPID[0]->setChecked(settings.value("nlcPID/x0").toBool() || settings.value("nlcPID/y0").toBool());
    CBnlcPID[1]->setChecked(settings.value("nlcPID/x1").toBool() || settings.value("nlcPID/y1").toBool());
    CBnlcPIDz[0]->setChecked(settings.value("nlcPID/z0").toBool());
    CBnlcPIDz[1]->setChecked(settings.value("nlcPID/z1").toBool());

    settings.beginReadArray("nlcPID");
    for (int i = 0; i < 2*3*6; i++) {
        settings.setArrayIndex(i);
        ((double *)nlcPID)[i] = settings.value("val").toDouble();
    }
    settings.endArray();

    nlcpid_loaded = true;

    nlcPIDSend();
}
void XYZCalibrations::nlcPIDSave()
{
    if (second_3030 || !nlcpid_loaded)
        return;

    QSettings settings("Param/calib.conf", QSettings::IniFormat);
    settings.setValue("lvm/0", CBlvm[0]->isChecked());
    settings.setValue("lvm/1", CBlvm[1]->isChecked());

    settings.setValue("nlcPID/x0", CBnlcPID[0]->isChecked());
    settings.setValue("nlcPID/y0", CBnlcPID[0]->isChecked());
    settings.setValue("nlcPID/z0", CBnlcPIDz[0]->isChecked());
    settings.setValue("nlcPID/x1", CBnlcPID[1]->isChecked());
    settings.setValue("nlcPID/y1", CBnlcPID[1]->isChecked());
    settings.setValue("nlcPID/z1", CBnlcPIDz[1]->isChecked());

    settings.beginWriteArray("nlcPID", 2*3*6);
    for (int i = 0; i < 2*3*6; i++) {
        settings.setArrayIndex(i);
        settings.setValue("val", ((double *)nlcPID)[i]);
    }
    settings.endArray();
}
void XYZCalibrations::nlcPIDSend()
{
    if (second_3030)
        return;

    nlcPIDon[0][0] = nlcPIDon[0][1] = CBnlcPID[0]->isChecked(); nlcPIDon[0][2] = CBnlcPIDz[0]->isChecked();
    nlcPIDon[1][0] = nlcPIDon[1][1] = CBnlcPID[1]->isChecked(); nlcPIDon[1][2] = CBnlcPIDz[1]->isChecked();

    _wThread->sendPIDnlc = true;
}


template<class T>
double NewtonResult(T arr[5], double x0, double value, double ac, DWORD max = 100)
{
    double x_new = x0;
    WORD N = 0;

    do {
        x0 = x_new;
        x_new = x0 -
                (((((arr[0] * x0 + arr[1]) * x0 + arr[2]) * x0 + arr[3]) * x0 + arr[4]) - value) /
                (((4*arr[0] * x0 + 3*arr[1]) * x0 + 2*arr[2]) * x0 + arr[3]);
        N++;
    } while ((fabs(x0 - x_new) > ac) && (N < max));

    return x_new;
}

inline
double polynom_at_pos(double pos, double coeff[5]) {
    double acc = coeff[0];
    for (int i = 1; i < 5; i++)
        acc = acc * pos + coeff[i];

    return acc;
}

QPicture buildNlcPic(double k[6])
{
    QPicture pic;

    double ShiftCL = NewtonResult(k, 10, 0, 0.001);
    double CalibRangeCL = NewtonResult(k, 50, 65535, 0.001) - ShiftCL;

    if (!qIsFinite(ShiftCL) || !qIsFinite(CalibRangeCL))
        return pic;

    QPainter p(&pic);
    QPen pen(Qt::red);
    pen.setCosmetic(true);
    p.setPen(pen);
    double v = 0;
    for (int i = 0; i < 100; i++) {
        double x = i/100.;
        double x1 = x * CalibRangeCL + ShiftCL;
        double v1 = polynom_at_pos(x1, k) / 65535.;
        p.drawLine(QPointF(x, v), QPointF(x + 1/100., v1));
        v = v1;
    }

    return pic;
}

void XYZCalibrations::nlcEdit()
{
    double *source = 0;
    NLCWidget *status = qobject_cast<NLCWidget *>(sender());
    if (!status)
        return;

    for (int i = 0; !source && i < 3; i++)
        if (status == viewPiezoCalibration[i])
            source = linePiezoCalibration[i];
    for (int i = 0; !source && i < 3; i++)
        if (status == viewSensorCalibration[i])
            source = lineSensorCalibration[i];
    for (int i = 0; !source && i < 3; i++)
        if (status == viewPiezoCalibration2[i])
            source = linePiezoCalibration2[i];
    for (int i = 0; !source && i < 3; i++)
        if (status == viewSensorCalibration2[i])
            source = lineSensorCalibration2[i];

    if (!source)
        return;

    for (int i = 0; i < 6; i++)
        lineNLCEdit[i]->setValue(source[i]);
    nlcEditUpd(); //viewNLCEdit->SetPic(status->m_pic);

    NLCEditWid->setProperty("source", QVariant::fromValue((void *)source));
    NLCEditWid->setProperty("status", QVariant::fromValue((void *)status));
    NLCEditWid->layout()->activate();
    NLCEditWid->move(mapToGlobal(rect().center() - NLCEditWid->rect().center()));
    NLCEditWid->show();
    if (NLCEditWid->focusWidget())
        NLCEditWid->focusWidget()->clearFocus();
}
void XYZCalibrations::nlcEditApply()
{
    NLCEditWid->hide();

    double *source = (double *)NLCEditWid->property("source").value<void *>();
    NLCWidget *status = qobject_cast<NLCWidget *>((QObject *)NLCEditWid->property("status").value<void *>());
    if (!status || !source)
        return;

    for (int i = 0; i < 6; i++)
        source[i] = lineNLCEdit[i]->value();
    status->SetPic(viewNLCEdit->m_pic);

    s_sendData();
}
void XYZCalibrations::nlcEditRst()
{
    for (int i = 0; i < 6; i++)
        lineNLCEdit[i]->setValue(0);
    lineNLCEdit[3]->setValue(1);
    nlcEditUpd();
}
void XYZCalibrations::nlcEditUpd()
{
    double source[6];
    for (int i = 0; i < 6; i++)
        source[i] = lineNLCEdit[i]->value();
    viewNLCEdit->SetPic(buildNlcPic(source));

    //check newton converges
    double CalibRange[5];
    double Shift[5];
    const double acc = 0.001;

    bool ok = true;
    int x[] = {1000, 2000, 5000, 10000, 50000};
    for (int i = 0; i < 5; i++) {
        int maxx = x[i];
        Shift[i] = NewtonResult(source, 10, 0, acc, maxx);
        CalibRange[i] = NewtonResult(source, 50, 65535, acc, maxx);
        if (i > 0) if (fabs(Shift[i-1] - Shift[i]) > acc || fabs(CalibRange[i-1] - CalibRange[i]) > acc) {
            ok = false;
            break;
        }
    }

    NLCEditValid->setText(ok ? Oscilloscope::tr("Valid curve") : Oscilloscope::tr("Invalid curve"));
    NLCEditValid->setStyleSheet(ok ? "background-color: green;" : "background-color: red;");
}

void XYZCalibrations::s_sendData()
{
    for (int i = 0; i < 3; i++) {
        for (int j = 0; j < 3; j++)
            Scanner_data_pid_cal_range[i+j*3] = linePID[i][j]->value();

        Scanner_data_pid_cal_range[45+i] = ((i < 2 ? CL_status_xy : CL_status_z) ? lineScannerRangeCL : lineScannerRangeOL)[i]->value();
        Scanner_data_scale[i] = (DWORD)qRound(lineSensorScale[i]->value());
        Scanner_data_Shift[i] = qRound(lineSensorScale0[i]->value());

        for (int j = 0; j < 6; j++) {
            Scanner_data_pid_cal_range[9 + i * 6 + j] = linePiezoCalibration[i][j];
            Scanner_data_pid_cal_range[27 + i * 6 + j] = lineSensorCalibration[i][j];
        }
    }

    free_move_speed = lineFreeSpeed->value();

    //------Second Scanner-------------------------------------------
    for (int i = 0; i < 3; i++) {
        for (int j = 0; j < 3; j++)
            Scanner2_data_pid_cal_range[i+j*3] = linePID2[i][j]->value();

        Scanner2_data_pid_cal_range[45+i] = ((i < 2 ? CL2_status_xy : CL2_status_z) ? lineScannerRangeCL2 : lineScannerRangeOL2)[i]->value();
        Scanner2_data_scale[i] = (DWORD)qRound(lineSensorScale2[i]->value());
        Scanner2_data_Shift[i] = qRound(lineSensorScale02[i]->value());

        for (int j = 0; j < 6; j++) {
            Scanner2_data_pid_cal_range[9 + i * 6 + j] = linePiezoCalibration2[i][j];
            Scanner2_data_pid_cal_range[27 + i * 6 + j] = lineSensorCalibration2[i][j];
        }
    }

    free_move_speed2 = lineFreeSpeed2->value();

    //Z3
    memset(Scanner3_data_pid_cal_range, 0, sizeof(float)*48);
    memset(Scanner3_data_scale, 0, sizeof(DWORD)*3);
    memset(Scanner3_data_Shift, 0, sizeof(long)*3);
    Scanner3_data_pid_cal_range[0] = linePIDProportionalZ3->value();
    Scanner3_data_pid_cal_range[1] = linePIDIntegralZ3->value();
    Scanner3_data_pid_cal_range[6] = linePIDProportionalZ3->value();
    Scanner3_data_pid_cal_range[7] = linePIDIntegralZ3->value();
    Scanner3_data_pid_cal_range[45] = Scanner3_data_pid_cal_range[47] =
            (CL3_status_z ? lineScannerRangeCLZ3 : lineScannerRangeOLZ3)->value();
    Scanner3_data_scale[2] = (DWORD)qRound(lineSensorScaleZ3->value());
    Scanner3_data_Shift[2] = qRound(lineSensorScaleZ03->value());
    free_move_speed3 = lineFreeSpeed3->value();

    emit sig_SentDataToEG3000();
}
void XYZCalibrations::s_updEEPData()
{
    for (int i = 0; i < 3; i++) {
        for (int j = 0; j < 3; j++)
            linePID[i][j]->setValue(Scanner_data_pid_cal_range[i+j*3]);

        if (Scanner_data_pid_cal_range[45+i] > 0) {
            lineScannerRangeOL[i]->setValue(Scanner_data_pid_cal_range[45+i]);
            lineScannerRangeCL[i]->setValue(Scanner_data_pid_cal_range[45+i]);
        } else
            Scanner_data_pid_cal_range[45+i] = ((i < 2 ? CL_status_xy : CL_status_z) ? lineScannerRangeCL : lineScannerRangeOL)[i]->value();

        lineSensorScale[i]->setValue(Scanner_data_scale[i]);
        lineSensorScale0[i]->setValue(Scanner_data_Shift[i]);

        for (int j = 0; j < 6; j++) {
            linePiezoCalibration[i][j] = Scanner_data_pid_cal_range[9 + i * 6 + j];
            lineSensorCalibration[i][j] = Scanner_data_pid_cal_range[27 + i * 6 + j];
        }
        viewPiezoCalibration[i]->SetPic(buildNlcPic(linePiezoCalibration[i]));
        viewSensorCalibration[i]->SetPic(buildNlcPic(lineSensorCalibration[i]));
    }

    lineFreeSpeed->setValue(free_move_speed);

    //------------Scanner2-----------------------
    for (int i = 0; i < 3; i++) {
        for (int j = 0; j < 3; j++)
            linePID2[i][j]->setValue(Scanner2_data_pid_cal_range[i+j*3]);

        if (Scanner2_data_pid_cal_range[45+i] > 0) {
            lineScannerRangeOL2[i]->setValue(Scanner2_data_pid_cal_range[45+i]);
            lineScannerRangeCL2[i]->setValue(Scanner2_data_pid_cal_range[45+i]);
        } else
            Scanner2_data_pid_cal_range[45+i] = ((i < 2 ? CL2_status_xy : CL2_status_z) ? lineScannerRangeCL2 : lineScannerRangeOL2)[i]->value();

        lineSensorScale2[i]->setValue(Scanner2_data_scale[i]);
        lineSensorScale02[i]->setValue(Scanner2_data_Shift[i]);

        for (int j = 0; j < 6; j++) {
            linePiezoCalibration2[i][j] = Scanner2_data_pid_cal_range[9 + i * 6 + j];
            lineSensorCalibration2[i][j] = Scanner2_data_pid_cal_range[27 + i * 6 + j];
        }
        viewPiezoCalibration2[i]->SetPic(buildNlcPic(linePiezoCalibration2[i]));
        viewSensorCalibration2[i]->SetPic(buildNlcPic(lineSensorCalibration2[i]));
    }

    lineFreeSpeed2->setValue(free_move_speed2);

    CL_NLS_Lamps(Scanner_type->currentIndex());

    sl_XYZ_range();
}

void XYZCalibrations::sl_Load_from_File()
{
    QString Filename = QFileDialog::getOpenFileName(this,
                                                    Oscilloscope::tr("Open Parameter File"), defaultFileName, Oscilloscope::tr("Parameter Files (*.par)"),
                                                    0, QFileDialog::DontUseNativeDialog);

    if (Filename.isEmpty())
        return;

    bool oldh = CBlvm[0]->isChecked();
    bool olds = CBlvm[1]->isChecked();
    nlcPIDLoad();

    if (oldh == CBlvm[0]->isChecked() && olds == CBlvm[1]->isChecked()) {
        //_wThread->setLVM(CBlvm[0]->isChecked(), CBlvm[1]->isChecked());

        read_calib_par(lastFileName = Filename);
    } else {
        lastFileName = Filename;
        lvmClicked();
    }
}

template <class T>
int readValueArray(QTextStream &s, T *array, int size, T default_value = 0)
{
    int i = 0;
    while (!s.atEnd() && i < size) {
        QStringList l = s.readLine().split('=');
        if (l.size() != 2)
            continue;

        array[i++] = l.last().toDouble();
    }
    int j = i;
    while (j < size)
        array[j++] = default_value;
    return i;
}

void XYZCalibrations::read_calib_par(QString file, bool autoload)
{
    QFile dataFile(QFile::exists(file) ? file : ":/config/default.par");
    dataFile.open(QIODevice::ReadOnly|QIODevice::Text);
    QTextStream inData(&dataFile);

    bool AutLoad;
    readValueArray(inData, &AutLoad, 1, false);
    if (autoload && AutLoad) {
        Load_from_Contr->click();
        dataFile.close();
        return; //if func called from automatic param load, and config wants us to get param from controller, we leave.
    }


    bool loaded1 = false;
    bool loaded2 = false;
    if (!second_3030) {
        QString name = file;
        name.truncate(name.lastIndexOf(".par"));
        float sz;
        {
            QString name1 = name + "_calib_head.bin";
            QString name2 = name + "_calib_head_lvm.bin";
            if (CBlvm[0]->isChecked()) {
                if (QFile::exists(name2)) name1 = name2;
            } else {
                if (!QFile::exists(name1)) name1 = name2;
            }
            if (QFile::exists(name1)) {
                QFile file(name1);
                file.open(QIODevice::ReadOnly);

                file.read((char *)Scanner_data_pid_cal_range, 48*sizeof(float));
                file.read((char *)Scanner_data_scale, 3*sizeof(DWORD));
                file.read((char *)Scanner_data_Shift, 3*sizeof(long));

                file.read((char *)&Nlc_status_ol_xy, 1); Nlc_status_ol_z = Nlc_status_ol_xy;
                file.read((char *)&Nlc_status_cl_xy, 1); Nlc_status_cl_z = Nlc_status_cl_xy;
                file.read((char *)&CL_status_xy, 1);
                file.read((char *)&CL_status_z, 1);

                for (int i = 0; i < 3; i++) {
                    file.read((char *)&sz, sizeof(float));
                    lineScannerRangeOL[i]->setValue(sz);
                }
                for (int i = 0; i < 3; i++) {
                    file.read((char *)&sz, sizeof(float));
                    lineScannerRangeCL[i]->setValue(sz);
                }
                for (int i = 0; i < 3; i++)
                    Scanner_data_pid_cal_range[45+i] = -1;

                file.read((char *)&free_move_speed, sizeof(float));

                loaded1 = (file.error() == QFile::NoError);
                file.close();
            }
        }
        {
            QString name1 = name + "_calib_stage.bin";
            QString name2 = name + "_calib_stage_lvm.bin";
            if (CBlvm[1]->isChecked()) {
                if (QFile::exists(name2)) name1 = name2;
            } else {
                if (!QFile::exists(name1)) name1 = name2;
            }
            if (QFile::exists(name1)) {
                QFile file(name1);
                file.open(QIODevice::ReadOnly);

                file.read((char *)Scanner2_data_pid_cal_range, 48*sizeof(float));
                file.read((char *)Scanner2_data_scale, 3*sizeof(DWORD));
                file.read((char *)Scanner2_data_Shift, 3*sizeof(long));

                file.read((char *)&Nlc2_status_ol_xy, 1); Nlc2_status_ol_z = Nlc2_status_ol_xy;
                file.read((char *)&Nlc2_status_cl_xy, 1); Nlc2_status_cl_z = Nlc2_status_cl_xy;
                file.read((char *)&CL2_status_xy, 1);
                file.read((char *)&CL2_status_z, 1);

                for (int i = 0; i < 3; i++) {
                    file.read((char *)&sz, sizeof(float));
                    lineScannerRangeOL2[i]->setValue(sz);
                }
                for (int i = 0; i < 3; i++) {
                    file.read((char *)&sz, sizeof(float));
                    lineScannerRangeCL2[i]->setValue(sz);
                }
                for (int i = 0; i < 3; i++)
                    Scanner2_data_pid_cal_range[45+i] = -1;

                file.read((char *)&free_move_speed2, sizeof(float));

                loaded2 = (file.error() == QFile::NoError);
                file.close();
            }
        }
    }


    if (!loaded1 || !loaded2) {
        readValueArray(inData, Scanner_data_pid_cal_range, 48);
        readValueArray(inData, Scanner_data_scale, 3, (DWORD)1);
        readValueArray(inData, Scanner_data_Shift, 3);
        for (int i = 0; i < 3; i++)
            if (Scanner_data_pid_cal_range[45+i] <= 0)
                Scanner_data_pid_cal_range[45+i] = 1;

        readValueArray(inData, &CL_status_xy, 1, false);

        readValueArray(inData, &Nlc_status_cl_xy, 1, false);
        Nlc_status_cl_z = Nlc_status_cl_xy;

        //-----------------
        readValueArray(inData, Scanner2_data_pid_cal_range, 48);
        readValueArray(inData, Scanner2_data_scale, 3, (DWORD)1);
        readValueArray(inData, Scanner2_data_Shift, 3);
        for (int i = 0; i < 3; i++)
            if (Scanner2_data_pid_cal_range[45+i] <= 0)
                Scanner2_data_pid_cal_range[45+i] = 1;

        readValueArray(inData, &CL2_status_xy, 1, false);

        readValueArray(inData, &Nlc2_status_cl_xy, 1, false);
        Nlc2_status_cl_z = Nlc2_status_cl_xy;

        //-----------------
        readValueArray(inData, &free_move_speed, 1, 100.f);
        readValueArray(inData, &free_move_speed2, 1, 100.f);

        readValueArray(inData, &CL_status_z, 1, false);

        readValueArray(inData, &Nlc_status_ol_xy, 1, false);
        Nlc_status_ol_z = Nlc_status_ol_xy;

        readValueArray(inData, &CL2_status_z, 1, false);

        readValueArray(inData, &Nlc2_status_ol_xy, 1, false);
        Nlc2_status_ol_z = Nlc2_status_ol_xy;

        double val[3];

        for (int i = 0; i < 3; i++) {
            readValueArray<double>(inData, val + i, 1, Scanner_data_pid_cal_range[45+i]);
            lineScannerRangeOL[i]->setValue(val[i]);
        }
        for (int i = 0; i < 3; i++) {
            readValueArray<double>(inData, val + i, 1, Scanner_data_pid_cal_range[45+i]);
            lineScannerRangeCL[i]->setValue(val[i]);
        }
        for (int i = 0; i < 3; i++)
            Scanner_data_pid_cal_range[45+i] = -1;

        for (int i = 0; i < 3; i++) {
            readValueArray<double>(inData, val + i, 1, Scanner2_data_pid_cal_range[45+i]);
            lineScannerRangeOL2[i]->setValue(val[i]);
        }
        for (int i = 0; i < 3; i++) {
            readValueArray<double>(inData, val + i, 1, Scanner2_data_pid_cal_range[45+i]);
            lineScannerRangeCL2[i]->setValue(val[i]);
        }
        for (int i = 0; i < 3; i++)
            Scanner2_data_pid_cal_range[45+i] = -1;
    }
    dataFile.close();

    if (!loaded1 && !loaded2 && !second_3030) { //if no binary dumps
        bool oldh = CBlvm[0]->isChecked();
        bool olds = CBlvm[1]->isChecked();
        CBlvm[0]->setChecked(false);  //always non-lvm
        CBlvm[1]->setChecked(false);
        write_calib_par(file, false); //dump to bin so that lvm won't overwrite and lose original calibrations
        CBlvm[0]->setChecked(oldh);
        CBlvm[1]->setChecked(olds);
    }


    s_updEEPData();

    if (autoload) {
        z3Load();
        s_sendData();
        sl_XYZ_range();
    } else
        s_sendData();
}

void XYZCalibrations::sl_Load_Extra_calib_from_File()
{
    //send legacy adc/dac calibs if present
    QSettings settings("Param/extracalib.conf", QSettings::IniFormat);
    if (settings.allKeys().isEmpty())
        return;

    for (int i = 0; i < 8; i++) {
        adcCalib[i*2] = settings_double(settings, "adc_gain/" + QString::number(i), 1.);
        adcCalib[i*2+1] = settings_double(settings, "adc_shift/" + QString::number(i), 0.);
    }
    for (int i = 0; i < 4; i++) {
        dacCalib[i*2] = settings_double(settings, "dac_gain/" + QString::number(i+1), 20.);
        dacCalib[i*2+1] = settings_double(settings, "dac_shift/" + QString::number(i+1), 0.);
    }

    emit sigUpdExCalibData();
}

void XYZCalibrations::eeprom_received()
{
    s_updEEPData();

    if (!eeprom_first_load)
        return;
    //skip_autocalib = !eeprom_first_load || skip_autocalib;
    eeprom_first_load = false;

    apply_calib_data();
}

void XYZCalibrations::lvmClicked()
{
    QWidgetList cwl;
    cwl << this << osc->DacsWidget;
    if (!osc->acp->needGuiInit) cwl << osc->acp->BSensorsAutoAdj << osc->acp->BPIDCalib << osc->acp->BSend << osc->acp->coeffOk;
    foreach (QWidget *wid, cwl)
        wid->setEnabled(false);

    bool upd = false;
    bool upd2 = false;
    const double nullpos = 10./(150 + 10);
    if (sender() != CBlvm[1]) { //head
        upd2 = upd2 || CL_status_xy || CL_status_z;
        CL_status_xy = false;
        CL_status_z = false;
        osc->DacsWidget->X->setValue(Scanner_data_pid_cal_range[45] * nullpos);
        osc->DacsWidget->Y->setValue(Scanner_data_pid_cal_range[46] * nullpos);
        osc->DacsWidget->Z->setValue(Scanner_data_pid_cal_range[47] * nullpos);
        upd = true;
    }
    if (sender() != CBlvm[0]) { //stage
        upd2 = upd2 || CL2_status_xy || CL2_status_z;
        CL2_status_xy = false;
        CL2_status_z = false;
        osc->DacsWidget->X2->setValue(Scanner2_data_pid_cal_range[45] * nullpos);
        osc->DacsWidget->Y2->setValue(Scanner2_data_pid_cal_range[46] * nullpos);
        osc->DacsWidget->Z2->setValue(Scanner2_data_pid_cal_range[47] * nullpos);
        upd = true;
    }
    QElapsedTimer timer;
    if (upd2) {
        CL_NLS_Lamps(Scanner_type->currentIndex());
        s_sendData();

        timer.start();
        while (_wThread->isRunning() && !timer.hasExpired(500)) {
            qApp->processEvents();
            SleepExport::msleep2(10);
        }
    }
    if (upd) {
        osc->DacsWidget->X->setNewValue();
        osc->DacsWidget->Y->setNewValue();
        osc->DacsWidget->Z->setNewValue();
        osc->DacsWidget->X2->setNewValue();
        osc->DacsWidget->Y2->setNewValue();
        osc->DacsWidget->Z2->setNewValue();

        double x_range = Scanner_data_pid_cal_range[45];//mkm
        double y_range = Scanner_data_pid_cal_range[46];//mkm
        double z_range = Scanner_data_pid_cal_range[47];//mkm
        double x_range2 = Scanner2_data_pid_cal_range[45];//mkm
        double y_range2 = Scanner2_data_pid_cal_range[46];//mkm
        double z_range2 = Scanner2_data_pid_cal_range[47];//mkm
        double time1 = qMax(x_range, qMax(y_range, z_range)) / free_move_speed;
        double time2 = qMax(x_range2, qMax(y_range2, z_range2)) / free_move_speed2;
        double timem = qMax(time1, time2);

        while (_wThread->isRunning() && _wThread->Action_5[0] && _wThread->Action_5[1]) {
            qApp->processEvents();
            SleepExport::msleep2(1);
        }

        int time = 100 + 1000 * qMin(10., timem);
        timer.start();
        while (_wThread->isRunning() && !timer.hasExpired(time)) {
            qApp->processEvents();
            SleepExport::msleep2(10);
        }
    }

    _wThread->setLVM(CBlvm[0]->isChecked(), CBlvm[1]->isChecked());
    while (_wThread->isRunning() && _wThread->updLVM) {
        qApp->processEvents();
        SleepExport::msleep2(1);
    }

    timer.start();
    while (_wThread->isRunning() && !timer.hasExpired(100)) {
        qApp->processEvents();
        SleepExport::msleep2(10);
    }


    read_calib_par(lastFileName);

    foreach (QWidget *wid, cwl)
        wid->setEnabled(true);

    apply_calib_data();
}

void XYZCalibrations::apply_calib_data()
{
    osc->DacsWidget->X->setValue(Scanner_data_pid_cal_range[45]/2);
    osc->DacsWidget->Y->setValue(Scanner_data_pid_cal_range[46]/2);
    osc->DacsWidget->Z->setValue(Scanner_data_pid_cal_range[47]/2);
    osc->DacsWidget->X->setNewValue();
    osc->DacsWidget->Y->setNewValue();
    osc->DacsWidget->Z->setNewValue();

    osc->DacsWidget->X2->setValue(Scanner2_data_pid_cal_range[45]/2);
    osc->DacsWidget->Y2->setValue(Scanner2_data_pid_cal_range[46]/2);
    osc->DacsWidget->Z2->setValue(Scanner2_data_pid_cal_range[47]/2);
    osc->DacsWidget->X2->setNewValue();
    osc->DacsWidget->Y2->setNewValue();
    osc->DacsWidget->Z2->setNewValue();

    osc->DacsWidget->Z3->setValue(Scanner3_data_pid_cal_range[47]/2);
    osc->DacsWidget->Z3->setNewValue();


    bool haveHead = false, haveStage = false;
    if (!second_3030)
    if (I_MODULE *eg3km = ModuleManager::instance->find1("eg3000")) {
        USB_READ_THREAD *_rThread = &((USB_THREAD_GROUP *)eg3km->basecast())->rThread;

        int time = 1000;
        QElapsedTimer timer;
        timer.start();
        while (_rThread->isRunning() && !_rThread->scannersConnectedUpdated && !_rThread->testMode && !timer.hasExpired(time)) {
            qApp->processEvents();
            SleepExport::msleep2(1);
        }
        if (timer.hasExpired(time))
            goto skip;

        haveHead = _rThread->headConnected;
        haveStage = _rThread->stageConnected;

        {
            QSettings settings("Param/gui.conf", QSettings::IniFormat);
            bool head_on = settings.value("display/oscilloscope/head", true).toBool();
            bool stage_on = settings.value("display/oscilloscope/stage", true).toBool();
            if (!head_on && haveHead) {
                haveHead = false;
                new Notification(tr("Head is connected, but disabled in settings."), Notification::Warn, 20);
            }
            if (!stage_on && haveStage) {
                haveStage = false;
                new Notification(tr("Stage is connected, but disabled in settings."), Notification::Warn, 20);
            }
        }
    }

    if (skip_autocalib) {
skip:
        skip_autocalib = false;
        return;
    }


    QSettings settings("Param/gui.conf", QSettings::IniFormat);
    bool sensors = settings.value("autocalib_sensors", false).toBool();
    bool pid = settings.value("autocalib_pid", false).toBool();

    bool need_something = (sensors && (second_3030 || haveHead || haveStage)) ||
            (pid && !second_3030 && (haveHead || haveStage));
    if (need_something) { //wait required, as next calibs send will stop active movements for another scanners;
        double x_range = Scanner_data_pid_cal_range[45];//mkm
        double y_range = Scanner_data_pid_cal_range[46];//mkm
        double z_range = Scanner_data_pid_cal_range[47];//mkm
        double x_range2 = Scanner2_data_pid_cal_range[45];//mkm
        double y_range2 = Scanner2_data_pid_cal_range[46];//mkm
        double z_range2 = Scanner2_data_pid_cal_range[47];//mkm
        double time1 = qMax(x_range, qMax(y_range, z_range)) / free_move_speed;
        double time2 = qMax(x_range2, qMax(y_range2, z_range2)) / free_move_speed2;
        double timem = qMax(time1, time2) / 2;

        while (_wThread->isRunning() && _wThread->Action_5[0] && _wThread->Action_5[1]) {
            qApp->processEvents();
            SleepExport::msleep2(1);
        }

        int time = 100 + 1000 * qMin(10., timem);
        QElapsedTimer timer;
        timer.start();
        while (_wThread->isRunning() && !timer.hasExpired(time)) {
            qApp->processEvents();
            SleepExport::msleep2(10);
        }
    }

    if (sensors) {
        if (!second_3030 && haveHead) {
            osc->acp->CBScanner->setCurrentIndex(0);
            osc->acp->SensorReCalc(true);
        }
        if (second_3030 || haveStage) {
            osc->acp->CBScanner->setCurrentIndex(1);
            osc->acp->SensorReCalc(true);
        }
    }
    if (pid) {
        if (!second_3030 && haveHead) {
            osc->acp->CBScanner->setCurrentIndex(0);
            osc->acp->PICalcAuto(true);
        }
        if (!second_3030 && haveStage) {
            osc->acp->CBScanner->setCurrentIndex(1);
            osc->acp->PICalcAuto(true);
        }
    }
}

void XYZCalibrations::AutoDataLoad()
{
    nlcPIDLoad();

    _wThread->setLVM(CBlvm[0]->isChecked(), CBlvm[1]->isChecked());

    QSettings settings("Param/controller.conf", QSettings::IniFormat);
    if (eeprom_first_load = settings.value("load_calib_from_flash", false).toBool()) {
        Load_from_Contr->click();
    } else {
        read_calib_par(lastFileName = defaultFileName, true);
        apply_calib_data();
    }
}

void XYZCalibrations::save_To_File()
{
    QString Filename = QFileDialog::getSaveFileName(this,
                                                    Oscilloscope::tr("Save Parameter File"), defaultFileName, Oscilloscope::tr("Parameter Files (*.par)"),
                                                    0, QFileDialog::DontUseNativeDialog);

    if (Filename.isEmpty())
        return;

    nlcPIDSave();

    write_calib_par(lastFileName = Filename);
}

template <class T>
void writeValueArray(QTextStream &s, T *array, int size)
{
    for (int i = 0; i < size; i++)
        s << "v=" << array[i] << endl;
}

void XYZCalibrations::write_calib_par(QString filename, bool need_old)
{
    if (need_old) {
        QFile dataFile(filename);
        dataFile.open(QIODevice::WriteOnly | QIODevice::Text);

        QTextStream outData(&dataFile);
        outData.setRealNumberPrecision(11);

        outData << "NanoScanTechnologies ltd." << endl;
        outData << "Compressed Parameter File for scanner parameters" << endl << endl;

        bool AutLoad = false;
        writeValueArray(outData, &AutLoad, 1);

        writeValueArray(outData, Scanner_data_pid_cal_range, 48);
        writeValueArray(outData, Scanner_data_scale, 3);
        writeValueArray(outData, Scanner_data_Shift, 3);

        writeValueArray(outData, &CL_status_xy, 1);

        if (Nlc_status_cl_z) Nlc_status_cl_xy = true;
        writeValueArray(outData, &Nlc_status_cl_xy, 1);

        //----------------
        writeValueArray(outData, Scanner2_data_pid_cal_range, 48);
        writeValueArray(outData, Scanner2_data_scale, 3);
        writeValueArray(outData, Scanner2_data_Shift, 3);

        writeValueArray(outData, &CL2_status_xy, 1);

        if (Nlc2_status_cl_z) Nlc2_status_cl_xy = true;
        writeValueArray(outData, &Nlc2_status_cl_xy, 1);

        //-----------------
        writeValueArray(outData, &free_move_speed, 1);
        writeValueArray(outData, &free_move_speed2, 1);

        writeValueArray(outData, &CL_status_z, 1);

        if (Nlc_status_ol_z) Nlc_status_ol_xy = true;
        writeValueArray(outData, &Nlc_status_ol_xy, 1);

        writeValueArray(outData, &CL2_status_z, 1);

        if (Nlc2_status_ol_z) Nlc2_status_ol_xy = true;
        writeValueArray(outData, &Nlc2_status_ol_xy, 1);

        double tmp[3];

        for (int i = 0; i < 3; i++)
            tmp[i] = lineScannerRangeOL[i]->value();
        writeValueArray(outData, tmp, 3);
        for (int i = 0; i < 3; i++)
            tmp[i] = lineScannerRangeCL[i]->value();
        writeValueArray(outData, tmp, 3);

        for (int i = 0; i < 3; i++)
            tmp[i] = lineScannerRangeOL2[i]->value();
        writeValueArray(outData, tmp, 3);
        for (int i = 0; i < 3; i++)
            tmp[i] = lineScannerRangeCL2[i]->value();
        writeValueArray(outData, tmp, 3);

        dataFile.close();
    }

    //binary dump
    QString name = filename;
    name.truncate(name.lastIndexOf(".par"));
    float sz;
    if (second_3030) return;
    {
        QFile file(name + "_calib_head" + (!CBlvm[0]->isChecked() ? "" : "_lvm") + ".bin");
        file.open(QIODevice::WriteOnly | QIODevice::Truncate);
        file.write((char *)Scanner_data_pid_cal_range, 48*sizeof(float));
        file.write((char *)Scanner_data_scale, 3*sizeof(DWORD));
        file.write((char *)Scanner_data_Shift, 3*sizeof(long));

        file.write((char *)&Nlc_status_ol_xy, 1);
        file.write((char *)&Nlc_status_cl_xy, 1);
        file.write((char *)&CL_status_xy, 1);
        file.write((char *)&CL_status_z, 1);

        for (int i = 0; i < 3; i++) {
            sz = lineScannerRangeOL[i]->value();
            file.write((char *)&sz, sizeof(float));
        }
        for (int i = 0; i < 3; i++) {
            sz = lineScannerRangeCL[i]->value();
            file.write((char *)&sz, sizeof(float));
        }

        file.write((char *)&free_move_speed, sizeof(float));

        file.close();
    }
    {
        QFile file(name + "_calib_stage" + (!CBlvm[1]->isChecked() ? "" : "_lvm") + ".bin");
        file.open(QIODevice::WriteOnly | QIODevice::Truncate);
        file.write((char *)Scanner2_data_pid_cal_range, 48*sizeof(float));
        file.write((char *)Scanner2_data_scale, 3*sizeof(DWORD));
        file.write((char *)Scanner2_data_Shift, 3*sizeof(long));

        file.write((char *)&Nlc2_status_ol_xy, 1);
        file.write((char *)&Nlc2_status_cl_xy, 1);
        file.write((char *)&CL2_status_xy, 1);
        file.write((char *)&CL2_status_z, 1);

        for (int i = 0; i < 3; i++) {
            sz = lineScannerRangeOL2[i]->value();
            file.write((char *)&sz, sizeof(float));
        }
        for (int i = 0; i < 3; i++) {
            sz = lineScannerRangeCL2[i]->value();
            file.write((char *)&sz, sizeof(float));
        }

        file.write((char *)&free_move_speed2, sizeof(float));

        file.close();
    }
}

void XYZCalibrations::z3Save()
{
    QSettings settings("Param/z3calib.conf", QSettings::IniFormat);

    settings.setValue("cl_z", CL3_status_z);
    settings.setValue("ol_speed", lineFreeSpeed3->value());

    settings.setValue("pid_p_z", linePIDProportionalZ3->value());
    settings.setValue("pid_i_z", linePIDIntegralZ3->value());
    settings.setValue("range_z", lineScannerRangeOLZ3->value());
    settings.setValue("range_cl_z", lineScannerRangeCLZ3->value());
    settings.setValue("cap_gain_z", lineSensorScaleZ3->value());
    settings.setValue("cap_shift_z", lineSensorScaleZ03->value());
}
void XYZCalibrations::z3Load()
{
    QSettings settings("Param/z3calib.conf", QSettings::IniFormat);

    CL3_status_z = settings.value("cl_z", false).toBool();
    lineFreeSpeed3->setValue(settings.value("ol_speed", 100).toDouble());

    linePIDProportionalZ3->setValue(settings.value("pid_p_z", 0.05).toDouble());
    linePIDIntegralZ3->setValue(settings.value("pid_i_z", 0.001).toDouble());
    lineScannerRangeOLZ3->setValue(settings.value("range_z", 100).toDouble());
    lineScannerRangeCLZ3->setValue(settings.value("range_cl_z", 100).toDouble());
    lineSensorScaleZ3->setValue(settings.value("cap_gain_z", 100).toDouble());
    lineSensorScaleZ03->setValue(settings.value("cap_shift_z", 0).toDouble());
}


void XYZCalibrations::sl_XYZ_range()
{
    emit sig_XYZ_range(Scanner_data_pid_cal_range[45], Scanner_data_pid_cal_range[46], Scanner_data_pid_cal_range[47]);
    emit sig_XYZ_range2(Scanner2_data_pid_cal_range[45], Scanner2_data_pid_cal_range[46], Scanner2_data_pid_cal_range[47]);
    emit sig_XYZ_range3(Scanner3_data_pid_cal_range[45], Scanner3_data_pid_cal_range[46], Scanner3_data_pid_cal_range[47]);

    sl_setCalibVC();
}

void XYZCalibrations::sl_setCalibVC()
{
    const int vch[3] = {srcDACX, srcDACX2, srcDACZ3-2};
    const int cch[3] = {srcCX, srcCX2, srcCZ3-2};
    bool * const vcl[3][3] = {{&CL_status_xy, &CL_status_xy, &CL_status_z},
                              {&CL2_status_xy, &CL2_status_xy, &CL2_status_z},
                              {&CL3_status_z, &CL3_status_z, &CL3_status_z}};
    DoubleEdit ** const vrange[3] = {lineScannerRangeOL, lineScannerRangeOL2, (&lineScannerRangeOLZ3)-2};
    DoubleEdit ** const crange[3] = {lineScannerRangeCL, lineScannerRangeCL2, (&lineScannerRangeCLZ3)-2};
    for (int i = 0; i < 3; i++) {
        for (int j = (i != 2 ? 0 : 2); j < 3; j++) {
            double vgain = (150+15) / 20.;
            double vshift = 10 - 20 * (15./(150+15)); //-15v shift
            const bool use_v_meters = !(*vcl[i][j]);
            if (use_v_meters) {
                vgain = vrange[i][j]->value()*1e-6 / 20.;
                vshift = 10;
            }
            input_gain_set(vch[i] + j, vgain);
            input_shift_set(vch[i] + j, vshift);
            input_unit_set(vch[i] + j, use_v_meters ? QObject::tr("m") : QObject::tr("V"));

            double cgain = crange[i][j]->value()*1e-6 / 20.;
            double cshift = 10;
            input_gain_set(cch[i] + j, cgain);
            input_shift_set(cch[i] + j, cshift);
            input_unit_set(cch[i] + j, QObject::tr("m"));
        }
    }

    emit input_calib_updated();
}


void XYZCalibrations::capBits(int n, long bits)
{
    if (n == 0) {
        bool new_csb = bool(bits & 2);
        if (new_csb) {
            if (!second_3030)
                _wThread->csb_speed_extended = true;
            sl_csb_speed(CBSpeedCSB->currentIndex());
        }
    }
}

void XYZCalibrations::reset()
{
    sl_Load_Extra_calib_from_File();
    AutoDataLoad();

    QSettings settings("Param/gui.conf", QSettings::IniFormat);
    if (!settings.value("display/spm", true).toBool())
        if (!settings.value("display/oscilloscope/head", true).toBool())
            _wThread->Set_FB_Scanner(scStage);
}
